﻿//*********************************************************
//
//    Copyright (c) Microsoft. All rights reserved.
//    This code is licensed under the Apache License
//    Version 2.0.
//
//    THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
//    ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
//    IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
//    PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.
//
//*********************************************************

using System;
using UProveCrypto.Math;

namespace UProveCrypto
{
    /// <summary>
    /// Class for U-Prove subgoup paramters.  See ParameterSet for more documentation.
    /// </summary>
    public static class ECParameterSets
    {
        /// <summary> OID for P-256</summary>
        public const string ParamSet_EC_P256_V1Name = "1.3.6.1.4.1.311.75.1.2.1";
        /// <summary> OID for P-384</summary>
        public const string ParamSet_EC_P384_V1Name = "1.3.6.1.4.1.311.75.1.2.2";
        /// <summary> OID for P-521</summary>
        public const string ParamSet_EC_P521_V1Name = "1.3.6.1.4.1.311.75.1.2.3";

        static ECParameterSets()
        {
            
            ECParameterSets.ParamSet_EC_P256_V1 = new ParameterSet(
                ECParameterSets.ParamSet_EC_P256_V1Name,
                P256.Gq,
                P256.IssuerGenerators,
                P256.Gd);

            ECParameterSets.ParamSet_EC_P384_V1 = new ParameterSet(
                ECParameterSets.ParamSet_EC_P384_V1Name,
                P384.Gq,
                P384.IssuerGenerators,
                P384.Gd);


            ECParameterSets.ParamSet_EC_P521_V1 = new ParameterSet(
                ECParameterSets.ParamSet_EC_P521_V1Name,
                P521.Gq,
                P521.IssuerGenerators,
                P521.Gd);
        }
        
        /// <summary>Recommended EC parameter set P-256.</summary>
        public static ParameterSet ParamSet_EC_P256_V1 { get; private set; }
        /// <summary>Recommended EC parameter set P-384.</summary>
        public static ParameterSet ParamSet_EC_P384_V1 { get; private set; }
        /// <summary>Recommended EC parameter set P-521.</summary>
        public static ParameterSet ParamSet_EC_P521_V1 { get; private set; }

        //
        // P-256
        //
        static class P256
        {

static readonly byte[] p = {
0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x01,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

static readonly byte[] a = {
0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x01,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc
};
static readonly byte[] b = {
0x5a, 0xc6, 0x35, 0xd8, 0xaa, 0x3a, 0x93, 0xe7,
0xb3, 0xeb, 0xbd, 0x55, 0x76, 0x98, 0x86, 0xbc,
0x65, 0x1d, 0x06, 0xb0, 0xcc, 0x53, 0xb0, 0xf6,
0x3b, 0xce, 0x3c, 0x3e, 0x27, 0xd2, 0x60, 0x4b
};

static readonly byte[] n = {
0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xbc, 0xe6, 0xfa, 0xad, 0xa7, 0x17, 0x9e, 0x84,
0xf3, 0xb9, 0xca, 0xc2, 0xfc, 0x63, 0x25, 0x51
};

static readonly byte[] h = {
0x01
};

// g
static readonly byte[] g_x = {
0x6b, 0x17, 0xd1, 0xf2, 0xe1, 0x2c, 0x42, 0x47,
0xf8, 0xbc, 0xe6, 0xe5, 0x63, 0xa4, 0x40, 0xf2,
0x77, 0x03, 0x7d, 0x81, 0x2d, 0xeb, 0x33, 0xa0,
0xf4, 0xa1, 0x39, 0x45, 0xd8, 0x98, 0xc2, 0x96
};
static readonly byte[] g_y = {
0x4f, 0xe3, 0x42, 0xe2, 0xfe, 0x1a, 0x7f, 0x9b,
0x8e, 0xe7, 0xeb, 0x4a, 0x7c, 0x0f, 0x9e, 0x16,
0x2b, 0xce, 0x33, 0x57, 0x6b, 0x31, 0x5e, 0xce,
0xcb, 0xb6, 0x40, 0x68, 0x37, 0xbf, 0x51, 0xf5
};

// g1 (counter = 5)
static readonly byte[] g1_x = {
0xf1, 0xb9, 0x86, 0xd5, 0xd1, 0x1f, 0x43, 0x48,
0x3a, 0xe7, 0x36, 0xe8, 0x86, 0xaf, 0x75, 0x0e,
0x87, 0x0d, 0x7f, 0x0c, 0x23, 0x12, 0xaa, 0xd8,
0xdb, 0x5c, 0x8a, 0x3e, 0x34, 0xf5, 0x39, 0x1e
};
static readonly byte[] g1_y = {
0x64, 0x34, 0x7b, 0x7f, 0x49, 0x31, 0x87, 0xa5,
0x3b, 0x37, 0x08, 0x94, 0xb8, 0xf8, 0xe3, 0x8f,
0xd2, 0x2c, 0xb9, 0x93, 0x02, 0x39, 0x3d, 0x79,
0xdc, 0xe2, 0x25, 0x91, 0x8e, 0xba, 0x61, 0xee
};

// g2 (counter = 1)
static readonly byte[] g2_x = {
0x15, 0x54, 0xcf, 0x98, 0x3e, 0x0b, 0x06, 0x0c,
0x78, 0x70, 0x5e, 0xd7, 0xd1, 0x4a, 0x49, 0x41,
0xb0, 0x2e, 0x60, 0x8c, 0xdb, 0x78, 0xf6, 0xa7,
0x5a, 0x52, 0x34, 0x59, 0x78, 0x14, 0x1f, 0xd3
};
static readonly byte[] g2_y = {
0x62, 0x54, 0x0e, 0x69, 0x0c, 0x8f, 0xa9, 0xfe,
0x10, 0x7e, 0x21, 0x41, 0xdf, 0xc6, 0x90, 0x7f,
0x74, 0xf5, 0xfe, 0xeb, 0xdf, 0x5b, 0x12, 0xd7,
0x15, 0x3b, 0x46, 0x35, 0xa2, 0xdf, 0x6a, 0x76
};

// g3 (counter = 0)
static readonly byte[] g3_x = {
0x32, 0x79, 0x1a, 0x77, 0x9e, 0x9a, 0xa4, 0x75,
0xba, 0x26, 0x66, 0xa0, 0xe4, 0x7a, 0x92, 0x8b,
0x21, 0xab, 0x19, 0x05, 0xfa, 0xaf, 0x48, 0xbb,
0x80, 0x62, 0xba, 0xe9, 0x00, 0x9e, 0xb2, 0x7d
};
static readonly byte[] g3_y = {
0x18, 0x74, 0xba, 0x86, 0xea, 0x19, 0x4f, 0xb1,
0x4d, 0xcc, 0xe9, 0xfa, 0x22, 0x36, 0x6f, 0x47,
0x35, 0xca, 0xea, 0x21, 0x19, 0xbe, 0xb6, 0x3f,
0x2b, 0xae, 0xc1, 0x9a, 0x9e, 0x93, 0xa5, 0x45
};

// g4 (counter = 2)
static readonly byte[] g4_x = {
0xc0, 0xef, 0xad, 0xb5, 0xc3, 0x01, 0x5e, 0x42,
0xc1, 0xd7, 0x1a, 0xc3, 0x90, 0xc4, 0xd2, 0x2a,
0x6f, 0x5d, 0x55, 0x2f, 0x63, 0xbb, 0xcc, 0x59,
0x19, 0x0a, 0xea, 0x6a, 0xee, 0x16, 0x35, 0x4a
};
static readonly byte[] g4_y = {
0x53, 0xf0, 0x13, 0x3e, 0xa4, 0x4d, 0xa2, 0x0c,
0x50, 0x9a, 0x4e, 0x5b, 0xe9, 0xb0, 0x27, 0xdb,
0xe1, 0x3e, 0x3a, 0x60, 0x43, 0x9d, 0xbe, 0x72,
0x08, 0x4b, 0x0c, 0x75, 0xa0, 0x49, 0x72, 0x3f
};

// g5 (counter = 1)
static readonly byte[] g5_x = {
0xbd, 0x5f, 0x29, 0xdf, 0x66, 0x40, 0x49, 0x3f,
0xf9, 0x6c, 0x6c, 0xbc, 0x49, 0xcb, 0x8e, 0x5f,
0x61, 0x46, 0x27, 0x92, 0xdb, 0x75, 0xf2, 0x0e,
0xf4, 0x9b, 0xf8, 0x6e, 0x26, 0x0d, 0xc9, 0x55
};
static readonly byte[] g5_y = {
0x20, 0x4c, 0x44, 0x0e, 0xf8, 0xc6, 0xeb, 0x2b,
0xec, 0x0c, 0x34, 0x3a, 0xce, 0x9c, 0x6d, 0x64,
0xe1, 0x88, 0xc8, 0xb4, 0xf0, 0x61, 0x3d, 0x64,
0x84, 0x6a, 0xdb, 0xdc, 0x3d, 0x8f, 0xdf, 0xad
};

// g6 (counter = 1)
static readonly byte[] g6_x = {
0xd9, 0x1a, 0xbd, 0xa2, 0x6e, 0xc5, 0xc3, 0x00,
0x1c, 0xf1, 0xca, 0x2c, 0x09, 0xad, 0x88, 0x66,
0x25, 0x58, 0x42, 0x6d, 0xc3, 0xb4, 0xd1, 0xb5,
0x01, 0xe7, 0xab, 0xc2, 0xdb, 0x08, 0x0c, 0xdc
};
static readonly byte[] g6_y = {
0x54, 0xeb, 0xb1, 0x7f, 0xed, 0x85, 0x5a, 0x36,
0xc1, 0xf7, 0x4a, 0xb8, 0x25, 0x62, 0x08, 0xe8,
0x63, 0x07, 0xa9, 0xf2, 0xb7, 0x56, 0xd7, 0xc8,
0x4b, 0x4f, 0xb9, 0x48, 0x5e, 0x0f, 0xf5, 0xf5
};

// g7 (counter = 4)
static readonly byte[] g7_x = {
0x86, 0xeb, 0x2c, 0x94, 0xe2, 0xb6, 0xd6, 0x20,
0xa3, 0x91, 0xb4, 0x08, 0x0d, 0xfe, 0x2b, 0x37,
0x7c, 0xc2, 0x0d, 0x98, 0x1b, 0x5b, 0xc0, 0xcc,
0xa9, 0x4e, 0x86, 0x56, 0x97, 0x95, 0x9e, 0xbe
};
static readonly byte[] g7_y = {
0x26, 0xac, 0x15, 0x89, 0xc5, 0x28, 0x80, 0xc3,
0xb8, 0xf8, 0x1d, 0x2b, 0xf3, 0x29, 0x76, 0x63,
0x60, 0x19, 0xf1, 0x6d, 0x8e, 0xfa, 0x1f, 0x4d,
0x20, 0x95, 0x0b, 0x99, 0x08, 0xce, 0xb7, 0xe1
};

// g8 (counter = 0)
static readonly byte[] g8_x = {
0x55, 0x53, 0x14, 0x8e, 0x44, 0x25, 0x26, 0x92,
0xd9, 0xe7, 0xea, 0x9c, 0x18, 0x94, 0x69, 0xdd,
0x2c, 0x0e, 0x8b, 0xd4, 0x49, 0x40, 0x5b, 0x6f,
0x3b, 0x1f, 0x27, 0x92, 0x45, 0xb3, 0x7f, 0x0d
};
static readonly byte[] g8_y = {
0x79, 0x0c, 0xa4, 0xce, 0x90, 0xe0, 0x48, 0xa7,
0x42, 0x5b, 0x66, 0x2a, 0x63, 0x16, 0x12, 0xd0,
0x22, 0x4f, 0x20, 0x8e, 0x4b, 0xe6, 0xe9, 0x07,
0xc3, 0xe7, 0xd9, 0x60, 0x7a, 0x99, 0x7f, 0x6d
};

// g9 (counter = 0)
static readonly byte[] g9_x = {
0x77, 0x66, 0x8d, 0x97, 0xbf, 0xf7, 0xd5, 0xda,
0x69, 0x5d, 0x6d, 0x72, 0xe4, 0xf8, 0x40, 0x20,
0x5d, 0xe2, 0x89, 0xce, 0x8f, 0xf1, 0xe9, 0x95,
0x24, 0x35, 0xb0, 0xb4, 0xdd, 0x4e, 0x22, 0x2e
};
static readonly byte[] g9_y = {
0x14, 0x76, 0x06, 0x0b, 0x33, 0xfe, 0x63, 0x6b,
0xb9, 0xb7, 0x5f, 0x10, 0x78, 0x5d, 0x4b, 0x43,
0x19, 0x05, 0xcd, 0x00, 0x6f, 0x83, 0x2b, 0xf7,
0x31, 0x03, 0xb9, 0xf8, 0x80, 0x37, 0x85, 0x56
};

// g10 (counter = 0)
static readonly byte[] g10_x = {
0x72, 0x9a, 0x72, 0xbe, 0x83, 0x75, 0x88, 0x8f,
0x67, 0xdf, 0x96, 0xd2, 0xa5, 0x2e, 0x1b, 0x38,
0x4a, 0xf1, 0xc6, 0x8f, 0xf8, 0xb7, 0x3c, 0xad,
0xf6, 0x29, 0x6c, 0x72, 0xc2, 0xc1, 0xfa, 0xb2
};
static readonly byte[] g10_y = {
0x01, 0x31, 0x20, 0xe6, 0x94, 0x2d, 0x07, 0x40,
0xa2, 0x5f, 0x8b, 0x87, 0x1e, 0x1f, 0x2f, 0xe9,
0xa8, 0x60, 0x49, 0x77, 0xd1, 0xda, 0xa1, 0x8a,
0xf0, 0xe4, 0xfe, 0xd5, 0x70, 0xc6, 0xea, 0x2e
};

// g11 (counter = 0)
static readonly byte[] g11_x = {
0xcf, 0xba, 0x01, 0x4e, 0xf2, 0x73, 0x4b, 0xb0,
0xd5, 0x18, 0x63, 0xa1, 0xe6, 0xae, 0x8e, 0xb4,
0xae, 0x18, 0x9f, 0x8c, 0x19, 0x43, 0x2a, 0xf4,
0x6d, 0x9f, 0x16, 0xfd, 0xd4, 0x3f, 0xbc, 0x18
};
static readonly byte[] g11_y = {
0x12, 0x56, 0xc7, 0x84, 0xf8, 0x27, 0xc3, 0x1a,
0xd2, 0x3d, 0x8d, 0x23, 0x36, 0x78, 0xce, 0x2e,
0xeb, 0xce, 0x34, 0x46, 0x29, 0xe7, 0xa5, 0xf7,
0xa6, 0xd9, 0x4a, 0xdc, 0x0f, 0xf4, 0x7a, 0x7e
};

// g12 (counter = 7)
static readonly byte[] g12_x = {
0x6c, 0x14, 0x07, 0xc4, 0x9a, 0x51, 0xf6, 0x76,
0x25, 0xeb, 0x8b, 0x29, 0x95, 0xac, 0x11, 0x94,
0x42, 0x88, 0x99, 0x5b, 0x3a, 0x81, 0x78, 0x9a,
0x5e, 0xb3, 0xe6, 0xbf, 0x4f, 0x2d, 0xed, 0x78
};
static readonly byte[] g12_y = {
0x16, 0xd8, 0x72, 0x49, 0x4f, 0xc1, 0x8d, 0x77,
0x40, 0x4f, 0x90, 0x6e, 0x58, 0x90, 0x21, 0x50,
0xe1, 0xfc, 0xdd, 0xa0, 0xcf, 0x21, 0x15, 0x16,
0xf6, 0xf1, 0x94, 0x15, 0xe8, 0x89, 0x2f, 0x26
};

// g13 (counter = 1)
static readonly byte[] g13_x = {
0xd9, 0x23, 0x1c, 0x31, 0x5b, 0xaf, 0x72, 0x24,
0x69, 0xf7, 0x4f, 0xba, 0x55, 0xba, 0x66, 0x17,
0x77, 0xe9, 0x1c, 0xa6, 0x32, 0x0a, 0x88, 0x25,
0xbd, 0xa1, 0xcb, 0xf0, 0xea, 0x20, 0x60, 0x92
};
static readonly byte[] g13_y = {
0x36, 0xe4, 0xcd, 0x12, 0x88, 0x08, 0x8d, 0xec,
0xee, 0xa8, 0xe7, 0xb6, 0xd2, 0x2c, 0xfd, 0x97,
0xb9, 0x9f, 0x87, 0xfa, 0xcc, 0x95, 0xf1, 0x89,
0x1f, 0xc6, 0xa2, 0x8b, 0xd8, 0x1e, 0x5f, 0x50
};

// g14 (counter = 1)
static readonly byte[] g14_x = {
0x35, 0x35, 0x87, 0x11, 0x38, 0x41, 0x06, 0xb8,
0x62, 0xa2, 0xcf, 0x0b, 0x40, 0x3e, 0x80, 0x55,
0x92, 0x0c, 0x75, 0x98, 0xbf, 0xb4, 0x99, 0x87,
0xa8, 0x9c, 0x35, 0x69, 0xe5, 0xa0, 0x5b, 0x61
};
static readonly byte[] g14_y = {
0x18, 0xed, 0xfa, 0x1d, 0xfc, 0x65, 0x3a, 0x05,
0x74, 0xca, 0x88, 0xfd, 0xaa, 0xec, 0xdf, 0xe9,
0xeb, 0x75, 0x30, 0x9a, 0xac, 0xbe, 0x92, 0x6c,
0x21, 0x10, 0xe9, 0x26, 0x78, 0xc8, 0x4e, 0x3d
};

// g15 (counter = 1)
static readonly byte[] g15_x = {
0x25, 0xd0, 0x5c, 0x26, 0x17, 0x72, 0x16, 0x6c,
0x08, 0x48, 0x3d, 0x00, 0x00, 0x3f, 0x44, 0x35,
0x20, 0xe9, 0x13, 0x24, 0xcb, 0xe9, 0x18, 0xfc,
0x34, 0x00, 0x8a, 0x93, 0x27, 0x16, 0xd7, 0xeb
};
static readonly byte[] g15_y = {
0x66, 0x8a, 0x13, 0xc5, 0xd1, 0x63, 0xf6, 0x64,
0x6b, 0xf2, 0xe8, 0xf4, 0x2d, 0x1f, 0x48, 0xe7,
0x9a, 0x9e, 0xad, 0x02, 0x09, 0x22, 0xb3, 0x83,
0x00, 0x6b, 0x67, 0x6d, 0x29, 0xd3, 0x5a, 0x42
};

// g16 (counter = 1)
static readonly byte[] g16_x = {
0xfc, 0x03, 0x5c, 0x85, 0xaa, 0x0e, 0x9c, 0x52,
0x7e, 0xa7, 0xdc, 0xa2, 0x6a, 0x2d, 0xb7, 0x4d,
0xc2, 0x50, 0xe8, 0xa5, 0xab, 0xe8, 0x53, 0xbb,
0xde, 0xd1, 0x59, 0x59, 0xd7, 0x23, 0x0f, 0x43
};
static readonly byte[] g16_y = {
0x65, 0xf0, 0x52, 0xa3, 0x82, 0xb2, 0xc7, 0x8c,
0xaa, 0x9f, 0xcf, 0xc9, 0x52, 0x09, 0x6f, 0x4c,
0xcc, 0x47, 0x72, 0x54, 0x6e, 0x57, 0x98, 0x64,
0x91, 0x23, 0xfe, 0xf9, 0x4e, 0xc9, 0x5a, 0xcc
};

// g17 (counter = 1)
static readonly byte[] g17_x = {
0x85, 0xb3, 0x87, 0x3f, 0xd9, 0x11, 0xbf, 0x06,
0xa9, 0x78, 0xfa, 0x40, 0xe2, 0x61, 0xe1, 0xc8,
0x56, 0xf6, 0x38, 0xca, 0x9e, 0xc8, 0xcb, 0xe8,
0x82, 0x6a, 0x60, 0x82, 0xc8, 0x45, 0x2d, 0x0f
};
static readonly byte[] g17_y = {
0x3c, 0xf0, 0x0d, 0x69, 0x58, 0x6f, 0x56, 0xbe,
0xd8, 0x49, 0xd5, 0xe9, 0xe2, 0x82, 0x5a, 0x00,
0x3c, 0xe5, 0x62, 0xaa, 0xb5, 0xf8, 0x1b, 0xd7,
0x18, 0xa4, 0xe9, 0x41, 0x98, 0x9e, 0x11, 0x01
};

// g18 (counter = 0)
static readonly byte[] g18_x = {
0x45, 0x49, 0xf8, 0xc6, 0x21, 0xea, 0xba, 0x57,
0xed, 0x23, 0x36, 0xd5, 0x19, 0x20, 0xf6, 0xfc,
0x4d, 0xc3, 0x4e, 0x04, 0x7d, 0xb1, 0x34, 0xc6,
0x19, 0x80, 0xe4, 0xe3, 0x58, 0xc5, 0xe3, 0x24
};
static readonly byte[] g18_y = {
0x39, 0xe8, 0xbe, 0x23, 0xf0, 0x40, 0x33, 0xa0,
0xf8, 0xbc, 0x43, 0xd5, 0xa1, 0x1b, 0x1e, 0x79,
0x8d, 0x25, 0xb5, 0xc7, 0x5d, 0x74, 0x0e, 0xfd,
0x30, 0x99, 0x85, 0xed, 0xc5, 0xde, 0xdb, 0x98
};

// g19 (counter = 0)
static readonly byte[] g19_x = {
0xb8, 0xad, 0x38, 0x6b, 0x54, 0xf9, 0x76, 0x6e,
0x5c, 0xb1, 0xa2, 0xf0, 0x50, 0xcb, 0xca, 0x2a,
0x22, 0x61, 0x9b, 0xa0, 0x08, 0xfd, 0xf9, 0x49,
0x6d, 0xf3, 0x8a, 0x6c, 0xea, 0x78, 0x4e, 0xb2
};
static readonly byte[] g19_y = {
0x5b, 0x33, 0x3a, 0x0c, 0xde, 0x9d, 0xdc, 0x8d,
0x65, 0x71, 0xb1, 0xca, 0xc4, 0x56, 0xa4, 0x71,
0x44, 0xc9, 0xc1, 0x6e, 0xce, 0x86, 0x6a, 0x53,
0x84, 0x94, 0xea, 0x0f, 0xea, 0xee, 0xf0, 0xac
};

// g20 (counter = 0)
static readonly byte[] g20_x = {
0x56, 0x62, 0x8c, 0x7d, 0x63, 0x66, 0xe1, 0xc4,
0xa9, 0x36, 0x1e, 0x5f, 0x7e, 0x49, 0x41, 0x5c,
0x80, 0xfd, 0xa1, 0x4c, 0x04, 0xf1, 0x06, 0xf0,
0x63, 0x8e, 0xc8, 0xcf, 0x59, 0xaa, 0x04, 0x85
};
static readonly byte[] g20_y = {
0x74, 0xfd, 0xc2, 0x60, 0x80, 0x2b, 0x6d, 0xf5,
0x5a, 0x64, 0x02, 0x33, 0x88, 0x95, 0x35, 0xcd,
0x04, 0xe0, 0xdf, 0x84, 0xb6, 0x6d, 0x9d, 0xa4,
0x64, 0x5d, 0xa3, 0x11, 0x93, 0x99, 0x50, 0x46
};

// g21 (counter = 1)
static readonly byte[] g21_x = {
0x8f, 0x1f, 0x5a, 0x0e, 0x34, 0x2e, 0x65, 0x57,
0xb9, 0x55, 0x35, 0x54, 0x38, 0x60, 0x8d, 0xb0,
0x9e, 0x4d, 0x23, 0x7e, 0xc7, 0x23, 0x0e, 0x2c,
0x83, 0x6b, 0xd5, 0xf3, 0xe9, 0x1c, 0x6c, 0x12
};
static readonly byte[] g21_y = {
0x2c, 0x1a, 0x21, 0x02, 0xa6, 0x9e, 0xf7, 0x4a,
0x00, 0x63, 0x53, 0xc2, 0xd2, 0xd1, 0xdd, 0x9d,
0xbd, 0xfa, 0xb0, 0x07, 0xfd, 0x08, 0xe7, 0xc8,
0x8e, 0xb8, 0x69, 0xa0, 0xa6, 0x69, 0xb1
};

// g22 (counter = 0)
static readonly byte[] g22_x = {
0xbe, 0xaf, 0x77, 0x57, 0xa3, 0xce, 0x43, 0xdc,
0x8d, 0x4a, 0x07, 0x32, 0xe1, 0xe3, 0x18, 0xf4,
0x97, 0x55, 0xe6, 0x1e, 0x5f, 0x57, 0xa8, 0x5b,
0xec, 0xcf, 0x21, 0xb7, 0xdc, 0xc8, 0x18, 0xe2
};
static readonly byte[] g22_y = {
0x40, 0xd2, 0x6c, 0x2a, 0xdc, 0x3f, 0x41, 0xd0,
0x91, 0x56, 0x02, 0x5a, 0x9d, 0xc3, 0x4f, 0xd3,
0xca, 0x6b, 0x96, 0x80, 0x9d, 0x3d, 0x7c, 0xf5,
0xf2, 0x8d, 0x00, 0xa1, 0xed, 0xbd, 0x69, 0x95
};

// g23 (counter = 2)
static readonly byte[] g23_x = {
0xe5, 0x13, 0xc3, 0xe5, 0x0e, 0xfa, 0x44, 0x36,
0x19, 0x9c, 0x5a, 0x51, 0xfd, 0x69, 0x1e, 0xa4,
0xdc, 0xab, 0xbc, 0x20, 0x2a, 0x80, 0x29, 0xba,
0x3d, 0xf0, 0x33, 0x6f, 0x12, 0xd8, 0x26, 0x63
};
static readonly byte[] g23_y = {
0x75, 0xf4, 0x2f, 0x58, 0x48, 0x0d, 0x2c, 0xad,
0x56, 0x9b, 0x0f, 0x13, 0xcb, 0xf3, 0x76, 0xc3,
0x91, 0x32, 0x71, 0xd9, 0xf7, 0x84, 0x42, 0x42,
0xb8, 0x70, 0x51, 0x9d, 0x2b, 0xe8, 0x39, 0x8e
};

// g24 (counter = 0)
static readonly byte[] g24_x = {
0xb4, 0x2b, 0x3b, 0x05, 0xbc, 0xaf, 0xbb, 0x72,
0x80, 0x0e, 0xe2, 0x42, 0xab, 0x4c, 0xb7, 0xab,
0xd7, 0x7f, 0x1f, 0xce, 0xac, 0x7c, 0xe1, 0xd3,
0x27, 0xee, 0xc2, 0x5b, 0x3d, 0xe6, 0xc4, 0x3d
};
static readonly byte[] g24_y = {
0x72, 0x5f, 0x5b, 0x3d, 0x0c, 0xdd, 0x1b, 0x86,
0xbd, 0x7a, 0x8b, 0xd6, 0x35, 0xc1, 0xac, 0xed,
0xba, 0xc9, 0x1d, 0x6c, 0x35, 0x16, 0x3e, 0xae,
0x66, 0x81, 0x07, 0x51, 0xf4, 0xd4, 0x62, 0x88
};

// g25 (counter = 0)
static readonly byte[] g25_x = {
0xc8, 0xa4, 0xa7, 0xdf, 0x6b, 0xef, 0x6c, 0x61,
0xef, 0x50, 0xbf, 0xfd, 0x9c, 0xfa, 0x7e, 0xfd,
0xe2, 0x25, 0x30, 0xf0, 0xb2, 0xd0, 0x37, 0x1e,
0x81, 0x9b, 0x80, 0xe8, 0x85, 0xd5, 0x92, 0xdd
};
static readonly byte[] g25_y = {
0x19, 0x6e, 0x7e, 0x0a, 0x81, 0xd0, 0x3b, 0x38,
0xa8, 0xf9, 0x91, 0x04, 0x81, 0x2f, 0x64, 0x78,
0x4b, 0x62, 0xd4, 0x19, 0x91, 0xf5, 0x66, 0xde,
0x27, 0x84, 0x7b, 0x6b, 0xb9, 0xba, 0xa2, 0x51
};

// g26 (counter = 4)
static readonly byte[] g26_x = {
0xa2, 0x2a, 0xf4, 0x5e, 0x5a, 0x7a, 0x9a, 0x9f,
0x94, 0x91, 0x0e, 0x8c, 0xdb, 0x5e, 0x64, 0x9e,
0x83, 0xc3, 0x8f, 0xc1, 0x36, 0x9f, 0x1c, 0xa9,
0xfa, 0x1d, 0x51, 0x88, 0x7c, 0x38, 0xdd, 0xf1
};
static readonly byte[] g26_y = {
0x75, 0x9b, 0xd3, 0x8c, 0x6e, 0x09, 0xfe, 0x2c,
0xd7, 0x5b, 0x4f, 0x35, 0x5f, 0x44, 0x20, 0xe2,
0xe7, 0xb2, 0xdf, 0xd9, 0xf7, 0x14, 0x7a, 0xa0,
0x3d, 0x53, 0x73, 0xb3, 0x61, 0x2b, 0x83, 0x89
};

// g27 (counter = 2)
static readonly byte[] g27_x = {
0x22, 0xf4, 0x7a, 0x6a, 0xae, 0xc1, 0x42, 0x35,
0x94, 0x81, 0xee, 0xa4, 0x90, 0x98, 0x88, 0x2b,
0x3e, 0xca, 0xc4, 0x62, 0x5b, 0x1d, 0x25, 0x62,
0xb0, 0x27, 0x18, 0x48, 0x76, 0x2c, 0x5d, 0xde
};
static readonly byte[] g27_y = {
0x3e, 0x0b, 0x7e, 0x0c, 0x51, 0xa0, 0x63, 0x30,
0x35, 0x80, 0xca, 0x25, 0xe3, 0x26, 0xae, 0x7e,
0x61, 0x08, 0x6e, 0xa6, 0xe4, 0xc4, 0x95, 0xd2,
0x51, 0x62, 0x86, 0x70, 0x39, 0xd9, 0xfe, 0x4c
};

// g28 (counter = 0)
static readonly byte[] g28_x = {
0xea, 0xe2, 0x4e, 0x9c, 0xbf, 0x4a, 0x8e, 0xb9,
0x2c, 0x1c, 0xc8, 0x0d, 0x75, 0xdc, 0xf4, 0x4c,
0x39, 0xdf, 0xe4, 0xed, 0xcf, 0x13, 0xc3, 0xe5,
0xe4, 0xb7, 0xba, 0x08, 0xc3, 0x29, 0x37, 0x8d
};
static readonly byte[] g28_y = {
0x2f, 0x7f, 0xff, 0xfa, 0x43, 0xa2, 0xd0, 0x26,
0x8c, 0x25, 0xe4, 0xf0, 0x86, 0x63, 0xfe, 0xf2,
0x6c, 0x57, 0x96, 0x2f, 0xd5, 0xf6, 0x23, 0x29,
0x2f, 0x06, 0x1e, 0xa1, 0x9c, 0x57, 0x10, 0xa1
};

// g29 (counter = 1)
static readonly byte[] g29_x = {
0xad, 0x92, 0xb0, 0x98, 0x52, 0x8a, 0xe2, 0x08,
0x57, 0x24, 0x74, 0xe3, 0xca, 0x2b, 0x1f, 0x6f,
0xbe, 0x13, 0x3c, 0xb4, 0xfa, 0xb5, 0xee, 0xba,
0x0e, 0x46, 0x10, 0x0c, 0x68, 0x4d, 0x5b, 0xbc
};
static readonly byte[] g29_y = {
0x47, 0x97, 0x86, 0x85, 0xfa, 0x8f, 0x41, 0xca,
0x52, 0x46, 0xbd, 0x63, 0x47, 0xba, 0x65, 0xf6,
0x70, 0xec, 0x65, 0xa1, 0x36, 0x16, 0x6c, 0x75,
0xe7, 0x93, 0x63, 0x46, 0xe1, 0x6a, 0xd7, 0x90
};

// g30 (counter = 0)
static readonly byte[] g30_x = {
0xdc, 0x5a, 0xbc, 0x9d, 0x9e, 0x2a, 0x04, 0xa7,
0xba, 0x38, 0x34, 0x6e, 0x82, 0x71, 0x19, 0xf5,
0x0f, 0xa3, 0x11, 0xb8, 0xcb, 0x4b, 0x12, 0xcf,
0x53, 0x60, 0x2f, 0x34, 0x82, 0xa6, 0x09, 0xc0
};
static readonly byte[] g30_y = {
0xe9, 0x4f, 0x73, 0xd5, 0xd9, 0x64, 0x19, 0x42,
0x18, 0x8f, 0xd0, 0xff, 0x64, 0xa7, 0x75, 0x10,
0x21, 0xfa, 0xf6, 0xcc, 0x9c, 0x4d, 0x2a, 0xa0,
0x31, 0x8e, 0x94, 0xf0, 0x59, 0x78, 0xbe
};

// g31 (counter = 0)
static readonly byte[] g31_x = {
0x5d, 0x00, 0x8b, 0x9b, 0xde, 0xbb, 0x38, 0x24,
0x93, 0x5b, 0xdc, 0x68, 0xa7, 0xac, 0x42, 0x6c,
0x55, 0x40, 0x58, 0xa9, 0xdc, 0x4e, 0xd8, 0xbe,
0xa2, 0xea, 0x74, 0xa9, 0x2d, 0xf4, 0x7f, 0xc3
};
static readonly byte[] g31_y = {
0x18, 0x05, 0xd5, 0xf8, 0xf0, 0x97, 0xea, 0x8b,
0x3b, 0x86, 0x08, 0xdc, 0x5f, 0x01, 0x6f, 0xd9,
0x09, 0x78, 0x1b, 0x75, 0x90, 0x0d, 0x53, 0xce,
0x8b, 0x65, 0x84, 0x65, 0x18, 0xca, 0x0b, 0xda
};

// g32 (counter = 0)
static readonly byte[] g32_x = {
0x4b, 0xff, 0x16, 0x06, 0x7e, 0x37, 0x79, 0x8f,
0xf3, 0xe3, 0x24, 0x2b, 0x11, 0xbe, 0x39, 0xf8,
0x3d, 0xd7, 0x45, 0x1e, 0xbe, 0x11, 0x01, 0xea,
0xc4, 0x88, 0x7a, 0x6f, 0x93, 0xd5, 0x02, 0x06
};
static readonly byte[] g32_y = {
0x06, 0x5e, 0x5e, 0x31, 0xe1, 0x50, 0x13, 0x60,
0x36, 0xe1, 0x92, 0x25, 0x49, 0xb9, 0xfd, 0x9a,
0x85, 0x59, 0x97, 0x12, 0x9f, 0x45, 0x66, 0xd3,
0xf5, 0xac, 0xf8, 0xa1, 0xe4, 0xd0, 0xac, 0x83
};

// g33 (counter = 0)
static readonly byte[] g33_x = {
0xae, 0xcb, 0xa7, 0xf0, 0x74, 0x51, 0x23, 0xd9,
0xc6, 0xa6, 0x0e, 0x9b, 0xd4, 0x61, 0xa8, 0x63,
0x61, 0x31, 0xb0, 0x95, 0xf5, 0x96, 0x17, 0x84,
0x9d, 0x33, 0x5d, 0x2a, 0x7d, 0x8b, 0x18, 0x7b
};
static readonly byte[] g33_y = {
0x5f, 0x62, 0xd5, 0xea, 0xf4, 0xa9, 0xa8, 0x92,
0x48, 0x8c, 0x0d, 0xe9, 0x5d, 0x8d, 0x85, 0xed,
0xa9, 0x03, 0x5b, 0x65, 0x97, 0xea, 0x26, 0x74,
0xd7, 0xa7, 0xee, 0x7d, 0x4a, 0x53, 0x5e, 0xbd
};

// g34 (counter = 2)
static readonly byte[] g34_x = {
0xa7, 0x4e, 0xcb, 0x80, 0x73, 0x24, 0x96, 0xe8,
0xf6, 0xce, 0x72, 0xf4, 0x55, 0x69, 0x37, 0xc2,
0x37, 0xe1, 0x9e, 0xfa, 0xc7, 0x56, 0x7c, 0x15,
0x1f, 0x38, 0x6b, 0x65, 0x06, 0x56, 0xa2, 0x26
};
static readonly byte[] g34_y = {
0x04, 0xf6, 0x61, 0x41, 0x53, 0x13, 0x28, 0x4d,
0x90, 0x44, 0x85, 0xe6, 0xf6, 0xdb, 0x8f, 0xe9,
0x47, 0x82, 0xb2, 0xba, 0x24, 0xc0, 0xcb, 0xa6,
0xca, 0x77, 0x55, 0x7e, 0xfc, 0xd8, 0xf0, 0x5e
};

// g35 (counter = 2)
static readonly byte[] g35_x = {
0xed, 0x0e, 0x96, 0x56, 0x69, 0x01, 0x7a, 0xa7,
0x1f, 0x34, 0x2e, 0xc8, 0xa0, 0x99, 0xbb, 0xf0,
0x1a, 0x0b, 0x9e, 0xab, 0x94, 0xf6, 0x26, 0x23,
0xec, 0xf9, 0x6b, 0xcc, 0x0e, 0x14, 0xe4, 0xab
};
static readonly byte[] g35_y = {
0x24, 0x4b, 0xf1, 0x25, 0x52, 0x3e, 0xf2, 0x97,
0x8d, 0xb0, 0x60, 0x06, 0xcd, 0xa7, 0xcf, 0x3e,
0x4d, 0x58, 0x39, 0x77, 0x11, 0xd9, 0x28, 0x97,
0x60, 0x3d, 0xba, 0xe2, 0x9b, 0x82, 0x86, 0x4b
};

// g36 (counter = 0)
static readonly byte[] g36_x = {
0x06, 0x9b, 0x84, 0x3b, 0xdb, 0xf0, 0x17, 0xd4,
0x16, 0xa7, 0x67, 0xd1, 0x34, 0xe1, 0xc2, 0xd4,
0x97, 0xfa, 0xd2, 0xcd, 0xaa, 0xe3, 0x6b, 0x27,
0x53, 0x70, 0xff, 0x51, 0x2a, 0x34, 0xbf, 0xa7
};
static readonly byte[] g36_y = {
0x3d, 0x3b, 0xe3, 0xd2, 0xe8, 0x6e, 0xb0, 0x7a,
0x87, 0x84, 0x9b, 0x2e, 0xf1, 0x6e, 0xe3, 0x03,
0x10, 0xb8, 0x6e, 0x63, 0xb3, 0x47, 0x81, 0x63,
0xfd, 0x06, 0xb6, 0x59, 0x2b, 0xbd, 0xe5, 0x45
};

// g37 (counter = 0)
static readonly byte[] g37_x = {
0x59, 0x2d, 0x48, 0x15, 0x8a, 0x63, 0x58, 0xa2,
0x90, 0x0d, 0x45, 0x3d, 0x79, 0xe8, 0x8d, 0x6b,
0xc2, 0x0b, 0x7f, 0xa8, 0xcb, 0x2b, 0xfc, 0xfc,
0xdf, 0xd0, 0x82, 0x96, 0x05, 0x25, 0xad, 0x83
};
static readonly byte[] g37_y = {
0x72, 0x31, 0xc3, 0xd1, 0xf8, 0x6f, 0xcc, 0x1b,
0x6c, 0x9e, 0x8c, 0x16, 0xae, 0x45, 0xa9, 0x35,
0x08, 0xc9, 0xc4, 0x9e, 0x8a, 0x74, 0x5e, 0x64,
0xb0, 0x76, 0x36, 0xfc, 0x6b, 0x03, 0x10, 0x3f
};

// g38 (counter = 0)
static readonly byte[] g38_x = {
0x18, 0xff, 0xac, 0x75, 0x07, 0xb8, 0xf0, 0x22,
0xeb, 0xa9, 0x72, 0x2a, 0xea, 0x93, 0xc6, 0xca,
0x74, 0x70, 0x82, 0x5a, 0x78, 0x7c, 0x1f, 0x98,
0x2b, 0x08, 0x3d, 0xda, 0x04, 0x90, 0xed, 0x32
};
static readonly byte[] g38_y = {
0x30, 0x4b, 0x83, 0x60, 0x4a, 0x94, 0xff, 0x8a,
0x27, 0x87, 0xb0, 0x47, 0xe8, 0x23, 0xe5, 0x0a,
0x64, 0xed, 0xca, 0x0b, 0x1d, 0xcc, 0xb9, 0x38,
0x11, 0x96, 0x59, 0x7a, 0x1c, 0x63, 0xb3, 0x62
};

// g39 (counter = 0)
static readonly byte[] g39_x = {
0xdd, 0xe5, 0xdf, 0xc2, 0x86, 0x7a, 0x61, 0xba,
0x2e, 0x04, 0x6d, 0xd5, 0x25, 0x76, 0xd3, 0xd3,
0x3a, 0x24, 0x17, 0x3e, 0x32, 0xd7, 0x16, 0xca,
0xf0, 0xd6, 0xbc, 0x4b, 0xd1, 0x19, 0x43, 0x74
};
static readonly byte[] g39_y = {
0x79, 0xb6, 0xe3, 0x0b, 0x18, 0x22, 0xd6, 0x1e,
0xad, 0xe5, 0x9b, 0x0a, 0xb3, 0xed, 0xbe, 0x8f,
0x42, 0x91, 0xc8, 0xe0, 0x81, 0xdd, 0xce, 0xde,
0xff, 0x00, 0xbc, 0x32, 0xeb, 0xfc, 0x1a, 0x93
};

// g40 (counter = 1)
static readonly byte[] g40_x = {
0xe0, 0xf7, 0x2a, 0x8c, 0x71, 0x39, 0x5e, 0x19,
0x06, 0x3b, 0x0e, 0x09, 0xf9, 0x47, 0xf8, 0x6c,
0x06, 0xf4, 0xb3, 0x00, 0xc8, 0x1d, 0x3b, 0xbb,
0xc4, 0x8d, 0xcb, 0x21, 0x9a, 0xb9, 0x60, 0xaa
};
static readonly byte[] g40_y = {
0x6f, 0x23, 0x1e, 0x0a, 0x53, 0x8c, 0x8f, 0x54,
0xc0, 0x66, 0xc9, 0x3e, 0x1a, 0xf8, 0x57, 0xbc,
0x3b, 0x1c, 0x41, 0x88, 0x02, 0x27, 0x4c, 0xbd,
0xf5, 0xe3, 0x87, 0xd8, 0x87, 0x36, 0xf5, 0x76
};

// g41 (counter = 0)
static readonly byte[] g41_x = {
0x38, 0x53, 0x88, 0x07, 0x8e, 0xa2, 0xb4, 0x79,
0x2d, 0xac, 0x8f, 0xbe, 0x0b, 0x47, 0x48, 0xb9,
0x98, 0x00, 0xca, 0x08, 0x66, 0x62, 0xfa, 0x8e,
0xab, 0xd6, 0x25, 0x96, 0xdd, 0x7e, 0x5c, 0x53
};
static readonly byte[] g41_y = {
0x4d, 0x21, 0x12, 0x11, 0x1d, 0x5b, 0xf4, 0x7b,
0xae, 0xd1, 0xc4, 0xa2, 0x68, 0x8c, 0xfa, 0x61,
0x6e, 0x7b, 0xbb, 0x64, 0xd4, 0x12, 0xf1, 0x6b,
0x37, 0x12, 0x88, 0xbf, 0xe9, 0x57, 0xea, 0x61
};

// g42 (counter = 0)
static readonly byte[] g42_x = {
0xb1, 0x08, 0xaa, 0x3e, 0x8b, 0xf1, 0xf7, 0x07,
0xf6, 0xba, 0x95, 0x56, 0xaa, 0x0f, 0x18, 0x71,
0x51, 0x97, 0x34, 0xa6, 0x98, 0x20, 0x3f, 0x75,
0x32, 0x92, 0x54, 0x43, 0xb2, 0x02, 0x0c, 0xbd
};
static readonly byte[] g42_y = {
0x5a, 0x75, 0xfa, 0xe7, 0xad, 0x0b, 0xe2, 0x35,
0x20, 0x73, 0x47, 0x79, 0xef, 0x11, 0xf3, 0x25,
0xdd, 0xe7, 0xa6, 0xed, 0xc6, 0x33, 0x36, 0xef,
0x9f, 0xb5, 0x86, 0x61, 0xfc, 0xcc, 0x46, 0xa5
};

// g43 (counter = 0)
static readonly byte[] g43_x = {
0x06, 0x05, 0xb3, 0x50, 0x5f, 0x77, 0xe7, 0x4b,
0x22, 0xea, 0x7e, 0x67, 0xc3, 0x33, 0x3f, 0xf3,
0xb7, 0xb7, 0x71, 0x73, 0x83, 0x89, 0xd3, 0x05,
0xaa, 0x59, 0x4d, 0x8f, 0x55, 0x02, 0x37, 0xdb
};
static readonly byte[] g43_y = {
0x74, 0x87, 0xad, 0xb2, 0xe0, 0x7c, 0x3a, 0xb9,
0x2e, 0x13, 0x86, 0x54, 0x67, 0x90, 0xa0, 0x11,
0x49, 0x7e, 0xb9, 0xfb, 0x98, 0x46, 0x71, 0x6b,
0x04, 0x79, 0x3d, 0xce, 0xa4, 0x30, 0xc7, 0xab
};

// g44 (counter = 0)
static readonly byte[] g44_x = {
0xd8, 0x18, 0x83, 0xa9, 0xcf, 0x1d, 0xc3, 0x04,
0x3c, 0x44, 0xf9, 0xf0, 0xf9, 0xff, 0x50, 0x2c,
0xd0, 0x45, 0xe4, 0x29, 0x4c, 0x37, 0x5a, 0x30,
0xa8, 0xa6, 0x5a, 0xbc, 0x0d, 0xd2, 0x82, 0x64
};
static readonly byte[] g44_y = {
0x1d, 0x75, 0xc9, 0x9e, 0xb4, 0x4e, 0x2d, 0x8b,
0x43, 0xa5, 0x3f, 0x69, 0xb6, 0x88, 0x1f, 0x96,
0x92, 0x94, 0x35, 0xe2, 0xb3, 0x85, 0x0a, 0x37,
0x01, 0xae, 0xd0, 0x26, 0xe8, 0x0a, 0x32, 0x91
};

// g45 (counter = 1)
static readonly byte[] g45_x = {
0x93, 0xec, 0x90, 0x87, 0x9c, 0xd2, 0xd8, 0x6a,
0x22, 0x76, 0xf4, 0x4b, 0x42, 0xdf, 0x73, 0x62,
0x83, 0xd2, 0x97, 0x47, 0x07, 0x59, 0xde, 0x0a,
0xf2, 0xc6, 0xc9, 0x2f, 0x16, 0x84, 0x82, 0xaf
};
static readonly byte[] g45_y = {
0x1f, 0x45, 0xf4, 0x80, 0xa0, 0xec, 0x76, 0x07,
0x51, 0x66, 0x79, 0xc2, 0xbb, 0x9f, 0x67, 0x7a,
0x89, 0xd4, 0x50, 0xec, 0x46, 0x9a, 0xc9, 0x30,
0xa1, 0x0d, 0x21, 0x3c, 0x1e, 0xb2, 0xa9, 0xcf
};

// g46 (counter = 0)
static readonly byte[] g46_x = {
0x4e, 0x9e, 0x9e, 0xb8, 0xe2, 0x67, 0xc0, 0xd6,
0x17, 0x60, 0xec, 0xab, 0xc9, 0xac, 0x19, 0xdd,
0xac, 0x5d, 0xb9, 0x5c, 0x28, 0x33, 0x4e, 0xc9,
0x9d, 0x49, 0xd7, 0x4d, 0x40, 0xb6, 0x6d, 0xaf
};
static readonly byte[] g46_y = {
0x5d, 0xd7, 0x1c, 0x92, 0xd3, 0x11, 0xec, 0x15,
0xd5, 0xe2, 0xe6, 0xd3, 0xb8, 0xd5, 0x13, 0x36,
0x41, 0x5a, 0x60, 0x8e, 0x14, 0x04, 0x8c, 0x86,
0xce, 0xec, 0x76, 0x4e, 0x6d, 0xe6, 0xdf, 0x49
};

// g47 (counter = 2)
static readonly byte[] g47_x = {
0xce, 0xb4, 0xca, 0x98, 0xf6, 0x20, 0x19, 0x59,
0x6b, 0x9b, 0xc6, 0x23, 0x4e, 0xa5, 0xc2, 0x02,
0x99, 0x90, 0xf0, 0x8d, 0x06, 0x8f, 0x27, 0xee,
0xf4, 0xfa, 0x7d, 0x98, 0x97, 0xbf, 0xaf, 0x62
};
static readonly byte[] g47_y = {
0x41, 0x60, 0xfb, 0xdd, 0xaf, 0x29, 0x86, 0xf3,
0xa1, 0x1e, 0x29, 0xb5, 0x89, 0xb9, 0xd9, 0x1d,
0x8b, 0x15, 0xc5, 0xf8, 0xbb, 0xf0, 0x2f, 0x7f,
0x17, 0x5f, 0x6e, 0xf8, 0xe7, 0xc2, 0xb1, 0xa4
};

// g48 (counter = 2)
static readonly byte[] g48_x = {
0x80, 0xe8, 0x70, 0x67, 0x09, 0xbd, 0x25, 0xa8,
0x49, 0x37, 0x41, 0x7e, 0x2d, 0x6a, 0x6d, 0xaf,
0xa8, 0x3d, 0x37, 0x38, 0xdf, 0xb4, 0x2f, 0x8e,
0xef, 0xa0, 0xfb, 0x52, 0x47, 0xd6, 0x99, 0x85
};
static readonly byte[] g48_y = {
0x6a, 0x8f, 0x2e, 0xa6, 0xb2, 0x30, 0x1e, 0x3a,
0xef, 0xbd, 0x82, 0x46, 0xf6, 0xeb, 0x97, 0xea,
0x0c, 0xe1, 0x15, 0x5c, 0xe0, 0xb7, 0x2c, 0x47,
0x1d, 0x01, 0xb0, 0xd0, 0xb8, 0x8d, 0xa2, 0xca
};

// g49 (counter = 0)
static readonly byte[] g49_x = {
0x13, 0xbd, 0x26, 0x06, 0x06, 0x67, 0xf8, 0xeb,
0x7e, 0x56, 0xe7, 0x82, 0x85, 0x4a, 0xf3, 0xb3,
0xe0, 0x10, 0xcf, 0x18, 0x25, 0xa6, 0x84, 0xbc,
0x72, 0xb2, 0x87, 0xea, 0x7b, 0x2c, 0x23, 0x4c
};
static readonly byte[] g49_y = {
0x18, 0x71, 0xc1, 0x5a, 0xa6, 0xf8, 0xcc, 0x3a,
0xda, 0x2d, 0x4b, 0xf6, 0xbb, 0x2b, 0xc6, 0x29,
0x6c, 0xa6, 0x58, 0x7c, 0x12, 0x2d, 0xf3, 0xb4,
0x7a, 0x9f, 0xaa, 0x30, 0x25, 0x86, 0x3a, 0x8c
};

// g50 (counter = 0)
static readonly byte[] g50_x = {
0x7d, 0x5e, 0x69, 0xba, 0xce, 0x92, 0x0e, 0x8e,
0xd2, 0xd0, 0xb4, 0x3a, 0xd1, 0x48, 0x49, 0xd7,
0x1e, 0x26, 0x72, 0x9c, 0xb3, 0x7f, 0x00, 0x9a,
0xe1, 0x4e, 0x6d, 0x8a, 0x06, 0x5e, 0x90, 0x79
};
static readonly byte[] g50_y = {
0x13, 0xd6, 0xc8, 0xd6, 0xae, 0x02, 0x73, 0xa1,
0x89, 0x01, 0x29, 0x77, 0x9f, 0xce, 0x34, 0xf0,
0xca, 0xf6, 0xf3, 0x53, 0xbf, 0xde, 0x9e, 0xe3,
0x37, 0x27, 0x86, 0x78, 0xc9, 0xb6, 0xe7, 0x58
};

// gt (counter = 4)
static readonly byte[] gt_x = {
0xe2, 0xab, 0x81, 0xde, 0xf5, 0x93, 0xe9, 0x99,
0xc9, 0x75, 0xa8, 0xa4, 0x86, 0x68, 0xb9, 0xa0,
0x7e, 0x55, 0x94, 0xcf, 0xd6, 0x8f, 0xac, 0x29,
0xf1, 0x7a, 0x81, 0x1c, 0xb2, 0x6b, 0x3e, 0x10
};
static readonly byte[] gt_y = {
0x75, 0x63, 0x11, 0xf8, 0x96, 0xc5, 0x03, 0xec,
0xdb, 0x2f, 0x60, 0x8a, 0x1c, 0xcb, 0xfa, 0x37,
0x8a, 0x95, 0xeb, 0x45, 0x78, 0xe6, 0x5f, 0x19,
0x0f, 0x1a, 0x8b, 0x54, 0x4d, 0x20, 0xb0, 0x82
};

// gd (counter = 0)
static readonly byte[] gd_x = {
0x4c, 0xa6, 0x25, 0x11, 0x8d, 0x0a, 0x05, 0xd0,
0x4d, 0x27, 0x5d, 0xae, 0x1f, 0xf0, 0x96, 0x36,
0x1e, 0xbe, 0xba, 0x34, 0x5c, 0x31, 0x27, 0x09,
0x82, 0xf7, 0x96, 0x63, 0x9b, 0x1c, 0xa5, 0x74
};
static readonly byte[] gd_y = {
0x14, 0x2d, 0x15, 0x0c, 0x85, 0x5b, 0xa9, 0xaa,
0x7d, 0xcc, 0x71, 0x82, 0x1a, 0x53, 0x8e, 0xdb,
0x54, 0x48, 0x36, 0xdf, 0x80, 0x50, 0x91, 0x26,
0x79, 0xcc, 0xd7, 0x23, 0x3f, 0xbb, 0xa6, 0x36
};

            private static readonly ECGroup ecGroup = ECGroup.CreateECGroup(p, a, b, g_x, g_y, n, ParamSet_EC_P256_V1Name, "P-256");

            internal static readonly Group Gq = ecGroup;
                
            internal static readonly GroupElement Gd = ecGroup.CreateGroupElement(gd_x, gd_y);

            internal static readonly GroupElement[] IssuerGenerators = {
                ecGroup.CreateGroupElement(g1_x, g1_y),
                ecGroup.CreateGroupElement(g2_x, g2_y),
                ecGroup.CreateGroupElement(g3_x, g3_y),
                ecGroup.CreateGroupElement(g4_x, g4_y),
                ecGroup.CreateGroupElement(g5_x, g5_y),
                ecGroup.CreateGroupElement(g6_x, g6_y),
                ecGroup.CreateGroupElement(g7_x, g7_y),
                ecGroup.CreateGroupElement(g8_x, g8_y),
                ecGroup.CreateGroupElement(g9_x, g9_y),
                ecGroup.CreateGroupElement(g10_x, g10_y),
                ecGroup.CreateGroupElement(g11_x, g11_y),
                ecGroup.CreateGroupElement(g12_x, g12_y),
                ecGroup.CreateGroupElement(g13_x, g13_y),
                ecGroup.CreateGroupElement(g14_x, g14_y),
                ecGroup.CreateGroupElement(g15_x, g15_y),
                ecGroup.CreateGroupElement(g16_x, g16_y),
                ecGroup.CreateGroupElement(g17_x, g17_y),
                ecGroup.CreateGroupElement(g18_x, g18_y),
                ecGroup.CreateGroupElement(g19_x, g19_y),
                ecGroup.CreateGroupElement(g20_x, g20_y),
                ecGroup.CreateGroupElement(g21_x, g21_y),
                ecGroup.CreateGroupElement(g22_x, g22_y),
                ecGroup.CreateGroupElement(g23_x, g23_y),
                ecGroup.CreateGroupElement(g24_x, g24_y),
                ecGroup.CreateGroupElement(g25_x, g25_y),
                ecGroup.CreateGroupElement(g26_x, g26_y),
                ecGroup.CreateGroupElement(g27_x, g27_y),
                ecGroup.CreateGroupElement(g28_x, g28_y),
                ecGroup.CreateGroupElement(g29_x, g29_y),
                ecGroup.CreateGroupElement(g30_x, g30_y),
                ecGroup.CreateGroupElement(g31_x, g31_y),
                ecGroup.CreateGroupElement(g32_x, g32_y),
                ecGroup.CreateGroupElement(g33_x, g33_y),
                ecGroup.CreateGroupElement(g34_x, g34_y),
                ecGroup.CreateGroupElement(g35_x, g35_y),
                ecGroup.CreateGroupElement(g36_x, g36_y),
                ecGroup.CreateGroupElement(g37_x, g37_y),
                ecGroup.CreateGroupElement(g38_x, g38_y),
                ecGroup.CreateGroupElement(g39_x, g39_y),
                ecGroup.CreateGroupElement(g40_x, g40_y),
                ecGroup.CreateGroupElement(g41_x, g41_y),
                ecGroup.CreateGroupElement(g42_x, g42_y),
                ecGroup.CreateGroupElement(g43_x, g43_y),
                ecGroup.CreateGroupElement(g44_x, g44_y),
                ecGroup.CreateGroupElement(g45_x, g45_y),
                ecGroup.CreateGroupElement(g46_x, g46_y),
                ecGroup.CreateGroupElement(g47_x, g47_y),
                ecGroup.CreateGroupElement(g48_x, g48_y),
                ecGroup.CreateGroupElement(g49_x, g49_y),
                ecGroup.CreateGroupElement(g50_x, g50_y),
                ecGroup.CreateGroupElement(gt_x, gt_y)
                };

        }

        //
        // P-384
        //
        static class P384
        {

static readonly byte[] p = {
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe,
0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff
};

static readonly byte[] a = {
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe,
0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xfc
};

static readonly byte[] b = {
0xb3, 0x31, 0x2f, 0xa7, 0xe2, 0x3e, 0xe7, 0xe4,
0x98, 0x8e, 0x05, 0x6b, 0xe3, 0xf8, 0x2d, 0x19,
0x18, 0x1d, 0x9c, 0x6e, 0xfe, 0x81, 0x41, 0x12,
0x03, 0x14, 0x08, 0x8f, 0x50, 0x13, 0x87, 0x5a,
0xc6, 0x56, 0x39, 0x8d, 0x8a, 0x2e, 0xd1, 0x9d,
0x2a, 0x85, 0xc8, 0xed, 0xd3, 0xec, 0x2a, 0xef
};

static readonly byte[] n = {
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xc7, 0x63, 0x4d, 0x81, 0xf4, 0x37, 0x2d, 0xdf,
0x58, 0x1a, 0x0d, 0xb2, 0x48, 0xb0, 0xa7, 0x7a,
0xec, 0xec, 0x19, 0x6a, 0xcc, 0xc5, 0x29, 0x73
};

static readonly byte[] h = {
0x01
};

// g
static readonly byte[] g_x = {
0xaa, 0x87, 0xca, 0x22, 0xbe, 0x8b, 0x05, 0x37,
0x8e, 0xb1, 0xc7, 0x1e, 0xf3, 0x20, 0xad, 0x74,
0x6e, 0x1d, 0x3b, 0x62, 0x8b, 0xa7, 0x9b, 0x98,
0x59, 0xf7, 0x41, 0xe0, 0x82, 0x54, 0x2a, 0x38,
0x55, 0x02, 0xf2, 0x5d, 0xbf, 0x55, 0x29, 0x6c,
0x3a, 0x54, 0x5e, 0x38, 0x72, 0x76, 0x0a, 0xb7
};
static readonly byte[] g_y = {
0x36, 0x17, 0xde, 0x4a, 0x96, 0x26, 0x2c, 0x6f,
0x5d, 0x9e, 0x98, 0xbf, 0x92, 0x92, 0xdc, 0x29,
0xf8, 0xf4, 0x1d, 0xbd, 0x28, 0x9a, 0x14, 0x7c,
0xe9, 0xda, 0x31, 0x13, 0xb5, 0xf0, 0xb8, 0xc0,
0x0a, 0x60, 0xb1, 0xce, 0x1d, 0x7e, 0x81, 0x9d,
0x7a, 0x43, 0x1d, 0x7c, 0x90, 0xea, 0x0e, 0x5f
};

// g1 (counter = 2)
static readonly byte[] g1_x = {
0x4a, 0xae, 0x57, 0x9d, 0xd5, 0x6d, 0x78, 0x09,
0x0b, 0x99, 0x21, 0xf3, 0x1b, 0xf7, 0x29, 0xf0,
0x74, 0x12, 0x1a, 0x3a, 0xdf, 0xfa, 0x2d, 0x31,
0xd0, 0x12, 0x15, 0xbe, 0xee, 0x1d, 0xc4, 0xdf,
0x9d, 0xf4, 0x63, 0xfd, 0x5e, 0x2b, 0x8f, 0x6c,
0x6b, 0x0a, 0x42, 0x16, 0x25, 0x8a, 0xc8, 0x44
};
static readonly byte[] g1_y = {
0x3c, 0x3b, 0x8a, 0x23, 0xc5, 0xd6, 0x6a, 0xa2,
0xf0, 0x96, 0x45, 0x21, 0x19, 0x0a, 0x92, 0x81,
0x45, 0x1e, 0x9a, 0xe3, 0xac, 0xe4, 0xb7, 0x37,
0x6e, 0x02, 0xd7, 0xb3, 0x94, 0x9e, 0x22, 0x74,
0xe8, 0x44, 0x8c, 0xad, 0xef, 0x7e, 0x51, 0x99,
0x17, 0x20, 0xb4, 0x9a, 0x45, 0xb0, 0x58, 0x05
};

// g2 (counter = 0)
static readonly byte[] g2_x = {
0x03, 0x2f, 0x08, 0x6e, 0xac, 0x7b, 0xeb, 0x55,
0xa0, 0xc9, 0x5e, 0x5a, 0xd9, 0x96, 0xe3, 0x9d,
0xde, 0x74, 0xb3, 0xb6, 0x6d, 0xfb, 0xc2, 0xb8,
0xf1, 0x25, 0x29, 0xd9, 0xdf, 0x4a, 0x2c, 0xb8,
0x4a, 0xa5, 0x4c, 0xe6, 0x9a, 0xb1, 0xfb, 0x22,
0xcd, 0x7f, 0xd7, 0x99, 0x67, 0xd2, 0x61, 0xc2
};
static readonly byte[] g2_y = {
0x16, 0xbb, 0xf0, 0x78, 0x65, 0x4e, 0x39, 0x16,
0x80, 0xbd, 0xb5, 0x74, 0x95, 0x01, 0x8c, 0xc8,
0xfe, 0x05, 0x13, 0x0a, 0xbf, 0xda, 0xa8, 0x4a,
0xb4, 0xaf, 0x90, 0xd0, 0xd2, 0xd6, 0xc0, 0x1f,
0xfd, 0xa8, 0xbc, 0x96, 0xcf, 0xcb, 0x00, 0x16,
0xf3, 0xdb, 0x13, 0xd8, 0x0a, 0xe8, 0xa2, 0xd0
};

// g3 (counter = 2)
static readonly byte[] g3_x = {
0xd6, 0x58, 0x3a, 0xfe, 0x48, 0x31, 0x1b, 0xec,
0x5c, 0x90, 0x16, 0x68, 0x25, 0x31, 0xc9, 0x35,
0xcf, 0x3f, 0xa9, 0x8e, 0x33, 0xd5, 0x03, 0x35,
0x48, 0x20, 0xc9, 0x9f, 0xb2, 0xe9, 0x02, 0xea,
0xcd, 0xb4, 0x19, 0x44, 0x12, 0x03, 0xe2, 0x87,
0xb0, 0xd3, 0x3a, 0xdb, 0xbe, 0x91, 0x2e, 0x33
};
static readonly byte[] g3_y = {
0x7f, 0x9e, 0x35, 0xc0, 0xc0, 0xda, 0xf2, 0xe6,
0x8e, 0xd1, 0x35, 0xeb, 0xc0, 0x79, 0x71, 0xd5,
0x8e, 0x0e, 0x6a, 0xc8, 0xda, 0x69, 0xf5, 0x4f,
0x0c, 0x09, 0x3e, 0x24, 0xdd, 0x3f, 0x62, 0x75,
0x1e, 0x81, 0x6b, 0x6f, 0x6e, 0x1f, 0xcb, 0x66,
0x22, 0x6c, 0x4f, 0x0b, 0x35, 0xf9, 0xac, 0xdc
};

// g4 (counter = 2)
static readonly byte[] g4_x = {
0x3e, 0xf8, 0x51, 0xb6, 0xe0, 0xa8, 0x4e, 0x24,
0xfc, 0x99, 0x9b, 0x05, 0x3c, 0xf6, 0xac, 0xf3,
0x2a, 0xdc, 0x94, 0x17, 0x84, 0xae, 0xf0, 0xde,
0x14, 0x82, 0x3e, 0xf7, 0xab, 0xbe, 0x7e, 0x7e,
0x49, 0xe7, 0xd8, 0x0c, 0xd3, 0x52, 0x95, 0xea,
0xdf, 0x33, 0x22, 0x65, 0xd4, 0xda, 0x16, 0x5f
};
static readonly byte[] g4_y = {
0x18, 0x28, 0x99, 0xd3, 0x46, 0x52, 0x62, 0x6c,
0x74, 0x74, 0x98, 0x93, 0x66, 0xb1, 0xf1, 0x7a,
0xd4, 0xd9, 0xc8, 0x12, 0x86, 0x3a, 0x67, 0xbc,
0xdf, 0x7d, 0x85, 0xc0, 0xed, 0x59, 0xcb, 0x73,
0x20, 0x66, 0x3e, 0xbb, 0x8f, 0xff, 0x3d, 0x5f,
0x56, 0x32, 0xe0, 0xa7, 0x5e, 0x00, 0x99, 0x66
};

// g5 (counter = 0)
static readonly byte[] g5_x = {
0x0b, 0x83, 0x44, 0x03, 0x1a, 0x95, 0x8a, 0x37,
0x4d, 0xe8, 0xee, 0x07, 0x11, 0xc1, 0x55, 0x54,
0x40, 0x4b, 0xfe, 0xdc, 0xa8, 0x0e, 0x48, 0x89,
0x92, 0x2b, 0xdb, 0x0c, 0xeb, 0xad, 0x3a, 0x30,
0xc9, 0x22, 0xf5, 0x49, 0x2d, 0x2c, 0xe4, 0x88,
0x4b, 0xa1, 0xc7, 0x7b, 0x57, 0x2f, 0xa0, 0xbd
};
static readonly byte[] g5_y = {
0x0d, 0x7d, 0x70, 0x7c, 0xec, 0x2e, 0xcb, 0xce,
0xd7, 0x6f, 0x32, 0xd4, 0x3c, 0xe7, 0xb5, 0x5c,
0xbd, 0x53, 0x27, 0x3f, 0x55, 0x60, 0x5b, 0x7b,
0x9b, 0xac, 0x8b, 0x3f, 0x52, 0x1b, 0xb1, 0x53,
0x96, 0x86, 0x96, 0x7b, 0xe1, 0x8b, 0x5a, 0xa4,
0x1a, 0x71, 0x65, 0xf7, 0x26, 0xab, 0x5d, 0xbb
};

// g6 (counter = 0)
static readonly byte[] g6_x = {
0xd8, 0x65, 0x19, 0x82, 0xf1, 0xcf, 0xab, 0x26,
0x72, 0x70, 0x21, 0x9a, 0xee, 0x25, 0x07, 0x36,
0xd5, 0x35, 0xc2, 0x89, 0xa3, 0x8c, 0x88, 0x5d,
0xf2, 0x8e, 0xeb, 0xf6, 0x0f, 0x76, 0x3a, 0x12,
0xa1, 0x66, 0x20, 0xac, 0xc5, 0x95, 0x69, 0x73,
0x08, 0xee, 0xd1, 0xdb, 0x05, 0xac, 0xd4, 0xf0
};
static readonly byte[] g6_y = {
0x2f, 0xc4, 0x3a, 0x11, 0xb8, 0x28, 0xc8, 0x54,
0x6a, 0xb8, 0xc1, 0xc6, 0xae, 0xb4, 0x1a, 0x68,
0x57, 0xc4, 0x81, 0x59, 0x04, 0x17, 0xb6, 0x59,
0xac, 0x8b, 0xd3, 0xee, 0x53, 0xc7, 0x0f, 0xd3,
0xf7, 0xaa, 0x13, 0xf0, 0x65, 0x70, 0x16, 0x88,
0x23, 0xaf, 0xfc, 0xe8, 0x4a, 0x5e, 0x86, 0x1d
};

// g7 (counter = 1)
static readonly byte[] g7_x = {
0xe4, 0x2b, 0xc1, 0x76, 0xe1, 0x3b, 0xde, 0xcf,
0x2b, 0x31, 0x60, 0x18, 0xa2, 0x2c, 0xf9, 0x51,
0x02, 0xb7, 0x1a, 0x1f, 0xa8, 0xac, 0x1c, 0xfe,
0x14, 0xee, 0xc5, 0x4c, 0xad, 0x25, 0x6a, 0x9a,
0xb9, 0xb4, 0xa7, 0xa8, 0xef, 0x3c, 0xfd, 0xe3,
0x38, 0x4f, 0xe4, 0xa1, 0xbc, 0x11, 0xc2, 0xf4
};
static readonly byte[] g7_y = {
0x4e, 0x26, 0x00, 0x27, 0x35, 0xc9, 0x26, 0x97,
0x87, 0x70, 0x57, 0x25, 0x60, 0x38, 0xe7, 0x93,
0x4c, 0x1d, 0x42, 0x81, 0x26, 0xf4, 0x1e, 0x77,
0x1b, 0x19, 0x08, 0x40, 0xdb, 0xac, 0x0f, 0x59,
0xbd, 0x5e, 0x07, 0x02, 0x25, 0x22, 0xb6, 0x38,
0xa2, 0x93, 0x3e, 0x14, 0x63, 0x39, 0xdc, 0xc6
};

// g8 (counter = 0)
static readonly byte[] g8_x = {
0xd7, 0x14, 0x10, 0xb3, 0xe3, 0xa2, 0x22, 0xaa,
0xfb, 0x7f, 0x53, 0xed, 0x4c, 0xd8, 0x29, 0x9e,
0x44, 0x27, 0x31, 0x20, 0x3b, 0xec, 0xf6, 0x43,
0xef, 0x81, 0xda, 0x37, 0x1d, 0x81, 0x42, 0x19,
0x1c, 0xf2, 0x5a, 0x27, 0x01, 0x95, 0x87, 0x29,
0x4d, 0x23, 0x87, 0x0c, 0x78, 0xfc, 0xa0, 0x49
};
static readonly byte[] g8_y = {
0x5d, 0xc9, 0x80, 0x0a, 0xd0, 0xed, 0xa0, 0x33,
0xab, 0xe1, 0x34, 0xdd, 0xd8, 0x94, 0xf2, 0x96,
0xb5, 0xae, 0xbd, 0x44, 0x58, 0xee, 0x42, 0x54,
0xce, 0x70, 0xf8, 0x4d, 0xfb, 0x0d, 0xe1, 0x8f,
0xca, 0x99, 0xbf, 0x79, 0x5c, 0xa7, 0xe2, 0x08,
0x18, 0x2e, 0xf6, 0xc4, 0x6d, 0x7b, 0xd4, 0x94
};

// g9 (counter = 1)
static readonly byte[] g9_x = {
0x68, 0x87, 0x2b, 0x16, 0x4a, 0x6e, 0x9b, 0x8f,
0x99, 0x66, 0x8b, 0x5b, 0xfd, 0x4a, 0xc0, 0x77,
0x0d, 0xea, 0x64, 0xe3, 0x77, 0x32, 0xa3, 0x84,
0xbc, 0x39, 0xc3, 0x24, 0x84, 0xb8, 0x60, 0x91,
0xcd, 0x47, 0xdd, 0xea, 0xe5, 0x26, 0xb1, 0x80,
0x65, 0xe8, 0x66, 0x3e, 0x1e, 0xcb, 0x8b, 0x80
};
static readonly byte[] g9_y = {
0x66, 0x60, 0x8a, 0xa7, 0xcf, 0x0a, 0xb5, 0x3c,
0x37, 0x50, 0x34, 0xdd, 0xe3, 0x97, 0x36, 0xdc,
0x81, 0xd6, 0x83, 0x1b, 0xc6, 0xee, 0x78, 0xc7,
0x14, 0xe0, 0x10, 0xaf, 0xa5, 0xc8, 0x54, 0x25,
0xa7, 0xed, 0x28, 0xef, 0x6f, 0x46, 0x2a, 0xeb,
0x4d, 0x79, 0x52, 0xea, 0xf4, 0x88, 0xd0, 0xc1
};

// g10 (counter = 0)
static readonly byte[] g10_x = {
0xa0, 0x4d, 0xec, 0xc7, 0xe9, 0xf0, 0xcf, 0x88,
0x93, 0x0a, 0xb2, 0x6c, 0x96, 0xd6, 0x95, 0x23,
0x76, 0xb4, 0xc3, 0xa3, 0xdb, 0x75, 0x25, 0x6e,
0xfd, 0xd4, 0x66, 0xf5, 0x1f, 0x7c, 0x01, 0x84,
0x1a, 0x5f, 0x4e, 0x6a, 0x9f, 0x11, 0x87, 0x7f,
0xf2, 0x86, 0xcb, 0xc7, 0x43, 0x06, 0xbb, 0xf1
};
static readonly byte[] g10_y = {
0x18, 0xc8, 0x09, 0x28, 0xac, 0x20, 0x82, 0xc0,
0x4f, 0x30, 0x0d, 0x31, 0xb2, 0xe6, 0xd0, 0x0e,
0xe6, 0x87, 0x72, 0x5e, 0xb8, 0x6f, 0x0d, 0xc7,
0xc7, 0xa8, 0xfb, 0x95, 0x99, 0x3c, 0xea, 0xca,
0x8a, 0xfc, 0xe4, 0x80, 0xed, 0xc7, 0x27, 0x84,
0x2e, 0xfe, 0xd1, 0x48, 0x88, 0x2d, 0xba, 0xa6
};

// g11 (counter = 1)
static readonly byte[] g11_x = {
0xe9, 0x80, 0x2c, 0xe5, 0x36, 0x1d, 0xd7, 0x9e,
0xb1, 0x4f, 0x00, 0x4d, 0x1e, 0x2a, 0x7d, 0xab,
0x4c, 0xa5, 0x58, 0x62, 0xb9, 0x37, 0x59, 0x3c,
0x86, 0x03, 0x5f, 0xce, 0x0d, 0x3a, 0x49, 0xc1,
0xa1, 0x34, 0x7e, 0x9d, 0x89, 0xe9, 0x34, 0x8b,
0xf8, 0x46, 0x0f, 0xe6, 0x46, 0x68, 0xaa, 0xe7
};
static readonly byte[] g11_y = {
0x29, 0x99, 0x29, 0x00, 0x81, 0x1b, 0x12, 0xc4,
0x2e, 0xfd, 0xe1, 0x23, 0xcf, 0x65, 0xb8, 0x0b,
0xb2, 0x49, 0xe4, 0x8e, 0x3f, 0x53, 0xe4, 0x47,
0x75, 0xcb, 0x1b, 0x36, 0x53, 0x6e, 0x28, 0x6b,
0x23, 0xe0, 0x8d, 0x4c, 0x59, 0xf5, 0x0d, 0xdd,
0x89, 0xb9, 0xdb, 0x4e, 0x01, 0x2a, 0x4f, 0x14
};

// g12 (counter = 1)
static readonly byte[] g12_x = {
0x85, 0x37, 0xb2, 0x9a, 0x8b, 0x60, 0xc6, 0x73,
0x94, 0xd3, 0x03, 0x78, 0xdb, 0x59, 0x0f, 0xc7,
0x04, 0xff, 0x36, 0x3a, 0x6c, 0x79, 0x01, 0xee,
0x29, 0xbb, 0x8b, 0x18, 0x3f, 0xdc, 0x8b, 0x0a,
0xea, 0xd5, 0xf4, 0x38, 0xe4, 0x43, 0x84, 0xd4,
0x1e, 0xd5, 0xf2, 0x6b, 0xe6, 0xa4, 0xc7, 0xe7
};
static readonly byte[] g12_y = {
0x75, 0xc3, 0xbe, 0x29, 0x5e, 0x38, 0x81, 0x3b,
0x27, 0x22, 0xbc, 0xfc, 0xba, 0xd4, 0x91, 0xc6,
0x2b, 0xd8, 0x1a, 0x14, 0x38, 0x86, 0x4d, 0x8f,
0x54, 0xa4, 0x8a, 0x43, 0x8a, 0x6a, 0x55, 0x6c,
0x5c, 0xc6, 0xbd, 0x95, 0x49, 0xc2, 0x5c, 0xe7,
0xe3, 0x9d, 0x98, 0x61, 0xa3, 0x0b, 0x6e, 0x08
};

// g13 (counter = 1)
static readonly byte[] g13_x = {
0xe4, 0x6b, 0xe8, 0xff, 0xb1, 0xa6, 0x12, 0x77,
0x57, 0x4b, 0x4d, 0x4e, 0x75, 0x20, 0xbb, 0x28,
0xdf, 0x1a, 0xa9, 0x2b, 0x75, 0x39, 0x0b, 0xf6,
0xaa, 0x81, 0x98, 0x84, 0xa4, 0x7d, 0xb6, 0x7e,
0x0a, 0x5a, 0x3f, 0x75, 0x4c, 0xef, 0x6d, 0xc5,
0x7d, 0x07, 0x25, 0xc7, 0x96, 0x80, 0x6d, 0x85
};
static readonly byte[] g13_y = {
0x1e, 0x0c, 0xfb, 0xc5, 0x09, 0x25, 0x82, 0xf7,
0x02, 0x00, 0x2f, 0xac, 0x85, 0xdd, 0x2f, 0x32,
0xef, 0x56, 0x8e, 0x00, 0x98, 0x01, 0xc3, 0xd7,
0x96, 0x11, 0xaa, 0x3a, 0xa0, 0xee, 0xbf, 0x2d,
0x55, 0x91, 0x01, 0x14, 0x45, 0x12, 0xfb, 0x2c,
0x1a, 0x59, 0x7f, 0x3f, 0x0b, 0x05, 0xf5, 0x43
};

// g14 (counter = 0)
static readonly byte[] g14_x = {
0xf6, 0xd6, 0xdf, 0x3c, 0x86, 0x7b, 0x88, 0x6a,
0x4b, 0xd3, 0x77, 0x56, 0x05, 0x6e, 0x72, 0x00,
0x9d, 0x2e, 0x26, 0x4c, 0xb2, 0x5d, 0xdd, 0x59,
0xc0, 0xb8, 0x3d, 0x4d, 0x0e, 0x40, 0x14, 0x4f,
0x64, 0x9f, 0x43, 0x57, 0xd4, 0x16, 0xa1, 0x77,
0x2f, 0x7a, 0x1e, 0x4e, 0x2b, 0xdd, 0xab, 0x15
};
static readonly byte[] g14_y = {
0x06, 0xd9, 0xb5, 0x0e, 0x84, 0x82, 0x46, 0xf3,
0x10, 0xb9, 0x2f, 0x01, 0x83, 0x5d, 0x53, 0xdb,
0xee, 0x8e, 0x27, 0xf8, 0x83, 0xaa, 0x6d, 0x25,
0x49, 0xe5, 0x27, 0xfe, 0x78, 0x08, 0xa9, 0xcb,
0x61, 0x92, 0x31, 0x75, 0xa8, 0xee, 0xcd, 0x33,
0x28, 0x57, 0x4a, 0x3e, 0xd7, 0xbb, 0x59, 0xba
};

// g15 (counter = 0)
static readonly byte[] g15_x = {
0x38, 0x8a, 0x6c, 0xb5, 0x5c, 0x5d, 0x08, 0xbc,
0xea, 0xd8, 0x21, 0x1c, 0xfd, 0x20, 0xe3, 0x2c,
0x78, 0xeb, 0x6f, 0x06, 0xd7, 0x92, 0x10, 0x1a,
0x00, 0xc0, 0xd7, 0x57, 0x48, 0x00, 0x46, 0xf9,
0xc4, 0xaa, 0x5d, 0xf8, 0x82, 0x17, 0x6b, 0xbc,
0x8d, 0x83, 0x1f, 0x72, 0x81, 0x4a, 0x79, 0x0c
};
static readonly byte[] g15_y = {
0x42, 0x7b, 0x89, 0x85, 0x18, 0x2f, 0x90, 0x36,
0x01, 0x9d, 0x28, 0x32, 0x56, 0x19, 0xb9, 0xca,
0x94, 0x42, 0x75, 0x08, 0x2b, 0xd2, 0xfd, 0x19,
0x85, 0x00, 0xc1, 0x7c, 0x9b, 0xb8, 0xae, 0x7d,
0x59, 0x1e, 0xfd, 0x64, 0xe1, 0x80, 0x70, 0xc4,
0xcf, 0x31, 0x64, 0xe0, 0x92, 0x6d, 0xfd, 0xcd
};

// g16 (counter = 1)
static readonly byte[] g16_x = {
0xf3, 0x0f, 0xbe, 0xec, 0x91, 0x29, 0x71, 0xdb,
0xaa, 0xd5, 0xed, 0x63, 0x3b, 0x5b, 0x2a, 0x37,
0x6a, 0xe6, 0x0e, 0x27, 0x86, 0xaf, 0x16, 0x96,
0x95, 0xaf, 0x00, 0xf6, 0xda, 0x9b, 0xbc, 0xfd,
0x9a, 0x43, 0x56, 0x40, 0x97, 0xc9, 0x02, 0x25,
0xc5, 0x4e, 0x2a, 0x63, 0xb9, 0xc0, 0x00, 0x4f
};
static readonly byte[] g16_y = {
0x31, 0xfe, 0xb5, 0x90, 0x3c, 0xb5, 0x67, 0x9f,
0xe9, 0x68, 0x6f, 0x17, 0x30, 0x3e, 0x8b, 0xcf,
0x83, 0x35, 0xfa, 0x07, 0xf6, 0xfd, 0xe0, 0x6b,
0x70, 0x62, 0xe2, 0xd3, 0x37, 0xf7, 0x2c, 0x7a,
0xa1, 0xad, 0xee, 0x5f, 0xbf, 0x5c, 0xb3, 0x74,
0x28, 0x44, 0xf0, 0x7e, 0x02, 0xfd, 0x47, 0x6d
};

// g17 (counter = 4)
static readonly byte[] g17_x = {
0xbc, 0x56, 0x18, 0x7e, 0x62, 0xb3, 0xa3, 0xc2,
0x46, 0xdf, 0x01, 0xd8, 0xf8, 0x85, 0xc3, 0x4d,
0x54, 0xff, 0x81, 0x42, 0x4a, 0xbd, 0x1d, 0x22,
0x7b, 0x03, 0x3f, 0x06, 0xec, 0xce, 0xc6, 0x27,
0x8d, 0xc0, 0x75, 0x9a, 0x16, 0xd9, 0x0f, 0x0c,
0xc5, 0x16, 0x16, 0xc5, 0x0e, 0x9a, 0x88, 0x45
};
static readonly byte[] g17_y = {
0x3a, 0x53, 0x2c, 0xd7, 0x4d, 0x1f, 0x73, 0xdc,
0x02, 0xbe, 0xfd, 0x8b, 0x00, 0x2d, 0xb3, 0x62,
0xeb, 0x13, 0x3b, 0x3d, 0x9c, 0xce, 0xc5, 0x45,
0x29, 0xf1, 0x5d, 0x73, 0x02, 0xda, 0x1d, 0x8b,
0x4c, 0x7b, 0x36, 0x65, 0x4e, 0x4f, 0x8d, 0x2a,
0x3e, 0x4d, 0xa5, 0xeb, 0x9b, 0x29, 0xa7, 0xe2
};

// g18 (counter = 1)
static readonly byte[] g18_x = {
0x89, 0x97, 0x72, 0x12, 0x6f, 0x98, 0x38, 0xec,
0x17, 0x89, 0x61, 0x50, 0x7c, 0xae, 0xd8, 0x25,
0x8b, 0x6f, 0x10, 0x2f, 0x5a, 0x77, 0x08, 0xba,
0xbf, 0x80, 0xdd, 0x1d, 0xcc, 0xdc, 0x70, 0x02,
0x1e, 0x4f, 0x41, 0xc2, 0xf7, 0x43, 0x8b, 0xeb,
0x67, 0xc9, 0xa2, 0xa9, 0xb4, 0xd5, 0x7f, 0x84
};
static readonly byte[] g18_y = {
0x71, 0xb9, 0xa6, 0xfd, 0xb9, 0x1e, 0xbd, 0x0a,
0x29, 0x2b, 0xdb, 0x71, 0x83, 0x77, 0x30, 0x8e,
0xde, 0xda, 0x06, 0x3d, 0x07, 0xcb, 0x03, 0x4e,
0x1b, 0xc8, 0x6e, 0xa2, 0xf6, 0x5f, 0xa2, 0x0f,
0x09, 0x35, 0xc6, 0xc8, 0xc3, 0x78, 0xca, 0xee,
0xfe, 0x64, 0xdd, 0xbb, 0x3a, 0xdc, 0x79, 0xed
};

// g19 (counter = 1)
static readonly byte[] g19_x = {
0xcb, 0x2e, 0xbf, 0x80, 0x7f, 0x1e, 0x6f, 0xe4,
0x11, 0xdf, 0x68, 0x98, 0xcd, 0xf6, 0x52, 0xcb,
0xb9, 0xbd, 0xdf, 0x39, 0x47, 0x35, 0x50, 0x11,
0x42, 0x9d, 0x11, 0x1b, 0xb2, 0x61, 0x8d, 0xc4,
0x6d, 0xef, 0xca, 0x46, 0x9a, 0x09, 0xc1, 0x97,
0x48, 0xcf, 0x1d, 0x09, 0xaa, 0x82, 0x19, 0xbe
};
static readonly byte[] g19_y = {
0x50, 0x23, 0x8a, 0x8b, 0x27, 0x07, 0xcd, 0xb8,
0x8c, 0x38, 0x1d, 0x57, 0x36, 0x9b, 0x4d, 0x78,
0x38, 0xd7, 0x89, 0x58, 0x76, 0xf9, 0xa3, 0xd8,
0x0a, 0x95, 0x56, 0xa5, 0xc7, 0x97, 0xa4, 0xd0,
0xdb, 0x83, 0x99, 0xfc, 0xd6, 0x57, 0xad, 0xd1,
0x93, 0x8b, 0x65, 0xc7, 0xaf, 0xad, 0x8a, 0x72
};

// g20 (counter = 0)
static readonly byte[] g20_x = {
0x95, 0xfa, 0x79, 0x5a, 0xa4, 0xf4, 0xc0, 0xda,
0x48, 0x64, 0x20, 0xfa, 0x94, 0x1b, 0x25, 0xd7,
0xf7, 0x0c, 0x80, 0x73, 0xb7, 0x8b, 0xcd, 0x88,
0x20, 0xd8, 0x14, 0x66, 0x89, 0xd8, 0x1e, 0x1d,
0xc2, 0xa4, 0x09, 0x8e, 0x86, 0xaf, 0xc2, 0x7b,
0x49, 0xc8, 0x6a, 0xef, 0xed, 0x1b, 0x0d, 0x61
};
static readonly byte[] g20_y = {
0x62, 0x3d, 0x37, 0xc1, 0x3d, 0xcb, 0x6e, 0xa9,
0x57, 0x33, 0x72, 0x88, 0x8a, 0xf0, 0x8b, 0xea,
0xcf, 0x94, 0xc8, 0xdf, 0xfc, 0x2c, 0xb6, 0x15,
0x03, 0x0a, 0xe6, 0x12, 0xb0, 0xcf, 0x14, 0x78,
0x75, 0x1d, 0xc3, 0xb5, 0x6a, 0x66, 0xa5, 0x1d,
0xb4, 0xb9, 0x8e, 0x26, 0x4e, 0xb0, 0x16, 0xdd
};

// g21 (counter = 0)
static readonly byte[] g21_x = {
0x22, 0xf4, 0x33, 0x02, 0x0d, 0xc1, 0x29, 0xc7,
0xbe, 0x74, 0x55, 0x8e, 0xf9, 0xc2, 0x91, 0xf3,
0x93, 0x8e, 0x78, 0x17, 0xb4, 0x7d, 0x4b, 0x41,
0xa5, 0x92, 0x21, 0xd8, 0x5b, 0x10, 0xff, 0xd1,
0xb8, 0x15, 0x91, 0x9f, 0xb3, 0x71, 0x7e, 0x3e,
0x7e, 0x15, 0xe9, 0x3f, 0xb9, 0x7f, 0x6f, 0x7c
};
static readonly byte[] g21_y = {
0x14, 0x45, 0x4c, 0xae, 0x0e, 0xec, 0xcb, 0xcc,
0x8e, 0x52, 0x15, 0x55, 0xfd, 0x2e, 0xc8, 0x22,
0xc2, 0x90, 0xc4, 0x64, 0x67, 0x1e, 0xa7, 0xaa,
0x99, 0xd5, 0x58, 0x90, 0x9f, 0xdc, 0xcc, 0xf6,
0x8e, 0x4d, 0x70, 0x95, 0xc4, 0x64, 0x7a, 0x16,
0xe4, 0x7f, 0x16, 0xc0, 0xb6, 0x88, 0x51, 0xc1
};

// g22 (counter = 0)
static readonly byte[] g22_x = {
0x45, 0x69, 0x82, 0xb2, 0x35, 0xd9, 0xd0, 0x13,
0xc9, 0x9b, 0x64, 0x09, 0x4d, 0x41, 0x29, 0x63,
0x1f, 0xb1, 0xc6, 0x21, 0x06, 0x28, 0x50, 0x5c,
0x74, 0x41, 0x33, 0xe6, 0xfa, 0x17, 0x5d, 0x14,
0x1f, 0xb4, 0xc0, 0x01, 0x05, 0xf8, 0x10, 0x28,
0x4c, 0x68, 0x80, 0xb4, 0x6a, 0x44, 0x06, 0xdf
};
static readonly byte[] g22_y = {
0x52, 0xb8, 0x49, 0x82, 0x56, 0x51, 0x6a, 0x4f,
0xef, 0xde, 0x13, 0xb5, 0xa7, 0xbb, 0xd7, 0x2d,
0x3f, 0x19, 0xaa, 0x00, 0xb3, 0x62, 0x6d, 0xec,
0xdd, 0x9c, 0xd1, 0xff, 0x7d, 0x17, 0x5c, 0xf7,
0x44, 0xe2, 0x24, 0x16, 0xf3, 0x51, 0xf6, 0x2e,
0x5d, 0x01, 0xbe, 0x65, 0x1c, 0xa8, 0x27, 0x47
};

// g23 (counter = 2)
static readonly byte[] g23_x = {
0x45, 0x1f, 0x77, 0xcb, 0xcf, 0x22, 0xbe, 0xe6,
0xa4, 0x07, 0x28, 0x7e, 0xf9, 0xa3, 0x6f, 0x29,
0x3f, 0xa8, 0x22, 0xf3, 0x95, 0xf6, 0x4c, 0x2e,
0xdc, 0xcb, 0x9a, 0xb5, 0xf8, 0xee, 0x3f, 0xde,
0x86, 0xef, 0xdd, 0xcf, 0x33, 0x02, 0xe8, 0xe9,
0x29, 0x37, 0x32, 0xa0, 0x58, 0x33, 0x28, 0x16
};
static readonly byte[] g23_y = {
0x77, 0x34, 0x01, 0xa1, 0x0f, 0xe5, 0x06, 0x9d,
0xbc, 0xc6, 0x87, 0x0c, 0xeb, 0x8b, 0x15, 0xd1,
0xcb, 0x35, 0x22, 0x7b, 0xd8, 0xaf, 0x7d, 0x70,
0xb6, 0x3d, 0x36, 0xe9, 0x56, 0x13, 0xde, 0xba,
0x2d, 0x60, 0x03, 0x83, 0x50, 0x27, 0x49, 0x3c,
0x04, 0x63, 0x0e, 0xdb, 0x27, 0x00, 0xb9, 0x65
};

// g24 (counter = 2)
static readonly byte[] g24_x = {
0xa4, 0xad, 0x50, 0xb7, 0xdb, 0xcf, 0xcd, 0xc4,
0x27, 0xe7, 0x2f, 0x85, 0x0a, 0x1b, 0xb0, 0x40,
0x28, 0x1c, 0x5f, 0x99, 0xb9, 0x14, 0x15, 0x1c,
0x47, 0xac, 0x48, 0xf9, 0xfb, 0x7b, 0x85, 0xa0,
0x58, 0x58, 0xf3, 0x03, 0x58, 0x8c, 0x57, 0xd2,
0xff, 0x66, 0xb5, 0x86, 0x71, 0x45, 0xfb, 0xdb
};
static readonly byte[] g24_y = {
0x0a, 0x56, 0x43, 0x36, 0xb5, 0xe3, 0x7a, 0xac,
0x39, 0xba, 0xa0, 0x87, 0x8c, 0x6c, 0x50, 0xd3,
0xd3, 0x6f, 0x54, 0x09, 0xf1, 0x02, 0xf9, 0xb8,
0x68, 0x73, 0x28, 0x09, 0x45, 0x62, 0xca, 0x62,
0x88, 0xb2, 0xb6, 0x9f, 0xee, 0x43, 0x89, 0x1a,
0xd5, 0x61, 0xd3, 0x2e, 0xd4, 0xbb, 0x20, 0xf8
};

// g25 (counter = 0)
static readonly byte[] g25_x = {
0xf1, 0xb4, 0xf1, 0x32, 0xb3, 0xc2, 0x9a, 0x9e,
0x34, 0x67, 0xa0, 0x22, 0x08, 0x17, 0xf2, 0x58,
0x78, 0x43, 0xe7, 0x75, 0x43, 0xe8, 0x12, 0xec,
0x52, 0x4d, 0x7d, 0x41, 0x3c, 0x6c, 0x20, 0xc0,
0x3e, 0xc3, 0x9b, 0x55, 0x83, 0x62, 0x20, 0x71,
0x7d, 0xdd, 0x9a, 0xf4, 0x2a, 0xff, 0xe6, 0x67
};
static readonly byte[] g25_y = {
0x34, 0x47, 0xa1, 0x34, 0x2d, 0x40, 0xa8, 0xc0,
0x94, 0x06, 0x03, 0x45, 0x10, 0x2e, 0x64, 0xb1,
0xb3, 0x87, 0x1b, 0x80, 0xef, 0x28, 0xd3, 0x27,
0x70, 0xf8, 0x49, 0xb5, 0x7e, 0x76, 0x96, 0x11,
0x8f, 0x59, 0x6b, 0x8b, 0x11, 0x9b, 0xab, 0xdd,
0x2d, 0x7c, 0xcf, 0x27, 0xb5, 0x5a, 0xcd, 0x17
};

// g26 (counter = 0)
static readonly byte[] g26_x = {
0x09, 0x70, 0x27, 0x9c, 0x58, 0x14, 0x7d, 0xd4,
0x0d, 0x2a, 0xaf, 0xeb, 0xce, 0x3e, 0x0b, 0xc1,
0x3e, 0xe4, 0xdd, 0xbb, 0x74, 0x69, 0x09, 0xcc,
0xe2, 0x84, 0x9b, 0xed, 0x98, 0xd8, 0xb7, 0xb3,
0x6e, 0xd9, 0x71, 0xa1, 0x7a, 0x8e, 0x75, 0xcd,
0xb5, 0xf3, 0xeb, 0xdc, 0x3e, 0x71, 0xb4, 0x7b
};
static readonly byte[] g26_y = {
0x2c, 0xc1, 0x7a, 0xcd, 0xa0, 0x96, 0xb5, 0x63,
0x93, 0xde, 0x63, 0xc6, 0x6b, 0x23, 0x42, 0xa0,
0x3e, 0x2e, 0x50, 0x4f, 0x12, 0x39, 0x80, 0x01,
0xb8, 0xd9, 0x0f, 0xe1, 0x58, 0x1d, 0x26, 0x96,
0xbf, 0x86, 0x43, 0x46, 0x58, 0x54, 0xe0, 0xd9,
0xad, 0xe9, 0xee, 0x21, 0x45, 0xf3, 0xc0, 0x2f
};

// g27 (counter = 5)
static readonly byte[] g27_x = {
0x2b, 0x5e, 0x17, 0xc2, 0x19, 0xe4, 0xea, 0x95,
0xaf, 0x39, 0xb2, 0x4a, 0x20, 0xdc, 0x4d, 0x50,
0x86, 0x70, 0x7e, 0xd8, 0x06, 0x6b, 0xd2, 0x98,
0xe3, 0x04, 0x7d, 0x64, 0x59, 0x82, 0x3f, 0xf1,
0x02, 0xef, 0x61, 0x7d, 0x46, 0x70, 0xd5, 0x9e,
0x0e, 0xfb, 0x96, 0x5e, 0x17, 0x9b, 0x01, 0xa6
};
static readonly byte[] g27_y = {
0x24, 0x20, 0x23, 0x56, 0x7c, 0x67, 0x0b, 0xcf,
0x21, 0xc9, 0xc2, 0xef, 0x69, 0xbb, 0x8b, 0x87,
0xd7, 0xa8, 0xdc, 0xd0, 0x34, 0xb6, 0x66, 0x06,
0x90, 0x08, 0x12, 0x39, 0x3e, 0xe9, 0x98, 0x9b,
0xc3, 0x7e, 0x25, 0xab, 0x67, 0x2a, 0x3e, 0xa4,
0xef, 0x3b, 0x02, 0x8f, 0x83, 0x49, 0x56, 0x23
};

// g28 (counter = 2)
static readonly byte[] g28_x = {
0x81, 0xfe, 0x66, 0xdc, 0x8e, 0x7c, 0x58, 0x3e,
0xd1, 0x52, 0x44, 0xb5, 0xd7, 0xec, 0x86, 0x97,
0x25, 0xbc, 0x35, 0x34, 0x6f, 0xf3, 0x06, 0x3e,
0x6a, 0x22, 0x48, 0x3e, 0x92, 0xd6, 0x70, 0x8b,
0x59, 0x8f, 0x41, 0xe4, 0x26, 0x1d, 0x4c, 0x52,
0xc8, 0x1f, 0xd7, 0x03, 0x85, 0x3c, 0x0b, 0x5d
};
static readonly byte[] g28_y = {
0x23, 0xf7, 0x5a, 0x94, 0xe5, 0xf8, 0x64, 0x85,
0x06, 0x97, 0x61, 0x9c, 0xab, 0xb8, 0xb3, 0x4e,
0xff, 0x56, 0x9f, 0x29, 0xd8, 0x82, 0x75, 0xb7,
0xa7, 0xae, 0x1d, 0x77, 0x80, 0x9e, 0x3a, 0x5a,
0x42, 0x5c, 0x82, 0xb5, 0xbd, 0xb8, 0xd7, 0xaa,
0x75, 0xf9, 0x5c, 0x53, 0xa3, 0x7c, 0xf9, 0x60
};

// g29 (counter = 0)
static readonly byte[] g29_x = {
0x46, 0xe4, 0xd4, 0xbc, 0x62, 0x81, 0xd4, 0x94,
0x3e, 0x0b, 0x5c, 0xea, 0xd8, 0x6e, 0xb8, 0xf5,
0x82, 0x1b, 0xff, 0xd4, 0x28, 0x88, 0xb5, 0x79,
0xdb, 0x71, 0x71, 0xc8, 0x24, 0x16, 0xbc, 0xf8,
0x63, 0x18, 0x7d, 0x86, 0x81, 0x55, 0xb4, 0xa0,
0xae, 0x1f, 0x3e, 0x44, 0xcf, 0x71, 0x07, 0xaf
};
static readonly byte[] g29_y = {
0x14, 0x64, 0x24, 0x6c, 0x54, 0xca, 0x3a, 0x8b,
0x70, 0xda, 0x3d, 0xa9, 0x3f, 0xfd, 0x88, 0x66,
0x33, 0x6e, 0x7a, 0xba, 0xda, 0x50, 0x9d, 0x39,
0x27, 0x6d, 0x08, 0x5a, 0x9c, 0x9d, 0xd5, 0x5c,
0x04, 0xff, 0xde, 0xf0, 0xc5, 0xec, 0x06, 0x98,
0x3a, 0xab, 0x0e, 0x63, 0xfc, 0x11, 0x86, 0xfd
};

// g30 (counter = 0)
static readonly byte[] g30_x = {
0xe3, 0xd3, 0x27, 0x37, 0x71, 0x44, 0xa4, 0x06,
0xff, 0x2d, 0x22, 0xfb, 0x26, 0xca, 0x63, 0x4d,
0x6a, 0x8a, 0x0e, 0xe4, 0xbb, 0x39, 0xaa, 0xd0,
0xb3, 0x5e, 0x66, 0x36, 0xac, 0xcc, 0xca, 0xf2,
0x0a, 0xca, 0x78, 0xfc, 0x1a, 0x02, 0xcc, 0x60,
0x48, 0x05, 0x33, 0x30, 0x34, 0x07, 0x8e, 0xf8
};
static readonly byte[] g30_y = {
0x47, 0xec, 0x98, 0xd3, 0x18, 0xf8, 0x71, 0xcb,
0x6f, 0xe4, 0x91, 0xbf, 0xda, 0xf2, 0x61, 0xd8,
0x62, 0xac, 0x92, 0xea, 0x5d, 0x26, 0x9b, 0x94,
0x18, 0x75, 0xe6, 0x36, 0xdc, 0x5c, 0xe2, 0xcc,
0x74, 0x66, 0x90, 0x8a, 0x60, 0xaf, 0x47, 0x9a,
0xa2, 0xc7, 0x0f, 0x94, 0xba, 0x0a, 0xd3, 0x03
};

// g31 (counter = 3)
static readonly byte[] g31_x = {
0xbd, 0x42, 0x54, 0x97, 0x21, 0x9e, 0x2b, 0xca,
0x3b, 0x47, 0x62, 0x9b, 0x19, 0xf7, 0x9d, 0x4d,
0xcb, 0x27, 0x62, 0x84, 0xb4, 0x5c, 0x3f, 0x5a,
0xc9, 0xb4, 0x7b, 0x76, 0xb9, 0x0e, 0xce, 0x92,
0x35, 0x65, 0x5c, 0xf9, 0x57, 0xf7, 0x7a, 0xda,
0x90, 0x2d, 0xd1, 0x75, 0xe4, 0x94, 0xda, 0x91
};
static readonly byte[] g31_y = {
0x29, 0x80, 0xe0, 0xc8, 0xd4, 0x4a, 0x32, 0x34,
0x52, 0x11, 0x55, 0x4b, 0xf1, 0xb7, 0xac, 0x1e,
0xde, 0xf0, 0xc5, 0xcc, 0x1a, 0x60, 0xd4, 0x2a,
0xd4, 0xe1, 0x9e, 0x9b, 0x9a, 0xaa, 0xe4, 0xcf,
0x1f, 0xbc, 0xa9, 0xeb, 0x01, 0xe1, 0x32, 0xdf,
0x66, 0xf8, 0x98, 0x57, 0xb9, 0x30, 0x30, 0x08
};

// g32 (counter = 0)
static readonly byte[] g32_x = {
0x2a, 0x6c, 0x57, 0x6d, 0xf2, 0xaf, 0x7b, 0x14,
0x05, 0xda, 0xad, 0xe9, 0xfb, 0x24, 0xbb, 0xfe,
0xb3, 0xfa, 0x9c, 0x55, 0x86, 0x03, 0x60, 0x88,
0xe0, 0x7d, 0x9d, 0xbd, 0x31, 0x55, 0xe9, 0xa9,
0x69, 0x22, 0xae, 0xe1, 0x85, 0xf7, 0x31, 0x57,
0x9c, 0x7d, 0x8c, 0xd1, 0xa7, 0xa3, 0x44, 0xe4
};
static readonly byte[] g32_y = {
0x44, 0x91, 0x6f, 0xe5, 0xda, 0x07, 0xe5, 0xad,
0xfe, 0xf1, 0x83, 0x7d, 0x5d, 0x3f, 0xc2, 0xae,
0xdd, 0xd5, 0xb0, 0x5b, 0x9c, 0xd9, 0xd4, 0x07,
0x15, 0xfe, 0xa4, 0x42, 0x11, 0x48, 0x6d, 0x82,
0xce, 0x95, 0xa2, 0x72, 0x96, 0xb5, 0xab, 0x3a,
0x90, 0x1f, 0x63, 0xd5, 0x01, 0xf2, 0xb1, 0xf7
};

// g33 (counter = 0)
static readonly byte[] g33_x = {
0xe7, 0xe5, 0xb4, 0x2c, 0x72, 0x2f, 0xb4, 0x7b,
0xd9, 0x2c, 0xa5, 0x81, 0xc0, 0x5a, 0xa0, 0x0f,
0x1c, 0x18, 0xc0, 0xff, 0x65, 0x49, 0x39, 0xf1,
0x9c, 0xcc, 0xb6, 0xed, 0x3c, 0x4d, 0xf9, 0xf7,
0xb0, 0xa0, 0x3b, 0xb4, 0x95, 0xc4, 0x4c, 0xbb,
0xbc, 0xe9, 0x3f, 0x48, 0x95, 0x88, 0x26, 0x37
};
static readonly byte[] g33_y = {
0x30, 0x92, 0xe3, 0x2a, 0xf3, 0x9c, 0x40, 0xc2,
0x4b, 0x5a, 0x11, 0x10, 0x0b, 0xe8, 0x10, 0x32,
0x9f, 0x05, 0x01, 0x6f, 0x79, 0x2f, 0xaf, 0x50,
0x87, 0x6f, 0x77, 0xf4, 0xf5, 0x61, 0xf1, 0x87,
0x66, 0xb0, 0x1d, 0xa4, 0x4f, 0xb2, 0xe5, 0x4e,
0x84, 0x94, 0x6f, 0x2a, 0x3c, 0x96, 0x70, 0xb0
};

// g34 (counter = 0)
static readonly byte[] g34_x = {
0xa6, 0x95, 0x9a, 0x68, 0xa6, 0xda, 0x7a, 0x46,
0x8d, 0x0e, 0xef, 0x82, 0x74, 0x92, 0xd3, 0x88,
0x8a, 0xf2, 0x81, 0x1d, 0x06, 0x27, 0xc1, 0x8a,
0xa9, 0x09, 0xdd, 0x0a, 0xe4, 0xb0, 0x20, 0xd1,
0x33, 0xe4, 0x8b, 0x91, 0x1b, 0x37, 0x7e, 0x05,
0xcc, 0xb7, 0x7f, 0xee, 0xea, 0x32, 0x66, 0x73
};
static readonly byte[] g34_y = {
0x70, 0xed, 0xeb, 0x46, 0x7c, 0x18, 0xd1, 0x59,
0x47, 0x48, 0x24, 0xdf, 0xbe, 0x59, 0x42, 0x9b,
0x4f, 0xad, 0xc9, 0x7c, 0xfe, 0x3f, 0x84, 0x98,
0x63, 0x07, 0xb4, 0xba, 0x50, 0xfa, 0xf2, 0xc4,
0xb6, 0x0d, 0x73, 0xf2, 0xb3, 0x95, 0xb5, 0x95,
0xbb, 0x6e, 0xcd, 0xd9, 0x6a, 0x4f, 0xe3, 0xe9
};

// g35 (counter = 5)
static readonly byte[] g35_x = {
0x99, 0x38, 0x12, 0x7c, 0x58, 0xef, 0x9e, 0xfe,
0x69, 0xcc, 0x43, 0xad, 0x75, 0x8b, 0x3a, 0xf2,
0x3b, 0xff, 0xde, 0x84, 0xf7, 0x18, 0x2f, 0x09,
0x00, 0x95, 0xe1, 0x18, 0x83, 0x4b, 0x67, 0x08,
0x7e, 0xe7, 0x06, 0xb4, 0x64, 0x40, 0x32, 0x38,
0x89, 0x25, 0x3a, 0x22, 0x3e, 0xcf, 0xc8, 0x65
};
static readonly byte[] g35_y = {
0x37, 0xf1, 0xc9, 0x42, 0x4f, 0x08, 0x60, 0xe3,
0xec, 0xe2, 0xf8, 0x10, 0x78, 0x00, 0xf6, 0xd8,
0x6c, 0x90, 0x9a, 0x9f, 0x44, 0x67, 0x9c, 0x35,
0x57, 0xec, 0xeb, 0x4b, 0x58, 0x14, 0xb4, 0xb3,
0x96, 0xbc, 0x8e, 0x9a, 0x9a, 0x78, 0xbb, 0xea,
0xb7, 0xc1, 0x31, 0x3f, 0x0a, 0x4b, 0xe3, 0x18
};

// g36 (counter = 0)
static readonly byte[] g36_x = {
0x60, 0xea, 0xe1, 0x69, 0xe1, 0x38, 0xdf, 0x49,
0x79, 0xde, 0xd5, 0x09, 0xba, 0xce, 0xed, 0x03,
0xe6, 0x34, 0x49, 0x10, 0xd3, 0x61, 0x35, 0xea,
0x66, 0x53, 0xef, 0xf5, 0x5c, 0x07, 0x6c, 0xfb,
0x54, 0xc1, 0xee, 0x9b, 0xc0, 0xb5, 0xe8, 0x12,
0x9c, 0x92, 0x9f, 0xda, 0x64, 0x3e, 0x82, 0xd9
};
static readonly byte[] g36_y = {
0x65, 0x21, 0x02, 0x12, 0xca, 0x85, 0x21, 0xa0,
0xf9, 0x02, 0x20, 0x63, 0xc3, 0xd8, 0x16, 0x34,
0xa1, 0xb4, 0x64, 0x9b, 0x02, 0x65, 0x19, 0x87,
0xfc, 0xe2, 0xa8, 0x5e, 0x3f, 0xae, 0xa6, 0x03,
0x5b, 0xbf, 0xa2, 0x5e, 0x4d, 0xe4, 0x76, 0x95,
0xd7, 0xef, 0x0b, 0xe7, 0xc6, 0x55, 0x29, 0xf4
};

// g37 (counter = 0)
static readonly byte[] g37_x = {
0xab, 0x09, 0xb0, 0x48, 0x5b, 0xd7, 0xea, 0xe2,
0x4b, 0xf8, 0x2a, 0xca, 0xf4, 0x8c, 0xb3, 0x84,
0x8a, 0x84, 0x1f, 0xe3, 0x82, 0x6c, 0xa2, 0xbf,
0xd9, 0x1d, 0x2a, 0xfe, 0xdf, 0xbe, 0xef, 0x07,
0xa9, 0x1d, 0x1d, 0x35, 0xb7, 0xc4, 0x22, 0x44,
0x53, 0x15, 0xd7, 0x46, 0x66, 0xa2, 0xdc, 0x5b
};
static readonly byte[] g37_y = {
0x6b, 0xe9, 0x10, 0x29, 0xfe, 0xd9, 0x01, 0x32,
0x98, 0x2e, 0x34, 0xf5, 0x81, 0xb8, 0x5c, 0xba,
0x8d, 0x69, 0x90, 0x62, 0xb7, 0x65, 0x78, 0x80,
0x86, 0x93, 0xf1, 0x34, 0xf2, 0x03, 0xfa, 0x86,
0x6a, 0x1f, 0xb2, 0xfa, 0xfc, 0x87, 0x77, 0x29,
0x2b, 0xb5, 0xc6, 0x2c, 0x13, 0x0a, 0x94, 0x10
};

// g38 (counter = 1)
static readonly byte[] g38_x = {
0xeb, 0xca, 0x9a, 0x02, 0x48, 0x2d, 0x74, 0x55,
0xd0, 0xe6, 0xaf, 0x49, 0x2e, 0x61, 0x1e, 0xfc,
0x90, 0x8c, 0x1f, 0xb8, 0x51, 0xb5, 0x8f, 0x33,
0x1b, 0x2a, 0xb0, 0x87, 0xa2, 0xd9, 0xb6, 0x48,
0xf3, 0x0f, 0x3f, 0x65, 0x53, 0x3f, 0x4c, 0x55,
0x09, 0x75, 0xbe, 0xb9, 0x16, 0x8b, 0x08, 0x49
};
static readonly byte[] g38_y = {
0x68, 0x5d, 0xf7, 0xb2, 0xc2, 0x75, 0x7a, 0x5b,
0x5b, 0xa9, 0x3d, 0x91, 0x28, 0x63, 0x88, 0xf6,
0x57, 0xb8, 0x9b, 0xd1, 0x2e, 0x36, 0xc8, 0xde,
0x9c, 0x95, 0x81, 0x8a, 0xa4, 0xfd, 0x35, 0x7b,
0x4c, 0x14, 0xdf, 0x72, 0xa2, 0xbc, 0x41, 0x36,
0xfd, 0x94, 0x76, 0x16, 0x0a, 0xaf, 0x65, 0x1c
};

// g39 (counter = 0)
static readonly byte[] g39_x = {
0xda, 0xb7, 0x9f, 0xea, 0x35, 0xa2, 0x88, 0x3a,
0x43, 0xc0, 0x28, 0x52, 0xd2, 0x8d, 0xc8, 0xb7,
0xa0, 0x2d, 0x5f, 0x33, 0x1c, 0xf1, 0xa5, 0xb9,
0x61, 0xaa, 0x2b, 0x39, 0xa0, 0x36, 0x75, 0x6b,
0xfc, 0x80, 0xf5, 0x1f, 0x94, 0x57, 0x14, 0x20,
0x9b, 0x0b, 0xe8, 0x81, 0xbb, 0x82, 0xc1, 0xb6
};
static readonly byte[] g39_y = {
0x4c, 0x2d, 0x06, 0x3d, 0x40, 0x5b, 0x45, 0x69,
0xaa, 0x6d, 0xc9, 0x29, 0x26, 0xe3, 0xfb, 0x36,
0x3b, 0x3f, 0xec, 0x76, 0x2d, 0x0c, 0x6f, 0x0d,
0xb2, 0x67, 0x07, 0x3c, 0xfe, 0x3f, 0x78, 0xba,
0xa2, 0xb8, 0xcb, 0x5d, 0x70, 0x00, 0xbc, 0x8f,
0xdb, 0x28, 0x1f, 0xf1, 0x48, 0xe6, 0x26, 0x76
};

// g40 (counter = 2)
static readonly byte[] g40_x = {
0x2e, 0xa5, 0x92, 0x5e, 0xed, 0xe0, 0x69, 0x6e,
0x05, 0xb5, 0x75, 0xd1, 0xca, 0x44, 0x66, 0xb1,
0xa7, 0x44, 0x90, 0xfc, 0x91, 0xd2, 0x12, 0x61,
0x25, 0x7e, 0x06, 0xc6, 0xc5, 0xd8, 0xd9, 0x58,
0xdc, 0x1b, 0x34, 0xa3, 0xc2, 0xd1, 0x46, 0x07,
0xed, 0xbc, 0x15, 0x19, 0x9e, 0xe3, 0x6e, 0x73
};
static readonly byte[] g40_y = {
0x76, 0x8d, 0xb9, 0xae, 0x8d, 0xe6, 0xa9, 0xd5,
0x1b, 0x73, 0x25, 0xf1, 0x74, 0x71, 0xed, 0x64,
0x22, 0xb8, 0x97, 0x7f, 0xbd, 0xfe, 0xfe, 0xfd,
0xeb, 0x0d, 0x5e, 0xb2, 0x87, 0x9c, 0xf5, 0xb9,
0x9d, 0xe7, 0xa0, 0x6b, 0xee, 0xf3, 0x5d, 0x4d,
0x4b, 0x47, 0x1e, 0x3d, 0xae, 0x96, 0xae, 0x94
};

// g41 (counter = 1)
static readonly byte[] g41_x = {
0x5f, 0xc2, 0x71, 0x13, 0xe1, 0x0e, 0xb4, 0x15,
0x11, 0x88, 0x06, 0x86, 0x94, 0x90, 0x56, 0x9e,
0x9a, 0x42, 0xd4, 0xe3, 0xce, 0xb2, 0x27, 0xd8,
0x93, 0x6d, 0xce, 0xa2, 0x7c, 0xad, 0xb1, 0x6a,
0x86, 0xfe, 0x6c, 0x11, 0x6e, 0x60, 0x75, 0x1a,
0xfc, 0x53, 0x54, 0xf2, 0x9f, 0x98, 0x24, 0xdd
};
static readonly byte[] g41_y = {
0x1c, 0x7e, 0x89, 0x0e, 0xa7, 0x6e, 0xca, 0xfe,
0x28, 0xb6, 0x54, 0x5a, 0x20, 0x05, 0xb5, 0xdf,
0x5f, 0xc2, 0x39, 0x18, 0xaa, 0x6d, 0x4c, 0x81,
0xf9, 0x68, 0x10, 0x8f, 0xf5, 0x65, 0xfb, 0xee,
0x01, 0x88, 0x2d, 0xee, 0x44, 0xb2, 0x3f, 0xb4,
0x24, 0x27, 0xd8, 0x94, 0x5f, 0x49, 0xf6, 0xf4
};

// g42 (counter = 0)
static readonly byte[] g42_x = {
0x19, 0xf7, 0x36, 0x91, 0x50, 0x14, 0xc1, 0x21,
0x61, 0x2f, 0xf0, 0xfb, 0xaa, 0x44, 0x79, 0xed,
0xba, 0x30, 0x84, 0x07, 0x36, 0xd0, 0x0f, 0x18,
0xd9, 0xc0, 0x8c, 0xfe, 0x17, 0x70, 0xae, 0xe9,
0x12, 0xb5, 0x16, 0xf8, 0x2a, 0x71, 0xf1, 0x76,
0x25, 0xbc, 0x10, 0xd5, 0x6e, 0xa7, 0x55, 0x86
};
static readonly byte[] g42_y = {
0x1a, 0x5a, 0x31, 0x18, 0x94, 0xe9, 0x0d, 0x01,
0x9f, 0xa1, 0x54, 0xc6, 0x89, 0x07, 0x06, 0x8d,
0x1a, 0x29, 0x07, 0xe3, 0x3f, 0xd6, 0x83, 0x61,
0x4f, 0xa0, 0x2c, 0xe9, 0xe2, 0x2d, 0x4c, 0xf4,
0x0f, 0x9c, 0xb7, 0x0f, 0xfa, 0x74, 0x38, 0x83,
0x41, 0xe9, 0x06, 0x35, 0xa2, 0xe2, 0x60, 0xf4
};

// g43 (counter = 1)
static readonly byte[] g43_x = {
0x29, 0x76, 0x37, 0x35, 0xc7, 0xf5, 0x6b, 0xf6,
0xbd, 0x7a, 0x1b, 0x6a, 0x1f, 0x2f, 0x87, 0xbc,
0x7c, 0xd4, 0x85, 0x92, 0x27, 0x0a, 0xf4, 0x65,
0x97, 0x05, 0x31, 0xd6, 0xd9, 0xfa, 0x9a, 0x29,
0x9c, 0x40, 0x73, 0xc2, 0xef, 0x5e, 0xd3, 0xf9,
0x60, 0x5c, 0x7d, 0xd4, 0x33, 0xb2, 0x08, 0xbc
};
static readonly byte[] g43_y = {
0x74, 0x84, 0xff, 0x65, 0x32, 0xc7, 0xfb, 0x29,
0x4e, 0xd4, 0x77, 0x4a, 0x62, 0x9a, 0xa2, 0xeb,
0xaf, 0xd0, 0x2b, 0x91, 0x6e, 0xcd, 0x91, 0x84,
0xc1, 0x34, 0xc7, 0x39, 0xcd, 0x2a, 0x59, 0x2a,
0x40, 0x98, 0x4c, 0xd2, 0x1a, 0x61, 0xc8, 0x80,
0x06, 0x7e, 0xef, 0x2c, 0x96, 0x0b, 0x23, 0xa2
};

// g44 (counter = 1)
static readonly byte[] g44_x = {
0x82, 0xde, 0xd0, 0x39, 0x57, 0x7f, 0x4e, 0x1d,
0xb3, 0x87, 0x22, 0x20, 0x0a, 0xdc, 0x9d, 0xee,
0xe4, 0x77, 0xc8, 0x92, 0x94, 0x84, 0x32, 0xea,
0x03, 0x82, 0xf3, 0xd3, 0x14, 0xdc, 0xcb, 0xb8,
0x94, 0x47, 0x18, 0xfb, 0xbf, 0x92, 0xc3, 0x1a,
0x89, 0xa8, 0xc1, 0x0d, 0xaa, 0x77, 0x8a, 0xb0
};
static readonly byte[] g44_y = {
0x25, 0x81, 0x02, 0x80, 0xfc, 0xf2, 0x80, 0x9f,
0xb8, 0xe4, 0x6e, 0xa5, 0xf7, 0x5f, 0x9c, 0x8c,
0xd4, 0xae, 0x3e, 0x56, 0x0a, 0xaf, 0xb5, 0xa0,
0xfe, 0x8b, 0xce, 0x7d, 0x8a, 0xc2, 0x81, 0x1e,
0x71, 0x3f, 0xdf, 0xca, 0x7c, 0x28, 0x14, 0x84,
0x1e, 0x64, 0xad, 0xc3, 0x2b, 0xfb, 0x66, 0x2d
};

// g45 (counter = 2)
static readonly byte[] g45_x = {
0x1b, 0x41, 0x50, 0x2e, 0xcd, 0x20, 0x81, 0x7d,
0x3c, 0x85, 0xe7, 0x00, 0x4e, 0x66, 0x96, 0x38,
0xc7, 0x25, 0xd1, 0x15, 0x8a, 0xa8, 0x03, 0xd1,
0x05, 0xab, 0xbf, 0x85, 0x95, 0x51, 0x85, 0x59,
0x8b, 0xd3, 0x16, 0xfb, 0xbc, 0x1e, 0x9d, 0xa4,
0xb0, 0xab, 0x46, 0x7e, 0x34, 0x17, 0x2d, 0x41
};
static readonly byte[] g45_y = {
0x0e, 0x37, 0x08, 0xf7, 0x21, 0xdc, 0x69, 0xf9,
0x2a, 0xf8, 0x11, 0x77, 0x6f, 0xb1, 0xe2, 0x48,
0x02, 0x51, 0xb2, 0x04, 0xf5, 0xb1, 0xcd, 0xf4,
0xad, 0xaa, 0xed, 0x66, 0xb0, 0x69, 0x8b, 0xd9,
0xfa, 0x66, 0x62, 0x3c, 0x5b, 0xf0, 0x56, 0xbe,
0xe3, 0x4f, 0x8c, 0xba, 0x26, 0x94, 0x4a, 0x7c
};

// g46 (counter = 1)
static readonly byte[] g46_x = {
0x4a, 0xe5, 0x76, 0xa4, 0x16, 0x41, 0xac, 0x9c,
0xee, 0x68, 0xd7, 0x61, 0x68, 0xca, 0x2d, 0xd3,
0xa5, 0xa8, 0xa1, 0xc2, 0xd1, 0xb0, 0x38, 0x2d,
0xf8, 0xc4, 0xbf, 0x77, 0x00, 0x8b, 0xe3, 0xb2,
0x2d, 0x16, 0xbf, 0xa9, 0xda, 0x4a, 0x5a, 0x1c,
0xd2, 0x77, 0xa3, 0x13, 0x69, 0xad, 0x04, 0xfe
};
static readonly byte[] g46_y = {
0x53, 0xc8, 0xce, 0x7a, 0x42, 0xd2, 0xb8, 0xd5,
0xb9, 0x40, 0x95, 0x93, 0x1a, 0x00, 0x5b, 0xd3,
0x20, 0x59, 0x39, 0x55, 0xe6, 0x41, 0xb9, 0xd4,
0x11, 0x02, 0xf8, 0x1f, 0x34, 0x46, 0x5e, 0xd9,
0x67, 0xd6, 0x99, 0x92, 0xb7, 0x81, 0x0e, 0x0b,
0x53, 0x68, 0xe1, 0x1f, 0xa9, 0xbf, 0x4b, 0xc2
};

// g47 (counter = 0)
static readonly byte[] g47_x = {
0xe2, 0xab, 0x87, 0x53, 0xb1, 0x1e, 0xc3, 0x52,
0xfb, 0xdc, 0x31, 0x81, 0x9a, 0xf1, 0x93, 0x7c,
0x1d, 0x72, 0x2d, 0x10, 0x0b, 0x6d, 0x8a, 0x0a,
0x9d, 0xfe, 0xaf, 0x5b, 0xfe, 0x26, 0x1f, 0x78,
0x80, 0x1a, 0x0b, 0x80, 0xa2, 0x02, 0x8c, 0x76,
0x7e, 0x57, 0x90, 0xce, 0xa9, 0x4e, 0xca, 0x1d
};
static readonly byte[] g47_y = {
0x04, 0xd7, 0xc7, 0x10, 0x08, 0xa5, 0x0d, 0x8b,
0x5f, 0x16, 0x9d, 0x23, 0xd5, 0x82, 0x16, 0x24,
0xf6, 0x87, 0x9d, 0x3d, 0x69, 0x24, 0x21, 0x8a,
0x94, 0xbb, 0xf8, 0xd6, 0x97, 0x25, 0x77, 0x07,
0x73, 0xaa, 0x87, 0xee, 0x4c, 0xef, 0x01, 0xce,
0x6e, 0x29, 0x43, 0x52, 0x49, 0x11, 0x7f, 0x6f
};

// g48 (counter = 1)
static readonly byte[] g48_x = {
0x37, 0x58, 0x8b, 0x4a, 0xca, 0x0f, 0xc5, 0xab,
0xf5, 0x29, 0x04, 0x60, 0x72, 0xe2, 0x33, 0xf7,
0x7c, 0x4b, 0x63, 0xd9, 0x7d, 0x2a, 0x33, 0x80,
0x0c, 0x10, 0x62, 0xc1, 0x19, 0x6c, 0x53, 0x09,
0x8e, 0x11, 0xf6, 0x43, 0xce, 0xc1, 0xc5, 0x4a,
0xba, 0xa6, 0xa9, 0xb2, 0x7d, 0x1d, 0xeb, 0x7b
};
static readonly byte[] g48_y = {
0x0c, 0x19, 0xce, 0xcb, 0x06, 0xa2, 0xdc, 0x7d,
0x0a, 0xe7, 0x6c, 0xe2, 0xc4, 0x50, 0x18, 0x05,
0x25, 0xe3, 0xdd, 0x02, 0xb7, 0x6d, 0x80, 0x97,
0xed, 0x44, 0x0b, 0xc8, 0xd9, 0x41, 0x05, 0xe6,
0xb6, 0xae, 0x57, 0xcc, 0xf3, 0x3c, 0x90, 0x2a,
0x40, 0xc4, 0x5f, 0xbc, 0xff, 0xd6, 0x06, 0x9f
};

// g49 (counter = 1)
static readonly byte[] g49_x = {
0xff, 0xed, 0xb2, 0xa5, 0x73, 0x5c, 0x6e, 0xab,
0x4d, 0x3a, 0x26, 0xab, 0x3f, 0x71, 0x6a, 0xd3,
0x65, 0x2f, 0x1f, 0xa7, 0x04, 0xea, 0x4c, 0x5f,
0x06, 0x4e, 0x09, 0xdf, 0x59, 0xe0, 0x64, 0xfe,
0x9a, 0xab, 0x89, 0xb6, 0x11, 0xd0, 0x52, 0x44,
0x24, 0xb2, 0xec, 0x4d, 0x35, 0x41, 0x35, 0x67
};
static readonly byte[] g49_y = {
0x8a, 0xf9, 0x13, 0xbc, 0xe6, 0x0e, 0x84, 0xea,
0x81, 0xc3, 0xd7, 0xaa, 0x63, 0x7d, 0x40, 0x77,
0xe9, 0xa5, 0xcc, 0x4f, 0xa6, 0xa8, 0x01, 0xda,
0xd9, 0xde, 0xa6, 0x9a, 0x56, 0xca, 0xce, 0x6f,
0xa3, 0x8b, 0x89, 0x1e, 0x3d, 0xfa, 0x07, 0xf2,
0xc6, 0xbf, 0xa4, 0x5c, 0x48, 0x0f, 0x42
};

// g50 (counter = 0)
static readonly byte[] g50_x = {
0xe5, 0x1a, 0x94, 0xf1, 0xbb, 0x05, 0x60, 0x52,
0xf4, 0x68, 0xe2, 0xf2, 0x1b, 0xbc, 0x5a, 0xd2,
0xf6, 0x72, 0xed, 0x3f, 0x83, 0x7d, 0xe0, 0x89,
0xbd, 0x59, 0xa9, 0xc7, 0x5a, 0x7b, 0xfb, 0x97,
0xfb, 0x87, 0x10, 0x3f, 0xff, 0x41, 0x5f, 0xb1,
0x94, 0xee, 0x8b, 0x57, 0xf2, 0xdc, 0xdb, 0x25
};
static readonly byte[] g50_y = {
0x1b, 0x5f, 0x77, 0x0e, 0xbc, 0x9f, 0x67, 0x41,
0xf9, 0x7d, 0x4e, 0x82, 0x7b, 0xfd, 0x6a, 0xcc,
0x59, 0x2a, 0xce, 0x08, 0x10, 0x7d, 0xb2, 0x6f,
0x3e, 0xb4, 0x28, 0xee, 0x18, 0xcd, 0x19, 0x7e,
0x72, 0xb9, 0x49, 0xe6, 0x17, 0x25, 0x45, 0xdf,
0xde, 0xe0, 0x46, 0x45, 0x57, 0x1a, 0x3d, 0xc7
};

// gt (counter = 0)
static readonly byte[] gt_x = {
0xae, 0x14, 0x1e, 0x91, 0x57, 0x8a, 0x26, 0x67,
0xf7, 0xb7, 0x90, 0x61, 0xe0, 0xa0, 0xf5, 0xb9,
0xe4, 0x59, 0xde, 0x30, 0x38, 0xc6, 0x97, 0x75,
0x3d, 0x2f, 0x7e, 0xe1, 0xc0, 0x8a, 0x66, 0x23,
0x16, 0xda, 0x0d, 0x04, 0xc5, 0xd2, 0x11, 0x5c,
0xfc, 0xbe, 0xd0, 0x03, 0xe5, 0x1b, 0x8e, 0x38
};
static readonly byte[] gt_y = {
0x4d, 0x4d, 0x60, 0xf7, 0x66, 0x51, 0x83, 0x48,
0x3c, 0x8b, 0xfb, 0x46, 0x6c, 0x36, 0xbf, 0x1e,
0x14, 0x83, 0x7a, 0x77, 0x53, 0xa0, 0xdd, 0x1d,
0xcc, 0x03, 0x6d, 0x91, 0xaf, 0x53, 0xc1, 0x0c,
0xfe, 0x76, 0x5a, 0xc6, 0x19, 0x08, 0x47, 0xd2,
0xf6, 0x68, 0x3b, 0x78, 0xe1, 0xe0, 0x9f, 0x0c
};

// gd (counter = 2)
static readonly byte[] gd_x = {
0xed, 0x14, 0xac, 0x90, 0x7c, 0xad, 0xce, 0xde,
0xdd, 0xbb, 0x77, 0x2a, 0x2c, 0x20, 0x96, 0x04,
0xfa, 0xd1, 0xb4, 0x3b, 0xd8, 0x5a, 0xcf, 0x2d,
0xf5, 0x0e, 0x84, 0x8b, 0xd9, 0xaa, 0x5b, 0x99,
0xb6, 0x5b, 0x6a, 0x87, 0xdb, 0xea, 0xc9, 0x0f,
0x33, 0x01, 0xe8, 0xc9, 0xd4, 0x5b, 0x03, 0x7f
};
static readonly byte[] gd_y = {
0x60, 0x44, 0x88, 0x78, 0xdc, 0x34, 0xae, 0x65,
0x67, 0x3d, 0x98, 0x2e, 0xf3, 0x00, 0xa5, 0x65,
0xac, 0x0b, 0x46, 0x51, 0x0b, 0xb9, 0x62, 0xf1,
0xa0, 0x0c, 0x45, 0x9d, 0xc9, 0x69, 0xe0, 0x49,
0xad, 0xd2, 0x11, 0x21, 0x06, 0x07, 0xdb, 0x5a,
0x06, 0xf1, 0x02, 0x50, 0x50, 0x67, 0xf5, 0x98
};

            private static readonly ECGroup ecGroup = ECGroup.CreateECGroup(p, a, b, g_x, g_y, n, ParamSet_EC_P384_V1Name, "P-384");

            internal static readonly Group Gq = ecGroup;

            internal static readonly GroupElement Gd = ecGroup.CreateGroupElement(gd_x, gd_y);

            internal static readonly GroupElement[] IssuerGenerators = {
                ecGroup.CreateGroupElement(g1_x, g1_y),
                ecGroup.CreateGroupElement(g2_x, g2_y),
                ecGroup.CreateGroupElement(g3_x, g3_y),
                ecGroup.CreateGroupElement(g4_x, g4_y),
                ecGroup.CreateGroupElement(g5_x, g5_y),
                ecGroup.CreateGroupElement(g6_x, g6_y),
                ecGroup.CreateGroupElement(g7_x, g7_y),
                ecGroup.CreateGroupElement(g8_x, g8_y),
                ecGroup.CreateGroupElement(g9_x, g9_y),
                ecGroup.CreateGroupElement(g10_x, g10_y),
                ecGroup.CreateGroupElement(g11_x, g11_y),
                ecGroup.CreateGroupElement(g12_x, g12_y),
                ecGroup.CreateGroupElement(g13_x, g13_y),
                ecGroup.CreateGroupElement(g14_x, g14_y),
                ecGroup.CreateGroupElement(g15_x, g15_y),
                ecGroup.CreateGroupElement(g16_x, g16_y),
                ecGroup.CreateGroupElement(g17_x, g17_y),
                ecGroup.CreateGroupElement(g18_x, g18_y),
                ecGroup.CreateGroupElement(g19_x, g19_y),
                ecGroup.CreateGroupElement(g20_x, g20_y),
                ecGroup.CreateGroupElement(g21_x, g21_y),
                ecGroup.CreateGroupElement(g22_x, g22_y),
                ecGroup.CreateGroupElement(g23_x, g23_y),
                ecGroup.CreateGroupElement(g24_x, g24_y),
                ecGroup.CreateGroupElement(g25_x, g25_y),
                ecGroup.CreateGroupElement(g26_x, g26_y),
                ecGroup.CreateGroupElement(g27_x, g27_y),
                ecGroup.CreateGroupElement(g28_x, g28_y),
                ecGroup.CreateGroupElement(g29_x, g29_y),
                ecGroup.CreateGroupElement(g30_x, g30_y),
                ecGroup.CreateGroupElement(g31_x, g31_y),
                ecGroup.CreateGroupElement(g32_x, g32_y),
                ecGroup.CreateGroupElement(g33_x, g33_y),
                ecGroup.CreateGroupElement(g34_x, g34_y),
                ecGroup.CreateGroupElement(g35_x, g35_y),
                ecGroup.CreateGroupElement(g36_x, g36_y),
                ecGroup.CreateGroupElement(g37_x, g37_y),
                ecGroup.CreateGroupElement(g38_x, g38_y),
                ecGroup.CreateGroupElement(g39_x, g39_y),
                ecGroup.CreateGroupElement(g40_x, g40_y),
                ecGroup.CreateGroupElement(g41_x, g41_y),
                ecGroup.CreateGroupElement(g42_x, g42_y),
                ecGroup.CreateGroupElement(g43_x, g43_y),
                ecGroup.CreateGroupElement(g44_x, g44_y),
                ecGroup.CreateGroupElement(g45_x, g45_y),
                ecGroup.CreateGroupElement(g46_x, g46_y),
                ecGroup.CreateGroupElement(g47_x, g47_y),
                ecGroup.CreateGroupElement(g48_x, g48_y),
                ecGroup.CreateGroupElement(g49_x, g49_y),
                ecGroup.CreateGroupElement(g50_x, g50_y),
                ecGroup.CreateGroupElement(gt_x, gt_y)
                };
        }

        //
        // P-521
        //
        static class P521
        {

static readonly byte[] p = {
0x01, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff
};

static readonly byte[] a = {
0x01, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xfc
};

static readonly byte[] b = {
0x51, 0x95, 0x3e, 0xb9, 0x61, 0x8e, 0x1c, 0x9a,
0x1f, 0x92, 0x9a, 0x21, 0xa0, 0xb6, 0x85, 0x40,
0xee, 0xa2, 0xda, 0x72, 0x5b, 0x99, 0xb3, 0x15,
0xf3, 0xb8, 0xb4, 0x89, 0x91, 0x8e, 0xf1, 0x09,
0xe1, 0x56, 0x19, 0x39, 0x51, 0xec, 0x7e, 0x93,
0x7b, 0x16, 0x52, 0xc0, 0xbd, 0x3b, 0xb1, 0xbf,
0x07, 0x35, 0x73, 0xdf, 0x88, 0x3d, 0x2c, 0x34,
0xf1, 0xef, 0x45, 0x1f, 0xd4, 0x6b, 0x50, 0x3f,
0x00
};

static readonly byte[] n = {
0x01, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
0xff, 0xfa, 0x51, 0x86, 0x87, 0x83, 0xbf, 0x2f,
0x96, 0x6b, 0x7f, 0xcc, 0x01, 0x48, 0xf7, 0x09,
0xa5, 0xd0, 0x3b, 0xb5, 0xc9, 0xb8, 0x89, 0x9c,
0x47, 0xae, 0xbb, 0x6f, 0xb7, 0x1e, 0x91, 0x38,
0x64, 0x09
};

static readonly byte[] h = {
0x01
};

// g
static readonly byte[] g_x = {
0xc6, 0x85, 0x8e, 0x06, 0xb7, 0x04, 0x04, 0xe9,
0xcd, 0x9e, 0x3e, 0xcb, 0x66, 0x23, 0x95, 0xb4,
0x42, 0x9c, 0x64, 0x81, 0x39, 0x05, 0x3f, 0xb5,
0x21, 0xf8, 0x28, 0xaf, 0x60, 0x6b, 0x4d, 0x3d,
0xba, 0xa1, 0x4b, 0x5e, 0x77, 0xef, 0xe7, 0x59,
0x28, 0xfe, 0x1d, 0xc1, 0x27, 0xa2, 0xff, 0xa8,
0xde, 0x33, 0x48, 0xb3, 0xc1, 0x85, 0x6a, 0x42,
0x9b, 0xf9, 0x7e, 0x7e, 0x31, 0xc2, 0xe5, 0xbd,
0x66
};
static readonly byte[] g_y = {
0x01, 0x18, 0x39, 0x29, 0x6a, 0x78, 0x9a, 0x3b,
0xc0, 0x04, 0x5c, 0x8a, 0x5f, 0xb4, 0x2c, 0x7d,
0x1b, 0xd9, 0x98, 0xf5, 0x44, 0x49, 0x57, 0x9b,
0x44, 0x68, 0x17, 0xaf, 0xbd, 0x17, 0x27, 0x3e,
0x66, 0x2c, 0x97, 0xee, 0x72, 0x99, 0x5e, 0xf4,
0x26, 0x40, 0xc5, 0x50, 0xb9, 0x01, 0x3f, 0xad,
0x07, 0x61, 0x35, 0x3c, 0x70, 0x86, 0xa2, 0x72,
0xc2, 0x40, 0x88, 0xbe, 0x94, 0x76, 0x9f, 0xd1,
0x66, 0x50
};

// g1 (counter = 3)
static readonly byte[] g1_x = {
0x01, 0x67, 0x5b, 0x76, 0xe4, 0xf2, 0x1c, 0xef,
0xb1, 0x1f, 0xfe, 0x81, 0xf3, 0x13, 0x7c, 0x19,
0xe6, 0x99, 0x92, 0xd1, 0x44, 0x03, 0x3f, 0x23,
0xf3, 0x93, 0x0d, 0x1d, 0x5b, 0x01, 0x3c, 0xa8,
0x16, 0x98, 0xa1, 0x55, 0xf2, 0x29, 0x85, 0x00,
0xca, 0xee, 0xe3, 0x6b, 0x13, 0xce, 0xf5, 0x2c,
0xa1, 0xc4, 0x21, 0x28, 0x6a, 0xe5, 0x9f, 0x68,
0x68, 0x4c, 0x85, 0x78, 0x62, 0x8b, 0xa1, 0x27,
0x3b, 0x65
};
static readonly byte[] g1_y = {
0xc6, 0x31, 0xe6, 0x0a, 0xbc, 0xbd, 0x97, 0x01,
0x36, 0x20, 0x1e, 0x5a, 0x8e, 0x43, 0x59, 0xf8,
0xcd, 0x31, 0x69, 0xcc, 0x39, 0x64, 0x56, 0xb9,
0xa1, 0x2d, 0x04, 0x31, 0x7b, 0x3b, 0xee, 0x9a,
0xa2, 0x7a, 0xc8, 0xfd, 0x84, 0xe6, 0x7a, 0x61,
0x93, 0x4c, 0x63, 0xd1, 0xd9, 0x5c, 0xd3, 0xf8,
0x90, 0x03, 0xfd, 0xb5, 0x7b, 0xfc, 0xbb, 0x71,
0xf8, 0x11, 0xfe, 0xb5, 0x96, 0xda, 0xed, 0x7e,
0x3a
};

// g2 (counter = 4)
static readonly byte[] g2_x = {
0x01, 0xe5, 0x65, 0xee, 0x28, 0x01, 0xf3, 0xd8,
0x22, 0x66, 0x47, 0x77, 0xb0, 0x8b, 0x43, 0xf2,
0xe5, 0x0b, 0x60, 0xaa, 0x66, 0x9e, 0xaf, 0x98,
0x70, 0xb3, 0x7f, 0xe8, 0xf7, 0x37, 0x73, 0xb8,
0xf9, 0x2e, 0xbe, 0x0c, 0x05, 0x19, 0xf4, 0x94,
0xbf, 0x93, 0x8d, 0x55, 0x07, 0x2d, 0xbe, 0xa5,
0x93, 0x95, 0x03, 0x6e, 0x5c, 0x0b, 0xa6, 0x8b,
0x88, 0x5a, 0xd2, 0xab, 0xb4, 0x7c, 0x64, 0x02,
0xb4, 0x63
};
static readonly byte[] g2_y = {
0x72, 0x33, 0xdb, 0xf7, 0x46, 0xdd, 0x2c, 0x2b,
0xf6, 0x2a, 0x86, 0x0f, 0x36, 0x45, 0x0c, 0xaf,
0x29, 0x5d, 0x88, 0xd3, 0x95, 0x52, 0x1d, 0x73,
0x56, 0xc9, 0x60, 0xf6, 0x55, 0x78, 0xfc, 0xc8,
0x2c, 0x29, 0xd6, 0xc0, 0x66, 0xf2, 0xc4, 0xa0,
0xf5, 0x48, 0xf8, 0x6c, 0xc1, 0xbc, 0x7e, 0xa1,
0x2e, 0xa7, 0xd3, 0x4c, 0xe2, 0x41, 0xd1, 0x98,
0xf9, 0x54, 0xd9, 0x8c, 0x4f, 0x66, 0x78, 0x78,
0xbd
};

// g3 (counter = 2)
static readonly byte[] g3_x = {
0x01, 0x63, 0xe4, 0x1f, 0x82, 0xb9, 0xd7, 0x56,
0xc7, 0x74, 0xaa, 0x84, 0xa2, 0x08, 0x9e, 0x7d,
0xab, 0x68, 0xa6, 0x44, 0x54, 0x91, 0x2d, 0x90,
0xe3, 0xcf, 0x37, 0xf9, 0x62, 0x3f, 0xbd, 0x7c,
0x74, 0x02, 0x41, 0x7f, 0x66, 0xae, 0x17, 0x51,
0x9d, 0xb4, 0xc6, 0xc6, 0x98, 0xbd, 0x85, 0x3d,
0xbf, 0xea, 0x2f, 0xd3, 0xec, 0x78, 0xef, 0xe8,
0x87, 0xd1, 0xbd, 0x88, 0x7c, 0xdd, 0xbb, 0xa0,
0xbb, 0x06
};
static readonly byte[] g3_y = {
0xb0, 0x81, 0x1b, 0xd8, 0x2a, 0xca, 0x45, 0x9f,
0x96, 0x9c, 0xb5, 0xe4, 0x64, 0x2e, 0x5b, 0x67,
0xb5, 0x2d, 0xfc, 0x96, 0x74, 0xe2, 0xba, 0x3d,
0x24, 0xec, 0xe8, 0x62, 0xfe, 0x79, 0xac, 0x51,
0x3e, 0xf6, 0x5e, 0x8c, 0x9d, 0x3f, 0x73, 0xbb,
0x28, 0xc8, 0x4f, 0x20, 0xd6, 0xb4, 0x57, 0xc1,
0x94, 0x14, 0x36, 0xe0, 0xf2, 0x80, 0xbc, 0xd7,
0x61, 0x23, 0xf7, 0x39, 0x7a, 0xf3, 0xa0, 0xfb,
0x8c
};

// g4 (counter = 0)
static readonly byte[] g4_x = {
0x1b, 0xc6, 0x8e, 0x0c, 0xcc, 0xd3, 0xe2, 0xf0,
0x62, 0x75, 0x89, 0x10, 0x62, 0x46, 0x31, 0x7f,
0xb7, 0x33, 0x46, 0xfe, 0xb9, 0x16, 0xe2, 0xca,
0x61, 0x3c, 0x38, 0x82, 0x54, 0xdd, 0x6f, 0x0f,
0x62, 0x1a, 0x9a, 0x5f, 0x56, 0xca, 0xc2, 0x5c,
0x35, 0x34, 0x90, 0xf4, 0x32, 0xe9, 0xdc, 0xa7,
0x6e, 0xf5, 0xa2, 0xff, 0xb1, 0x85, 0x1d, 0xf8,
0x92, 0x08, 0x2c, 0x6d, 0x58, 0x3f, 0x4e, 0xaa,
0x31
};
static readonly byte[] g4_y = {
0x42, 0x5a, 0xbc, 0x21, 0xdb, 0x65, 0x8c, 0xb7,
0xd3, 0x32, 0x5d, 0x74, 0x90, 0xd5, 0x37, 0x25,
0x58, 0xf7, 0x62, 0x00, 0x13, 0x69, 0x25, 0x0b,
0x80, 0x86, 0xe4, 0xd4, 0x97, 0x07, 0x06, 0xeb,
0x4b, 0x51, 0xbf, 0x76, 0xc6, 0x43, 0xaa, 0x57,
0xbf, 0xa3, 0xce, 0x5f, 0x6b, 0x78, 0x31, 0xe8,
0xc1, 0x67, 0xff, 0x29, 0xe2, 0xc6, 0x95, 0x8b,
0x52, 0x86, 0xbd, 0xe4, 0xf5, 0xf3, 0x20, 0xeb,
0xeb
};

// g5 (counter = 0)
static readonly byte[] g5_x = {
0x65, 0x11, 0x0b, 0x80, 0xb2, 0xf3, 0xba, 0x19,
0xab, 0x7f, 0x5a, 0x6f, 0xbf, 0xf5, 0x63, 0x3b,
0xb6, 0xb3, 0x14, 0x85, 0x63, 0xbd, 0x1a, 0x29,
0xac, 0x1b, 0xb9, 0xd7, 0x48, 0x80, 0x0d, 0x4f,
0x28, 0xf0, 0x4d, 0xee, 0x10, 0xe8, 0x8f, 0x86,
0xaf, 0xbe, 0xab, 0x7e, 0x10, 0xfc, 0xcf, 0xf6,
0x99, 0x03, 0xd3, 0x3f, 0x7c, 0xbe, 0xf6, 0xfd,
0xdf, 0x65, 0x15, 0x96, 0x78, 0x2c, 0xd7, 0x6c,
0x57
};
static readonly byte[] g5_y = {
0xef, 0xa9, 0xbb, 0x04, 0xc1, 0xda, 0x53, 0xc5,
0xd9, 0xa7, 0xe9, 0xa2, 0x6d, 0x33, 0xb6, 0x41,
0x61, 0x3c, 0xd3, 0x74, 0x28, 0x47, 0x5f, 0xdc,
0xf8, 0x21, 0x94, 0x8c, 0x45, 0x72, 0xdb, 0x44,
0x4f, 0x99, 0x7f, 0x90, 0x76, 0x81, 0xf0, 0xc4,
0x66, 0x2d, 0x8b, 0x08, 0xb5, 0x2d, 0x04, 0xa1,
0x09, 0x21, 0x54, 0x36, 0xb0, 0xa3, 0xe4, 0x8f,
0x00, 0x96, 0x3b, 0x99, 0xc4, 0x90, 0xe8, 0xfa,
0x39
};

// g6 (counter = 0)
static readonly byte[] g6_x = {
0x01, 0xf9, 0xff, 0x75, 0xfe, 0x4c, 0xa3, 0xf2,
0xb6, 0x5f, 0x66, 0x85, 0xa2, 0x3f, 0x62, 0x01,
0xed, 0x46, 0x0e, 0x59, 0xd5, 0x17, 0x29, 0x04,
0x05, 0xaf, 0xc7, 0xf7, 0x97, 0xaf, 0x27, 0x2a,
0x06, 0xf4, 0x50, 0x97, 0x17, 0xd1, 0x89, 0x00,
0x5c, 0xe9, 0x3b, 0xe3, 0x32, 0x13, 0x21, 0x84,
0xed, 0x4f, 0xf6, 0x67, 0x97, 0x66, 0xe6, 0x17,
0x26, 0x7e, 0x9d, 0x77, 0x01, 0x74, 0xe9, 0x39,
0x8b, 0x02
};
static readonly byte[] g6_y = {
0xe8, 0x7e, 0xbc, 0x53, 0x59, 0x51, 0xba, 0x18,
0xb9, 0xde, 0xff, 0xd8, 0x08, 0xe9, 0x7b, 0x33,
0x7c, 0xc4, 0xec, 0x0b, 0x29, 0x44, 0xae, 0xaa,
0xec, 0xd3, 0x49, 0x34, 0x51, 0x0f, 0xee, 0x96,
0x06, 0xec, 0x6f, 0xf2, 0x8b, 0x4c, 0x49, 0xcc,
0x0b, 0xb6, 0xe8, 0xc5, 0x0c, 0x2a, 0xbd, 0xc9,
0x52, 0xae, 0x95, 0xf6, 0x84, 0x02, 0x7b, 0x15,
0xa1, 0x0d, 0x64, 0xe7, 0x6c, 0xc6, 0x14, 0xf5,
0xe5
};

// g7 (counter = 4)
static readonly byte[] g7_x = {
0xca, 0x01, 0x2f, 0x9a, 0x52, 0xa6, 0x8e, 0x79,
0xad, 0x2e, 0x8e, 0x23, 0x82, 0x89, 0xb5, 0x2f,
0x7d, 0xb5, 0xc3, 0x7b, 0x7e, 0x50, 0xf1, 0xff,
0x00, 0xd2, 0xe4, 0x12, 0x2f, 0xe4, 0xb9, 0x29,
0xd6, 0x17, 0x5a, 0x95, 0xd7, 0xf2, 0x4c, 0x9e,
0x22, 0x06, 0x9d, 0x49, 0xeb, 0x9d, 0xe2, 0xa1,
0x1b, 0xd3, 0x83, 0x2f, 0xd9, 0xd4, 0x6d, 0xa0,
0xed, 0xd0, 0xdb, 0xa0, 0x88, 0x6f, 0xb3, 0x30,
0xa5
};
static readonly byte[] g7_y = {
0x25, 0x5a, 0xd2, 0xcf, 0x14, 0x40, 0xe3, 0x12,
0x10, 0xf3, 0x8f, 0xa7, 0x0f, 0xfc, 0xdc, 0xf6,
0x32, 0xe5, 0x3f, 0xb8, 0x07, 0x2f, 0xb0, 0xea,
0x5e, 0xf1, 0x1b, 0x6c, 0x59, 0x0e, 0x0a, 0xa8,
0x16, 0x16, 0x3b, 0xa2, 0x88, 0x14, 0x86, 0x54,
0xbf, 0xb5, 0x4b, 0x71, 0xa4, 0xe3, 0x87, 0xd3,
0xc6, 0x3e, 0x8e, 0x4c, 0x12, 0xf5, 0x47, 0x9d,
0x45, 0x29, 0xf9, 0xfc, 0x0b, 0x2c, 0x93, 0x59,
0xf6
};

// g8 (counter = 0)
static readonly byte[] g8_x = {
0x1f, 0xfa, 0x43, 0x8f, 0xf0, 0x48, 0xba, 0x53,
0x9c, 0xeb, 0x37, 0xe2, 0x33, 0x4e, 0xca, 0x04,
0xc8, 0x2a, 0xf7, 0xe9, 0xbe, 0x21, 0x0c, 0xe5,
0x15, 0x72, 0xdb, 0x53, 0xcb, 0x41, 0xe9, 0xb3,
0x26, 0xb9, 0xa4, 0x14, 0x8a, 0x24, 0x99, 0x1e,
0x66, 0x08, 0xfd, 0x75, 0x3b, 0x2f, 0x34, 0xad,
0xfe, 0x13, 0xab, 0x87, 0xe6, 0x2d, 0xc8, 0xf0,
0x77, 0xe5, 0xba, 0xef, 0x43, 0x7f, 0x58, 0xf6,
0x0d
};
static readonly byte[] g8_y = {
0x9f, 0x70, 0xff, 0x23, 0xa4, 0x66, 0x18, 0xa3,
0x4b, 0xa0, 0x00, 0xe7, 0x90, 0x54, 0x4b, 0xa1,
0xbe, 0xbb, 0x18, 0xf9, 0xf9, 0x97, 0x4f, 0x7e,
0x1e, 0x08, 0x44, 0xa5, 0x5b, 0x9a, 0x27, 0x01,
0x83, 0xa1, 0xb6, 0xbb, 0x90, 0x5b, 0xa8, 0x76,
0x81, 0xf9, 0xaa, 0x63, 0xcb, 0x39, 0xb5, 0x7f,
0xaa, 0xc9, 0xbb, 0x93, 0x3e, 0x95, 0x26, 0xb8,
0x91, 0xb2, 0x10, 0x99, 0x91, 0xc2, 0x59, 0xfb,
0xc2
};

// g9 (counter = 0)
static readonly byte[] g9_x = {
0x30, 0x1c, 0xb6, 0x07, 0x5d, 0x05, 0x6b, 0xc6,
0x5e, 0x6a, 0xe3, 0x01, 0xc0, 0x76, 0x75, 0x2e,
0x07, 0xef, 0x2d, 0x21, 0xf1, 0x82, 0xce, 0xde,
0x42, 0xab, 0x6f, 0x56, 0xf0, 0x89, 0xdb, 0x5d,
0x50, 0x66, 0x0f, 0x91, 0x1a, 0x7f, 0x85, 0x79,
0x73, 0x37, 0xa1, 0xa9, 0x50, 0x56, 0xd4, 0x61,
0x3b, 0xa7, 0x5c, 0x7d, 0x69, 0xa5, 0xbc, 0xba,
0xec, 0x23, 0x62, 0xe9, 0x4c, 0x65, 0x96, 0x45,
0x65
};
static readonly byte[] g9_y = {
0x2d, 0xee, 0x2d, 0xcd, 0x02, 0xae, 0x5e, 0x3d,
0x50, 0x34, 0x0b, 0x38, 0x86, 0xdb, 0x25, 0xc0,
0x53, 0x6e, 0xad, 0xeb, 0x28, 0x7d, 0x22, 0xab,
0x85, 0x24, 0xfd, 0xf3, 0x2d, 0x40, 0xd2, 0x01,
0x9a, 0x92, 0x02, 0xf9, 0xda, 0x8f, 0xa0, 0x79,
0x27, 0x52, 0xda, 0xfa, 0x23, 0xa0, 0x42, 0xc0,
0x84, 0xf4, 0xd0, 0x8c, 0xe0, 0x79, 0x2f, 0x46,
0x00, 0x3b, 0xf4, 0x77, 0xf0, 0xa8, 0x6c, 0xac,
0x85
};

// g10 (counter = 0)
static readonly byte[] g10_x = {
0x40, 0x7f, 0xc8, 0xe9, 0x84, 0xc8, 0x46, 0x01,
0xf3, 0xe4, 0x19, 0x52, 0xfe, 0xc1, 0x61, 0x42,
0x30, 0x17, 0xe0, 0x73, 0xd3, 0x37, 0x83, 0xff,
0x05, 0xa5, 0x0e, 0x0a, 0x11, 0xb9, 0x4c, 0x4e,
0xbe, 0xd0, 0xc8, 0xd7, 0x67, 0x1e, 0x51, 0x80,
0xf9, 0x93, 0xb1, 0xe9, 0xd1, 0xe3, 0x57, 0xd2,
0x91, 0x77, 0x27, 0x1e, 0xb2, 0x9e, 0xfb, 0x96,
0xcd, 0xef, 0x4e, 0xdf, 0x9d, 0x8c, 0x6b, 0x74,
0x40
};
static readonly byte[] g10_y = {
0x43, 0x1d, 0xf6, 0x9b, 0x02, 0x20, 0x04, 0xca,
0xf4, 0xc7, 0x1e, 0xce, 0x51, 0x69, 0x32, 0x02,
0xb5, 0x10, 0xf4, 0x02, 0x07, 0xe4, 0x85, 0x6e,
0x94, 0x09, 0x82, 0x15, 0xc0, 0x1f, 0x55, 0xc1,
0x9a, 0x07, 0x28, 0xc7, 0x48, 0x1c, 0xdf, 0xc2,
0x4b, 0x2f, 0x80, 0xd1, 0xd2, 0x29, 0xec, 0xda,
0xc1, 0xf4, 0xd8, 0x3e, 0x8c, 0x46, 0x6c, 0xd3,
0x58, 0x6f, 0x4e, 0x0d, 0x13, 0xff, 0x25, 0xe9,
0x88
};

// g11 (counter = 3)
static readonly byte[] g11_x = {
0x01, 0xdb, 0xac, 0x6b, 0xe0, 0x86, 0x07, 0x87,
0xde, 0xf0, 0x3f, 0xfc, 0x54, 0x72, 0xdc, 0x23,
0xfa, 0x06, 0xb1, 0x5a, 0xf0, 0x1c, 0xbf, 0x8b,
0xee, 0x8f, 0x5a, 0x0e, 0x71, 0xae, 0x28, 0xb5,
0xec, 0xac, 0x96, 0xfd, 0xa8, 0x2e, 0x85, 0x74,
0x74, 0x1c, 0x21, 0x9f, 0xa6, 0x2d, 0x31, 0xaf,
0xd6, 0x58, 0xbb, 0x95, 0x78, 0xd5, 0x9b, 0x27,
0x9d, 0xf0, 0x94, 0x90, 0x63, 0x12, 0x26, 0x65,
0xcf, 0xb3
};
static readonly byte[] g11_y = {
0x36, 0xda, 0x1a, 0x1f, 0xbf, 0xa6, 0x5c, 0x4e,
0xe9, 0x1d, 0x83, 0x6c, 0x2f, 0x42, 0x4a, 0xb2,
0x84, 0x5c, 0xd6, 0xeb, 0x79, 0x62, 0x50, 0x06,
0x65, 0xf7, 0x45, 0xc7, 0xa8, 0x3c, 0x4d, 0xa9,
0x9e, 0x07, 0xcf, 0x2f, 0x57, 0xd9, 0x9e, 0x57,
0x69, 0xbf, 0xa7, 0x0b, 0xdb, 0xde, 0x84, 0xed,
0xe5, 0x62, 0x50, 0x4b, 0x8e, 0x84, 0xb2, 0xc1,
0xe1, 0x59, 0xf0, 0xdb, 0xb9, 0xeb, 0x58, 0x54,
0x9b
};

// g12 (counter = 0)
static readonly byte[] g12_x = {
0xed, 0xa6, 0x25, 0xde, 0x05, 0x83, 0x40, 0x21,
0x3a, 0x85, 0x9d, 0xf9, 0x74, 0x81, 0xc0, 0x62,
0x23, 0x33, 0x7a, 0xb4, 0x2e, 0x45, 0x59, 0x9b,
0xb5, 0x43, 0x19, 0x60, 0xd3, 0xb9, 0xc4, 0x33,
0xd8, 0x7e, 0x6d, 0xf0, 0x1b, 0xdc, 0x89, 0x5a,
0x3c, 0x10, 0x0c, 0xf5, 0xbd, 0x35, 0x7e, 0x42,
0xd4, 0xcd, 0xbb, 0x9e, 0xeb, 0x47, 0xe1, 0x92,
0x98, 0x02, 0xc2, 0x0e, 0xfc, 0x65, 0xe9, 0xe0,
0x44
};
static readonly byte[] g12_y = {
0xc9, 0x0a, 0x0e, 0xfa, 0x16, 0xc0, 0xbb, 0x69,
0x8e, 0x2b, 0x67, 0xb5, 0xbd, 0x49, 0x8a, 0x02,
0xd8, 0x69, 0x5c, 0x73, 0xf2, 0x62, 0x6a, 0x96,
0xdd, 0x85, 0x37, 0x1d, 0xc8, 0xd2, 0x36, 0xaf,
0x6f, 0x23, 0x42, 0x93, 0xd0, 0x96, 0x22, 0xaa,
0x10, 0x58, 0x85, 0x9b, 0x36, 0x96, 0x1f, 0xd5,
0x9f, 0x5e, 0x28, 0xcb, 0x48, 0x0c, 0x20, 0x26,
0x24, 0xbe, 0xf4, 0x31, 0x86, 0xfc, 0xcd, 0x39,
0x6e
};

// g13 (counter = 0)
static readonly byte[] g13_x = {
0x05, 0xf8, 0xa4, 0x99, 0xb9, 0x3c, 0x10, 0xc4,
0x9e, 0x1e, 0x29, 0xac, 0x7b, 0xa4, 0x26, 0x71,
0xfb, 0x46, 0xae, 0x8d, 0x03, 0x20, 0xba, 0xcb,
0x5f, 0x01, 0x8f, 0x45, 0x0a, 0x58, 0x77, 0xc0,
0xc6, 0x36, 0x56, 0xee, 0xd3, 0xc8, 0x04, 0x3b,
0xfe, 0x12, 0xc9, 0x34, 0x6a, 0x6a, 0xe1, 0xf8,
0x1d, 0xdb, 0x90, 0xf9, 0x6a, 0x0f, 0xae, 0xf9,
0xde, 0xf3, 0xa9, 0x86, 0xbe, 0xcc, 0xa8, 0x56,
0xfd
};
static readonly byte[] g13_y = {
0x63, 0x49, 0x5d, 0x5e, 0xca, 0x21, 0xf2, 0x70,
0xe3, 0xd5, 0x99, 0xfe, 0xf3, 0x49, 0x22, 0x30,
0xec, 0xa6, 0x4c, 0x71, 0x2b, 0xb3, 0x10, 0x0b,
0x14, 0xf4, 0xae, 0x2b, 0x62, 0x18, 0xab, 0x68,
0x35, 0x55, 0x78, 0xca, 0x73, 0xa2, 0x6a, 0xf4,
0xc8, 0x02, 0x02, 0x1f, 0x3a, 0x29, 0x9c, 0x84,
0xd3, 0x43, 0xaa, 0xef, 0x23, 0x7c, 0x1a, 0xd2,
0x05, 0x03, 0x53, 0x2a, 0x46, 0x8d, 0xbe, 0xf9,
0xa5
};

// g14 (counter = 1)
static readonly byte[] g14_x = {
0xaa, 0x86, 0xfc, 0xd1, 0x1c, 0xb4, 0xe5, 0x15,
0x60, 0x21, 0xf5, 0x11, 0xc2, 0x14, 0x5d, 0xf1,
0x98, 0xd8, 0xfc, 0x07, 0xe4, 0x37, 0xf7, 0x49,
0x56, 0x04, 0xef, 0xd3, 0x87, 0x10, 0x0d, 0x6a,
0xc3, 0x38, 0x93, 0x4e, 0x43, 0x07, 0x37, 0x6f,
0xeb, 0x1b, 0x9c, 0xd2, 0x31, 0xf7, 0xe7, 0x4e,
0x1f, 0xf7, 0x68, 0x16, 0x51, 0x89, 0x44, 0x2d,
0x6a, 0x85, 0x18, 0x58, 0xdb, 0x44, 0xc3, 0x94,
0x2f
};
static readonly byte[] g14_y = {
0x46, 0xe3, 0x45, 0x90, 0x6c, 0x1f, 0xa9, 0x61,
0xa7, 0x0c, 0xec, 0x67, 0x1f, 0xe9, 0x84, 0xad,
0x62, 0x92, 0x6e, 0x26, 0xaf, 0x0e, 0x49, 0xdb,
0xaf, 0x78, 0x08, 0x1b, 0xd4, 0xc8, 0x0a, 0xf1,
0x42, 0x6c, 0xb3, 0x97, 0xe9, 0xd4, 0x14, 0xb7,
0xc2, 0xe8, 0x30, 0x05, 0xbe, 0x4a, 0xb1, 0x57,
0x1f, 0x28, 0x43, 0xac, 0xd4, 0x83, 0x4d, 0x16,
0x0b, 0x38, 0xf4, 0x70, 0x4c, 0x1c, 0x2d, 0x1a,
0x37
};

// g15 (counter = 3)
static readonly byte[] g15_x = {
0x60, 0xba, 0x95, 0x38, 0x7d, 0x52, 0xd6, 0xf4,
0x22, 0x5f, 0xf4, 0xc2, 0x40, 0xc4, 0xeb, 0x06,
0x1d, 0x6d, 0xa6, 0x01, 0x94, 0x67, 0xf4, 0xa8,
0x45, 0x84, 0xfa, 0xe8, 0xe6, 0x85, 0xea, 0xec,
0x07, 0x8f, 0x00, 0x64, 0xc6, 0xfc, 0x0b, 0x5b,
0x5e, 0x47, 0x9b, 0xda, 0x8b, 0x19, 0x50, 0x8f,
0x64, 0xaa, 0x6f, 0x6b, 0xd3, 0x21, 0xba, 0x3b,
0xb5, 0x59, 0x9b, 0x1f, 0xda, 0x0b, 0xcf, 0x0f,
0x42
};
static readonly byte[] g15_y = {
0xd0, 0xfa, 0xd9, 0x20, 0x1e, 0x63, 0x4b, 0x41,
0xf0, 0xb9, 0xfc, 0x75, 0x97, 0xb7, 0x7b, 0x2f,
0xe4, 0x87, 0xe1, 0x2b, 0x83, 0x36, 0x9f, 0xe7,
0xff, 0x0b, 0x65, 0xab, 0xdd, 0xac, 0x92, 0xa0,
0x2d, 0xeb, 0x1a, 0xc5, 0x4c, 0x8d, 0x6e, 0x0f,
0x5e, 0x55, 0x25, 0x84, 0x30, 0xff, 0x61, 0x8f,
0xd1, 0xcb, 0xf0, 0x39, 0x49, 0x0a, 0x03, 0xe2,
0x6e, 0x61, 0x10, 0x4d, 0x80, 0x9d, 0xe4, 0x1e,
0x45
};

// g16 (counter = 4)
static readonly byte[] g16_x = {
0x01, 0xec, 0xae, 0x88, 0x27, 0xa7, 0xb0, 0x94,
0x7d, 0xaa, 0x18, 0x5e, 0x92, 0x90, 0x0e, 0x5c,
0x5e, 0xf2, 0x5d, 0xec, 0x1f, 0x08, 0x21, 0xb8,
0xa6, 0x97, 0x4b, 0xda, 0x81, 0x92, 0xc0, 0x7b,
0x3a, 0x3c, 0x67, 0x69, 0xad, 0xfd, 0xb8, 0x57,
0xc7, 0x3c, 0x96, 0x74, 0xea, 0xd2, 0x6a, 0x3b,
0x57, 0xa5, 0x6f, 0xa6, 0xc1, 0x5a, 0x47, 0x32,
0xc4, 0xbd, 0x63, 0x06, 0x1e, 0x1c, 0xb8, 0xd7,
0x2e, 0xa2
};
static readonly byte[] g16_y = {
0x75, 0xd0, 0x1f, 0x5d, 0x3b, 0x38, 0x8a, 0x16,
0x54, 0x6b, 0xc8, 0x3e, 0xfa, 0x7f, 0x97, 0x43,
0x2f, 0xd0, 0x07, 0x1e, 0xc7, 0xe5, 0x92, 0x93,
0x23, 0x2b, 0xe9, 0xc6, 0xaa, 0x48, 0xd1, 0xae,
0x5f, 0xa7, 0x4b, 0xa0, 0xdd, 0x8b, 0x99, 0xeb,
0x84, 0x09, 0xb5, 0x4d, 0x6d, 0x81, 0x1d, 0x16,
0x4a, 0xf4, 0x30, 0x85, 0xef, 0x86, 0x0a, 0x3a,
0x27, 0x77, 0x7c, 0x03, 0x36, 0x24, 0xa8, 0xf6,
0x69
};

// g17 (counter = 0)
static readonly byte[] g17_x = {
0xb4, 0x1a, 0x80, 0xd0, 0xcd, 0xae, 0x1d, 0x86,
0xbf, 0x03, 0x21, 0xa9, 0xad, 0xe6, 0xbf, 0x3c,
0xc1, 0xd6, 0x0b, 0xbc, 0x32, 0xa8, 0xcd, 0x8a,
0xd2, 0x32, 0xc8, 0x22, 0x8e, 0x9c, 0x15, 0xe7,
0x5c, 0xc3, 0x72, 0xd6, 0xca, 0x6e, 0x5a, 0xe7,
0x0d, 0xe5, 0xfe, 0x52, 0x02, 0x88, 0x39, 0x38,
0xb4, 0x2d, 0xf2, 0x23, 0x66, 0x43, 0x56, 0x9c,
0x0e, 0xe6, 0xa7, 0x3e, 0x09, 0x64, 0x54, 0x8c,
0x1d
};
static readonly byte[] g17_y = {
0x44, 0xbb, 0x4e, 0xa0, 0x65, 0x9a, 0x27, 0xe6,
0x12, 0xac, 0x00, 0x88, 0x79, 0x49, 0x40, 0x0f,
0xb8, 0xf4, 0x1c, 0x9e, 0xb5, 0xda, 0xff, 0x89,
0xb5, 0xee, 0x4c, 0xe9, 0x79, 0xfa, 0x1d, 0x73,
0xb5, 0xea, 0xda, 0x1f, 0x60, 0x37, 0xc8, 0xe0,
0x16, 0xe4, 0x4c, 0x1c, 0x8a, 0x4f, 0x9d, 0xc2,
0xaf, 0x48, 0x7c, 0xd9, 0x2a, 0x19, 0xf4, 0x50,
0xbd, 0x71, 0xbb, 0x4d, 0x72, 0x3e, 0x9d, 0x04,
0xa9
};

// g18 (counter = 0)
static readonly byte[] g18_x = {
0xd7, 0xea, 0x7e, 0x80, 0x45, 0x29, 0xf3, 0xa0,
0x88, 0x9f, 0xec, 0xa6, 0x32, 0xcc, 0x69, 0x07,
0xa2, 0x43, 0x89, 0xbc, 0x0b, 0xa1, 0xee, 0x0f,
0x64, 0x8d, 0x52, 0xe0, 0xc0, 0xd9, 0x17, 0x44,
0x35, 0x9c, 0xa9, 0x45, 0xe0, 0x29, 0x42, 0x04,
0xa9, 0xd8, 0xa9, 0x05, 0x6c, 0xfc, 0x16, 0xa0,
0x13, 0x84, 0xf4, 0x3c, 0x7b, 0x0f, 0xe2, 0x07,
0xf3, 0x7a, 0x81, 0xad, 0x15, 0x84, 0x55, 0x83,
0x1b
};
static readonly byte[] g18_y = {
0x85, 0xfa, 0x66, 0xee, 0xde, 0x6d, 0x45, 0x2d,
0x4c, 0x80, 0xfb, 0x92, 0x43, 0xbb, 0xbc, 0xd5,
0x50, 0x0d, 0x7c, 0x37, 0xa8, 0x7a, 0xc6, 0x87,
0xa4, 0x9c, 0xa0, 0x20, 0xbb, 0xf0, 0x00, 0x15,
0xac, 0xd5, 0xd4, 0x27, 0xb2, 0xe2, 0x0e, 0x73,
0x3a, 0x05, 0x84, 0x9c, 0xc1, 0x3d, 0x2c, 0x20,
0x45, 0xc5, 0x33, 0x3d, 0x7e, 0xc1, 0x0c, 0xc8,
0xf2, 0xa6, 0x3f, 0x63, 0xfe, 0x78, 0x1c, 0xbd,
0x42
};

// g19 (counter = 2)
static readonly byte[] g19_x = {
0x6c, 0xe8, 0x7f, 0x9f, 0x5b, 0x5b, 0x6f, 0xa0,
0xfb, 0xc5, 0x1c, 0x2e, 0x3a, 0x20, 0x11, 0x53,
0x0b, 0xf2, 0x16, 0xcc, 0x03, 0xcd, 0xea, 0x0c,
0x35, 0x27, 0x33, 0xa0, 0xda, 0xb6, 0xef, 0xa7,
0xff, 0x77, 0xef, 0xb2, 0x16, 0x98, 0xd6, 0xa2,
0xa3, 0x06, 0x97, 0xe3, 0x5b, 0x12, 0x8c, 0x6e,
0xe2, 0xa0, 0x2e, 0xbc, 0x94, 0x6b, 0x7d, 0x01,
0xd6, 0xae, 0xf9, 0xc4, 0x38, 0x1c, 0x12, 0x92,
0x5f
};
static readonly byte[] g19_y = {
0xd6, 0x03, 0x10, 0x16, 0x60, 0x32, 0x50, 0x13,
0x5e, 0x19, 0xf6, 0xbb, 0x80, 0x86, 0xb7, 0x51,
0x9d, 0x22, 0x3b, 0x08, 0x3e, 0xb8, 0x33, 0xf4,
0xee, 0xdb, 0x01, 0x8c, 0x5a, 0x82, 0x6c, 0xe2,
0x09, 0x29, 0x25, 0xb6, 0x8f, 0x8b, 0x3c, 0x82,
0x57, 0x68, 0xb5, 0x87, 0xe7, 0x40, 0x0c, 0x57,
0xbe, 0x2c, 0xbd, 0x5a, 0xf2, 0xda, 0x4d, 0x9a,
0xdb, 0xfd, 0xae, 0x09, 0x64, 0x1f, 0x3e, 0xb6,
0xc2
};

// g20 (counter = 1)
static readonly byte[] g20_x = {
0x01, 0x07, 0x4f, 0x17, 0x16, 0xca, 0x1d, 0xf0,
0x0f, 0x49, 0x49, 0x0c, 0x85, 0xf9, 0xc4, 0xbe,
0x23, 0x8a, 0xe3, 0x52, 0x9e, 0xc6, 0xb2, 0xc4,
0x9b, 0xad, 0xcf, 0x70, 0x2f, 0x90, 0xb4, 0xbc,
0xcf, 0x91, 0x51, 0x46, 0xfb, 0xcf, 0xa0, 0x7c,
0xf2, 0x04, 0x98, 0x67, 0x7e, 0x82, 0xf4, 0x75,
0x72, 0xbe, 0xd9, 0xe2, 0xbb, 0x58, 0x54, 0x68,
0xa9, 0x64, 0x7d, 0xf1, 0xaa, 0x65, 0xb1, 0x0b,
0x4b, 0xee
};
static readonly byte[] g20_y = {
0xe3, 0x73, 0x5d, 0x88, 0x1a, 0x47, 0x13, 0x98,
0xc8, 0x55, 0x4e, 0xcf, 0x7d, 0x25, 0xf5, 0xc8,
0x71, 0x90, 0xc6, 0xde, 0x1a, 0x04, 0xb5, 0xc1,
0x3e, 0xd1, 0x68, 0xe2, 0x37, 0x86, 0x91, 0x26,
0x6f, 0xb9, 0x8f, 0x5b, 0xd5, 0xf3, 0x36, 0x0c,
0xb6, 0x69, 0xcd, 0x9a, 0x4d, 0xb3, 0x2f, 0x42,
0xf6, 0x02, 0x51, 0x5a, 0x74, 0x2d, 0xfe, 0xc7,
0xad, 0xa8, 0x42, 0xd5, 0x95, 0xda, 0x6c, 0x6e,
0x3d
};

// g21 (counter = 1)
static readonly byte[] g21_x = {
0x4e, 0x0a, 0xa3, 0xb9, 0xc2, 0xe7, 0x14, 0x73,
0x86, 0xfa, 0x7b, 0xed, 0x54, 0x3e, 0xbd, 0xd4,
0x15, 0xe8, 0xc3, 0x32, 0x13, 0xc1, 0x5e, 0x28,
0xaa, 0x8a, 0x00, 0x3e, 0x5d, 0xb3, 0x78, 0xe8,
0x0c, 0xb1, 0xda, 0xee, 0x60, 0x02, 0xa5, 0x14,
0xa6, 0x73, 0x9c, 0x8f, 0x61, 0x8d, 0x71, 0x10,
0x4d, 0xbc, 0xc9, 0xd1, 0x6c, 0x71, 0x91, 0xee,
0x65, 0xc9, 0x67, 0xcc, 0x30, 0x68, 0xe7, 0x1f,
0x24
};
static readonly byte[] g21_y = {
0x08, 0x43, 0x7e, 0x93, 0xbf, 0x86, 0xdb, 0xd0,
0xea, 0xba, 0x38, 0x3e, 0xf3, 0x07, 0xdd, 0xb9,
0x8d, 0x27, 0xae, 0x63, 0xe7, 0x90, 0x33, 0x8e,
0x85, 0xb2, 0x78, 0x16, 0x80, 0xab, 0x0b, 0xbf,
0xe1, 0x71, 0x24, 0x78, 0x56, 0x8e, 0x12, 0xac,
0xed, 0xb9, 0x76, 0xa8, 0xf3, 0x8a, 0x25, 0x5c,
0x87, 0x0a, 0xc5, 0xf9, 0x7f, 0x45, 0x14, 0x1c,
0x09, 0x77, 0x46, 0x37, 0x70, 0x47, 0x92, 0x71,
0x65
};

// g22 (counter = 2)
static readonly byte[] g22_x = {
0x01, 0x26, 0xb1, 0xe6, 0x05, 0xe3, 0xba, 0xe4,
0x56, 0xc9, 0x86, 0xad, 0xdb, 0x6e, 0xba, 0x0f,
0x75, 0x02, 0x29, 0xb3, 0xf9, 0x71, 0x23, 0x5e,
0xa8, 0x27, 0xa2, 0xe1, 0x6f, 0x45, 0xf3, 0xa9,
0x38, 0x76, 0x5a, 0xf0, 0xa0, 0x81, 0x2f, 0xa2,
0xaf, 0x01, 0x99, 0xe2, 0xe1, 0x80, 0x16, 0x79,
0xe4, 0x3a, 0x1e, 0xf9, 0x42, 0x17, 0x32, 0x59,
0x5d, 0xba, 0x7e, 0x71, 0xd9, 0xf6, 0xd6, 0x8c,
0x30, 0xcf
};
static readonly byte[] g22_y = {
0xfb, 0x86, 0x40, 0xe4, 0x9a, 0x5f, 0xa0, 0x9c,
0x4f, 0xcd, 0xf2, 0xc3, 0x76, 0x11, 0xfc, 0x39,
0x7f, 0x5f, 0x7b, 0x59, 0x91, 0x70, 0x9c, 0xc6,
0x78, 0x3b, 0x92, 0x88, 0x5e, 0xe9, 0x79, 0x84,
0x74, 0x44, 0x51, 0xb5, 0x33, 0xd5, 0x71, 0x0c,
0x8b, 0x80, 0x1d, 0x5e, 0x29, 0xce, 0x2e, 0x37,
0x3f, 0xc3, 0x25, 0xf3, 0x29, 0xa7, 0x50, 0x2d,
0x50, 0x3f, 0x42, 0x3e, 0xb2, 0xa3, 0xa9, 0xd4,
0x28
};

// g23 (counter = 0)
static readonly byte[] g23_x = {
0x80, 0x94, 0xe9, 0x4e, 0x19, 0x83, 0xa7, 0x61,
0x1a, 0x0c, 0xd6, 0x24, 0xd9, 0x4c, 0xfa, 0x1b,
0x3c, 0x1e, 0x18, 0xeb, 0xd2, 0x76, 0x7f, 0x4f,
0x41, 0x46, 0xc5, 0x73, 0xdf, 0x36, 0x82, 0x81,
0x69, 0x3d, 0xdf, 0xa5, 0x26, 0x21, 0x7a, 0x0d,
0x93, 0x5d, 0x22, 0x56, 0xa5, 0x26, 0xc8, 0x0f,
0xe5, 0x41, 0x80, 0xd6, 0x3b, 0x44, 0xc5, 0xa3,
0xc4, 0x4a, 0x52, 0x8b, 0xb2, 0x11, 0xd0, 0xd4,
0x0b
};
static readonly byte[] g23_y = {
0x2c, 0x57, 0xee, 0xff, 0xc5, 0x12, 0xd8, 0x30,
0xef, 0xf9, 0x0f, 0x8a, 0xa4, 0x1f, 0x86, 0x71,
0x57, 0x92, 0x6a, 0xe5, 0x37, 0x27, 0xb3, 0x65,
0x1d, 0x29, 0x6a, 0x67, 0x03, 0x87, 0x8f, 0x96,
0x73, 0x8f, 0xb3, 0x08, 0xf8, 0x5c, 0xb1, 0x53,
0x0b, 0x39, 0xed, 0x24, 0xcc, 0xb6, 0x0e, 0xaa,
0xef, 0x80, 0xaa, 0x22, 0xf1, 0xca, 0x5e, 0x9a,
0xc7, 0x94, 0x11, 0x0a, 0xc6, 0xcd, 0x04, 0x8d,
0xa9
};

// g24 (counter = 0)
static readonly byte[] g24_x = {
0x01, 0xdd, 0xc9, 0xa9, 0x91, 0xf6, 0x18, 0x0a,
0xbe, 0xfa, 0xae, 0x93, 0x51, 0xed, 0xce, 0x52,
0x4d, 0xb8, 0xb0, 0xd7, 0xe4, 0x27, 0xc3, 0xeb,
0xa5, 0x88, 0x17, 0xa5, 0x08, 0x39, 0x6f, 0x2d,
0xbf, 0x6e, 0x1a, 0x75, 0x43, 0x62, 0x7e, 0x05,
0xc3, 0xbb, 0xe6, 0x5d, 0xfe, 0x40, 0xd0, 0xb3,
0x31, 0x89, 0x40, 0x1e, 0xf1, 0x9a, 0xb1, 0xb1,
0x6f, 0xf3, 0x3b, 0xbc, 0x44, 0x3a, 0x5f, 0xe5,
0x47, 0xf9
};
static readonly byte[] g24_y = {
0xfa, 0x05, 0xcc, 0xbe, 0x21, 0xe0, 0xbb, 0xa1,
0x88, 0x81, 0xdc, 0x66, 0x2a, 0xb2, 0x9f, 0x3d,
0x9c, 0x96, 0x11, 0x0b, 0x2a, 0xb6, 0x0f, 0xa5,
0x50, 0x32, 0x21, 0x37, 0x7a, 0xcf, 0xb0, 0x48,
0x51, 0x47, 0x9a, 0x96, 0x12, 0xda, 0xb9, 0x94,
0xda, 0x56, 0x72, 0x57, 0x1d, 0xc5, 0xc4, 0x39,
0x3f, 0x75, 0xb2, 0xc8, 0x4a, 0xab, 0x3f, 0xa6,
0xa0, 0xc4, 0x49, 0x7f, 0x0e, 0x55, 0xb1, 0x2c,
0xb2
};

// g25 (counter = 1)
static readonly byte[] g25_x = {
0x01, 0xeb, 0x90, 0x35, 0x84, 0x06, 0x30, 0xac,
0x6f, 0xd0, 0xa3, 0x01, 0x73, 0xa9, 0x47, 0x93,
0xf7, 0xe8, 0x1f, 0x34, 0x45, 0x5e, 0x2e, 0x1a,
0x60, 0x0f, 0xde, 0xc0, 0x9b, 0xb6, 0x63, 0x61,
0x63, 0xc6, 0xc4, 0xb2, 0x41, 0x75, 0xb5, 0x60,
0x53, 0xe6, 0x7c, 0x3e, 0x37, 0x4d, 0x35, 0x73,
0x9b, 0xe3, 0x18, 0x19, 0xcd, 0xe7, 0xbe, 0xfc,
0x42, 0xeb, 0xc3, 0xae, 0xa0, 0xab, 0xdd, 0x03,
0xef, 0xc7
};
static readonly byte[] g25_y = {
0xda, 0x55, 0x65, 0xb4, 0xf6, 0x16, 0xdc, 0xe1,
0x8b, 0xe9, 0x65, 0x7f, 0x38, 0x6f, 0x0c, 0x27,
0x92, 0xfc, 0xf0, 0xf6, 0x02, 0xc6, 0xac, 0xf8,
0x4f, 0xcb, 0xc9, 0x76, 0xbb, 0x27, 0x6d, 0x6d,
0x84, 0x97, 0x78, 0x75, 0xdc, 0x12, 0xf8, 0x87,
0x65, 0x4a, 0xc7, 0xc2, 0x87, 0x3c, 0x3a, 0xef,
0xf1, 0xc0, 0xd4, 0x5c, 0xdd, 0xce, 0x4c, 0xd6,
0x33, 0x28, 0x59, 0x8a, 0x67, 0xb0, 0xb2, 0x22,
0xbf
};

// g26 (counter = 1)
static readonly byte[] g26_x = {
0x01, 0x2a, 0x8c, 0x69, 0x5c, 0x84, 0x96, 0x14,
0xe9, 0xa6, 0xa1, 0xd4, 0xb3, 0x98, 0x8b, 0x6a,
0x56, 0xbb, 0xce, 0xb7, 0x80, 0xd1, 0xcd, 0x2b,
0x2b, 0xe3, 0x37, 0xb1, 0x72, 0x0a, 0x27, 0x93,
0x6f, 0xb0, 0x5e, 0xcf, 0xa2, 0x1a, 0xac, 0x9e,
0x99, 0xb7, 0x03, 0x1b, 0xda, 0x94, 0x21, 0xb5,
0xb3, 0xb9, 0x55, 0x5b, 0xaa, 0xda, 0xc9, 0x6b,
0xab, 0xce, 0x8a, 0x7e, 0xe4, 0xe1, 0xe2, 0x65,
0x75, 0x27
};
static readonly byte[] g26_y = {
0x0c, 0x2c, 0x28, 0x76, 0x3f, 0xd1, 0xe3, 0x92,
0xd0, 0x7a, 0xe2, 0xed, 0x96, 0x99, 0x12, 0xdb,
0x6c, 0x78, 0xb8, 0xbb, 0xfa, 0x21, 0x6e, 0x56,
0x52, 0xc6, 0xfb, 0x47, 0x73, 0x1e, 0xc4, 0x55,
0x93, 0x8d, 0xa2, 0x5c, 0x17, 0xfb, 0x58, 0x2e,
0xa5, 0x02, 0x28, 0x10, 0x30, 0x84, 0x97, 0x26,
0xf4, 0x78, 0xca, 0xd1, 0xa8, 0xb5, 0x6b, 0xe9,
0x21, 0xa9, 0x0b, 0x53, 0x61, 0x0b, 0x62, 0xa0,
0x9c
};

// g27 (counter = 1)
static readonly byte[] g27_x = {
0x01, 0x30, 0xb2, 0x74, 0x65, 0xad, 0x69, 0xc6,
0x73, 0x2c, 0xda, 0x11, 0x80, 0x50, 0xcd, 0xdd,
0xa2, 0x88, 0x1a, 0x08, 0x70, 0xf1, 0xee, 0x9c,
0xeb, 0xd8, 0x1a, 0x56, 0x56, 0x75, 0x5c, 0xb6,
0xb0, 0x80, 0xa4, 0x28, 0x93, 0x8b, 0xae, 0xd1,
0xf7, 0xd6, 0x77, 0x2a, 0x8f, 0x55, 0x92, 0x34,
0xfb, 0xa2, 0x0f, 0xd3, 0x00, 0xf9, 0x1d, 0x88,
0xcf, 0xf9, 0x9b, 0x79, 0x4f, 0x1f, 0xad, 0x64,
0x79, 0xd4
};
static readonly byte[] g27_y = {
0x8c, 0x65, 0x73, 0xa8, 0x9d, 0x9b, 0x48, 0xd9,
0xee, 0x84, 0x74, 0xd7, 0x47, 0x52, 0x6b, 0xcd,
0x74, 0xb2, 0x0d, 0x54, 0x5c, 0x42, 0xab, 0x75,
0x50, 0xd6, 0xf4, 0x90, 0x61, 0xfc, 0xcb, 0x09,
0x57, 0x9c, 0x39, 0xe6, 0x70, 0xbc, 0x04, 0xbe,
0x08, 0x20, 0xbd, 0x2f, 0xc3, 0x9a, 0x54, 0x11,
0x48, 0x60, 0x41, 0xba, 0x94, 0xe4, 0xb4, 0x05,
0x6a, 0x19, 0x6b, 0x07, 0x6a, 0x22, 0x2b, 0xe0,
0x1b
};

// g28 (counter = 1)
static readonly byte[] g28_x = {
0x01, 0x2c, 0xae, 0xae, 0x62, 0xf4, 0x2d, 0x13,
0x14, 0x1f, 0xc2, 0x8b, 0xf5, 0x23, 0x2f, 0x95,
0xdb, 0xe9, 0xdc, 0xd7, 0x21, 0xb2, 0x59, 0x90,
0x41, 0x50, 0x7d, 0xa3, 0xa5, 0xc1, 0x46, 0x7a,
0x19, 0xfa, 0x46, 0x9b, 0xa0, 0x91, 0x40, 0x10,
0x66, 0x22, 0xe6, 0x77, 0x71, 0xf3, 0x32, 0xe1,
0x9b, 0x3c, 0x17, 0x3a, 0xfd, 0x1e, 0x23, 0xd8,
0x45, 0xdb, 0x96, 0x00, 0xbe, 0x07, 0x55, 0xf8,
0x5e, 0xf6
};
static readonly byte[] g28_y = {
0xec, 0x2f, 0x91, 0x95, 0xd2, 0xbf, 0x0f, 0x74,
0x9f, 0x1a, 0x73, 0xdc, 0x13, 0xb7, 0x61, 0x89,
0x24, 0xb3, 0xd3, 0x79, 0xdb, 0x21, 0x29, 0x13,
0x71, 0x88, 0xb9, 0xe8, 0xe5, 0x53, 0x35, 0xc0,
0x96, 0xd6, 0x3a, 0x27, 0xf5, 0xbb, 0x02, 0xf5,
0xb4, 0xf5, 0xac, 0xd7, 0xc5, 0xc4, 0xba, 0x9a,
0x7b, 0xfb, 0xf8, 0xbe, 0xaf, 0x1c, 0x37, 0x47,
0xd9, 0x4a, 0xb2, 0xfc, 0x5f, 0x13, 0x10, 0xfe,
0xd8
};

// g29 (counter = 0)
static readonly byte[] g29_x = {
0xc2, 0x18, 0x2b, 0xa6, 0xe8, 0x20, 0xe8, 0xad,
0x89, 0xac, 0x8f, 0x62, 0x55, 0x21, 0x60, 0xab,
0xca, 0xbd, 0xd1, 0x47, 0xe6, 0x3f, 0x76, 0x9a,
0xdc, 0x69, 0xa6, 0x23, 0xeb, 0xec, 0x89, 0x10,
0x2a, 0xa8, 0x3d, 0xfa, 0x2d, 0x6d, 0x6a, 0xf6,
0x26, 0x3b, 0x49, 0x7a, 0x71, 0x50, 0x61, 0x0e,
0xba, 0x6a, 0xd9, 0x82, 0x97, 0xa2, 0x78, 0xeb,
0x81, 0xd6, 0x62, 0xc2, 0xde, 0xf3, 0x8d, 0xe0,
0x40
};
static readonly byte[] g29_y = {
0x9a, 0xe7, 0xe0, 0x67, 0x15, 0x72, 0x61, 0x9b,
0x73, 0x0b, 0x33, 0xab, 0xdc, 0x8c, 0x9a, 0x46,
0x08, 0xd6, 0x41, 0xe1, 0x63, 0x45, 0x63, 0x10,
0xbe, 0x94, 0xcf, 0x55, 0xae, 0x4c, 0x92, 0x5c,
0x6f, 0xe6, 0x3a, 0x08, 0x1f, 0xb1, 0x41, 0x9b,
0x66, 0xc3, 0xa8, 0xfc, 0x19, 0x9e, 0x7e, 0xcb,
0x11, 0x4a, 0x9b, 0xde, 0xb1, 0x3e, 0xe7, 0xac,
0xf3, 0xac, 0x7e, 0xdd, 0x43, 0xf7, 0x22, 0x35,
0x18
};

// g30 (counter = 0)
static readonly byte[] g30_x = {
0x51, 0x7c, 0x31, 0x26, 0x55, 0x63, 0x5d, 0x41,
0xdf, 0x5e, 0xfe, 0x8f, 0x8a, 0xf3, 0x06, 0x77,
0xe0, 0xe3, 0xcf, 0x5a, 0xdd, 0x48, 0xaa, 0xb1,
0x57, 0xf4, 0x5d, 0xff, 0x62, 0x15, 0xcd, 0xcf,
0x5f, 0x5c, 0x1a, 0xd4, 0xdc, 0x34, 0x74, 0xbc,
0xdd, 0x40, 0x1b, 0xfa, 0x8b, 0x77, 0x10, 0x3b,
0x20, 0xc9, 0xc2, 0x89, 0x25, 0xa1, 0x8e, 0xbf,
0xea, 0xcd, 0x8d, 0x2b, 0x0a, 0x1d, 0x52, 0xd4,
0x8d
};
static readonly byte[] g30_y = {
0x5d, 0x68, 0x30, 0xd3, 0x3c, 0x7e, 0xdf, 0x55,
0x4d, 0x9f, 0x14, 0xc9, 0xe6, 0xee, 0x50, 0x29,
0x66, 0xf8, 0x33, 0xa1, 0xc5, 0xef, 0x50, 0x6b,
0x6e, 0xf4, 0x4b, 0x91, 0x71, 0x66, 0x4e, 0x99,
0xe5, 0xcc, 0x1c, 0x8b, 0x09, 0xd7, 0x04, 0xff,
0x9e, 0x72, 0xd0, 0xc6, 0x0e, 0x1f, 0x7c, 0xe6,
0xe6, 0xf8, 0x98, 0x9a, 0x88, 0xa0, 0x3b, 0x1c,
0x30, 0x07, 0x34, 0xf0, 0x94, 0xc2, 0xc5, 0x62,
0x4c
};

// g31 (counter = 4)
static readonly byte[] g31_x = {
0x6f, 0x20, 0x17, 0xda, 0x1d, 0x5c, 0x88, 0x27,
0x45, 0x84, 0x75, 0x2c, 0x8a, 0xb8, 0xc7, 0xef,
0x8a, 0xe7, 0xf6, 0xa9, 0xec, 0xeb, 0x3a, 0x8b,
0x3b, 0x9a, 0xf1, 0x6b, 0x5e, 0x03, 0x47, 0xb0,
0x1f, 0xa6, 0x62, 0xb0, 0xdb, 0x27, 0xfb, 0xab,
0x75, 0x74, 0x9a, 0xe6, 0x83, 0x57, 0x13, 0x5f,
0x3d, 0xd2, 0x54, 0x4d, 0x2e, 0x1a, 0x7d, 0x95,
0x94, 0xbf, 0xd7, 0xd7, 0xdd, 0xf9, 0xd6, 0x9c,
0x25
};
static readonly byte[] g31_y = {
0xd6, 0x21, 0x95, 0xe1, 0x4e, 0x6b, 0x29, 0x65,
0x41, 0x47, 0xdb, 0x19, 0xd0, 0x2c, 0xd7, 0x33,
0x81, 0x02, 0x0d, 0x7a, 0x55, 0x59, 0x88, 0x0e,
0x3e, 0x2f, 0x41, 0x34, 0x3e, 0xf3, 0x9b, 0xd2,
0x9b, 0x61, 0x3d, 0x8b, 0x55, 0xae, 0x33, 0xd5,
0x59, 0xdf, 0x36, 0xa1, 0x7d, 0x1f, 0xa4, 0xda,
0xe0, 0xb5, 0xa3, 0x52, 0xf4, 0x59, 0x16, 0x8f,
0x94, 0xd0, 0xa5, 0xb4, 0x56, 0x14, 0x70, 0x27,
0x2c
};

// g32 (counter = 0)
static readonly byte[] g32_x = {
0x01, 0xf6, 0x66, 0x42, 0x5f, 0xa8, 0xbd, 0xe9,
0x8b, 0xff, 0x13, 0x81, 0x61, 0x79, 0x6c, 0x12,
0xe0, 0xe1, 0xe4, 0x85, 0xea, 0xda, 0xcd, 0x0e,
0xd5, 0xad, 0xab, 0x17, 0xba, 0x92, 0xb3, 0xa2,
0x93, 0x28, 0xfe, 0x05, 0x71, 0x96, 0xd6, 0x7a,
0x03, 0x6c, 0x92, 0x2f, 0x08, 0x74, 0x36, 0xe9,
0x81, 0x87, 0xa4, 0x25, 0xb9, 0x22, 0x6a, 0xe7,
0x5a, 0x0a, 0x91, 0xb8, 0x48, 0xad, 0xa6, 0x7a,
0x9b, 0xe0
};
static readonly byte[] g32_y = {
0xb7, 0x8d, 0x81, 0x52, 0xba, 0xae, 0x1f, 0xe2,
0x90, 0x18, 0x35, 0x64, 0xec, 0x0c, 0xde, 0xdd,
0xef, 0x62, 0x49, 0x22, 0x0b, 0x98, 0x6d, 0xf8,
0xbb, 0xf3, 0xcb, 0x9a, 0x04, 0x7f, 0x06, 0x62,
0x79, 0xd3, 0xcd, 0x07, 0x3b, 0x5d, 0x8f, 0x18,
0x50, 0x8a, 0x17, 0xf7, 0xaf, 0xcc, 0xf4, 0x07,
0x58, 0x32, 0x4f, 0xe2, 0x51, 0xc5, 0x4e, 0x2b,
0xbf, 0xdd, 0x2b, 0xe1, 0x6d, 0xdc, 0xd3, 0x5a,
0xcc
};

// g33 (counter = 2)
static readonly byte[] g33_x = {
0x1c, 0x78, 0xaa, 0x54, 0x76, 0xf6, 0xcc, 0x9b,
0x6d, 0x22, 0xc3, 0xa2, 0x58, 0xbe, 0xbc, 0x1d,
0xf2, 0xce, 0x2e, 0x39, 0x1d, 0xc7, 0x75, 0xb8,
0x14, 0xce, 0xce, 0x94, 0xf6, 0x4f, 0xed, 0x29,
0xc0, 0xc9, 0xa4, 0x0a, 0xe7, 0xb7, 0x8f, 0x21,
0x81, 0xaf, 0x16, 0x3d, 0xec, 0x83, 0x18, 0x67,
0x84, 0x93, 0xd1, 0xb9, 0x85, 0x26, 0x06, 0xe5,
0x8f, 0x96, 0x37, 0x54, 0x65, 0x21, 0x41, 0xcd,
0x1b
};
static readonly byte[] g33_y = {
0x3e, 0x4c, 0x46, 0x81, 0x42, 0xcc, 0x1b, 0x6f,
0xb6, 0x41, 0x17, 0xc6, 0x69, 0x7f, 0x29, 0xc3,
0x36, 0xfd, 0xb2, 0x72, 0xa4, 0x83, 0x64, 0x23,
0x39, 0xd8, 0x38, 0x22, 0x06, 0x03, 0xe1, 0x1d,
0x4c, 0x6d, 0xc8, 0x0b, 0x7b, 0x1e, 0x5d, 0x13,
0x21, 0xab, 0x11, 0xa0, 0x2d, 0x7d, 0x81, 0xa2,
0x5a, 0x14, 0xee, 0xd5, 0x80, 0x8c, 0x7a, 0x09,
0x54, 0x81, 0xc7, 0xb9, 0x21, 0x6d, 0x07, 0x65,
0x80
};

// g34 (counter = 2)
static readonly byte[] g34_x = {
0x01, 0x4b, 0xe3, 0x06, 0xc2, 0xa3, 0x0c, 0x0b,
0x90, 0xa7, 0x41, 0xa3, 0x7d, 0xe3, 0x82, 0xb5,
0x65, 0xa0, 0x78, 0x74, 0x71, 0xdb, 0xd3, 0x72,
0x6f, 0x3a, 0xde, 0xde, 0x7a, 0xb1, 0xfd, 0xae,
0x81, 0xbe, 0x59, 0x5a, 0xd9, 0x5e, 0x66, 0xd7,
0xe2, 0xdc, 0xf5, 0x38, 0x75, 0x18, 0xba, 0xdd,
0xe3, 0x12, 0x8e, 0x18, 0xf3, 0x2b, 0xc1, 0x98,
0xcb, 0x70, 0x0f, 0xa3, 0xb8, 0xb0, 0xfc, 0xf3,
0x87, 0x9b
};
static readonly byte[] g34_y = {
0xc8, 0x67, 0xc1, 0x21, 0xcb, 0xb3, 0x42, 0x53,
0xc6, 0xd4, 0xed, 0x65, 0xc7, 0x0b, 0xe2, 0x2d,
0x16, 0x17, 0xb6, 0x90, 0x0c, 0xab, 0xce, 0xe9,
0x28, 0xe7, 0x6b, 0x93, 0xbd, 0x16, 0xba, 0x13,
0xd4, 0x93, 0xd1, 0xc7, 0x1e, 0x62, 0x70, 0x94,
0xdd, 0x14, 0xd6, 0xa1, 0xb2, 0xcb, 0xae, 0xa6,
0x45, 0x05, 0xcc, 0x55, 0x24, 0x9c, 0x97, 0x86,
0x97, 0x17, 0x03, 0x50, 0x27, 0x23, 0xdb, 0xeb,
0x23
};

// g35 (counter = 3)
static readonly byte[] g35_x = {
0x01, 0xf9, 0x73, 0x25, 0xc4, 0x55, 0xdd, 0x3b,
0x04, 0xd5, 0x8a, 0x79, 0xc7, 0x6d, 0x80, 0xe4,
0x19, 0xd6, 0xe7, 0x58, 0x27, 0x84, 0x7a, 0x4b,
0x51, 0x48, 0xcc, 0x32, 0x7c, 0x34, 0x0d, 0xe6,
0x67, 0x0a, 0xef, 0xe5, 0x99, 0x88, 0xc1, 0xb6,
0x14, 0x9a, 0xd3, 0x9d, 0x59, 0x49, 0x8a, 0xfe,
0x43, 0x4f, 0xac, 0x28, 0x6f, 0x51, 0x96, 0x7d,
0xe9, 0xe1, 0x26, 0xe5, 0xa2, 0x3a, 0x31, 0x06,
0x1c, 0x82
};
static readonly byte[] g35_y = {
0x1a, 0xa4, 0x22, 0xac, 0x52, 0xf5, 0x08, 0x43,
0x93, 0xf4, 0x64, 0x07, 0x68, 0xed, 0xf0, 0xc2,
0x85, 0x8f, 0x79, 0x08, 0x6f, 0x92, 0x4c, 0xe3,
0xc1, 0x80, 0xb7, 0xf1, 0x1d, 0x91, 0xc8, 0xd4,
0x23, 0x5f, 0x9f, 0x74, 0x57, 0x59, 0x3e, 0xc8,
0x21, 0xdd, 0xa0, 0x6a, 0x10, 0x79, 0x2b, 0xd6,
0x9e, 0xc3, 0x41, 0xba, 0x84, 0xb1, 0x26, 0x08,
0x34, 0x2c, 0xc7, 0x15, 0xd1, 0xcd, 0x34, 0x2f,
0xfc
};

// g36 (counter = 5)
static readonly byte[] g36_x = {
0xaf, 0x4b, 0x0e, 0x42, 0x74, 0x00, 0xa8, 0x64,
0xaf, 0x89, 0x9a, 0x2a, 0x56, 0xcf, 0x65, 0x0e,
0xf4, 0xd4, 0xa8, 0x66, 0x58, 0x86, 0x02, 0x7b,
0xbc, 0xce, 0x7c, 0x10, 0x04, 0xdd, 0x72, 0x5e,
0xb4, 0xbb, 0x62, 0xd1, 0xb7, 0x3c, 0x5e, 0x28,
0x5c, 0xf2, 0xae, 0xd3, 0xcd, 0x49, 0x0c, 0x06,
0xff, 0x2a, 0x16, 0x13, 0xeb, 0x66, 0x1f, 0x8c,
0x0a, 0x86, 0x73, 0x8c, 0xae, 0xa2, 0xf5, 0xd4,
0xd9
};
static readonly byte[] g36_y = {
0x34, 0x94, 0xaa, 0x6e, 0x02, 0xe8, 0x14, 0x3a,
0x85, 0x7a, 0xfc, 0x9a, 0x6c, 0x79, 0xe8, 0x6c,
0x31, 0xbb, 0x5d, 0xe7, 0xa0, 0x1b, 0xfc, 0xca,
0x4a, 0x57, 0x57, 0x2f, 0xef, 0x37, 0x3a, 0x34,
0x36, 0xe3, 0x22, 0x99, 0x44, 0xcc, 0xa1, 0x5e,
0x52, 0x65, 0x28, 0xa2, 0x2b, 0x0f, 0xb1, 0xa3,
0x51, 0x56, 0x95, 0xa8, 0xe4, 0xdf, 0x14, 0x18,
0x8f, 0x81, 0xc5, 0x55, 0xe7, 0x80, 0x7a, 0x3f,
0x78
};

// g37 (counter = 0)
static readonly byte[] g37_x = {
0x35, 0x93, 0x98, 0x49, 0x17, 0x82, 0x4d, 0xf5,
0xc9, 0x47, 0x02, 0xcf, 0xd7, 0xb0, 0x5b, 0x3e,
0xb1, 0xab, 0xf7, 0x3c, 0xba, 0xb6, 0x2a, 0x43,
0xdf, 0x96, 0x71, 0xdf, 0x61, 0x26, 0xa1, 0x65,
0xa2, 0x72, 0x7d, 0x27, 0x99, 0x7a, 0x2b, 0x91,
0x1d, 0xc1, 0x52, 0x8f, 0xe4, 0x22, 0x63, 0x47,
0x7c, 0xe2, 0xe8, 0xee, 0x8b, 0xf1, 0xec, 0x22,
0xc9, 0x32, 0x45, 0x9a, 0xda, 0xa3, 0x75, 0x8e,
0x0d
};
static readonly byte[] g37_y = {
0x3e, 0x00, 0x09, 0x55, 0xd5, 0x68, 0xb3, 0x35,
0xff, 0x04, 0xc7, 0xf0, 0x11, 0xf8, 0xb0, 0xae,
0xae, 0x2e, 0xbd, 0x06, 0x6f, 0x85, 0xbd, 0xe6,
0x88, 0x82, 0xfb, 0x84, 0x0d, 0x05, 0xb2, 0x91,
0xc9, 0x4f, 0x26, 0x6e, 0x82, 0x66, 0x4f, 0x23,
0x9b, 0x96, 0x8a, 0xcb, 0x22, 0x04, 0x11, 0xb5,
0x46, 0xd4, 0x6c, 0xb6, 0xce, 0x72, 0xd9, 0x8c,
0xd7, 0xa6, 0xde, 0xc5, 0x4e, 0x24, 0x69, 0xb6,
0x32
};

// g38 (counter = 0)
static readonly byte[] g38_x = {
0x01, 0x93, 0x20, 0xc9, 0x93, 0x9d, 0xf2, 0x47,
0x11, 0xf9, 0x84, 0x4f, 0x03, 0x12, 0xaa, 0xb8,
0x0e, 0x44, 0x76, 0x99, 0xe4, 0xd0, 0x72, 0xa2,
0xd9, 0xe8, 0x0a, 0x29, 0xda, 0xfe, 0x86, 0xca,
0xf0, 0x58, 0xfa, 0xb3, 0xa8, 0xa8, 0xd4, 0x97,
0x55, 0x15, 0x86, 0x4c, 0x2b, 0x93, 0x1c, 0x7f,
0x7d, 0xb3, 0xb2, 0x5f, 0x2f, 0x5f, 0x9e, 0x5f,
0x8b, 0x53, 0x15, 0xa6, 0x81, 0xb3, 0x93, 0xb0,
0x8a, 0x7f
};
static readonly byte[] g38_y = {
0xab, 0x49, 0x05, 0x0e, 0xa4, 0x99, 0x25, 0xfc,
0x1c, 0xb8, 0x6a, 0xce, 0xdf, 0x42, 0x19, 0x65,
0x9b, 0x24, 0x90, 0xe9, 0x1b, 0x09, 0xf4, 0x25,
0xbf, 0xf0, 0x51, 0x60, 0x0a, 0x9e, 0x31, 0x46,
0x4e, 0x2e, 0xa9, 0x69, 0x74, 0x54, 0x5b, 0x7c,
0x74, 0x58, 0x38, 0xa7, 0x5a, 0x49, 0xc8, 0x8b,
0xca, 0xcd, 0x83, 0x7a, 0xc8, 0x95, 0x7f, 0xc1,
0x4f, 0x71, 0x1d, 0xea, 0x93, 0x07, 0x0f, 0x3d,
0x99
};

// g39 (counter = 1)
static readonly byte[] g39_x = {
0x01, 0x2b, 0xa5, 0xb5, 0x32, 0xac, 0xb5, 0xce,
0xee, 0x9c, 0x91, 0xa3, 0x25, 0x69, 0x57, 0x86,
0xbf, 0x7e, 0x18, 0x42, 0xb8, 0x31, 0xf8, 0xda,
0x30, 0xd3, 0x60, 0xd4, 0x28, 0xae, 0x53, 0x3e,
0xf2, 0xd1, 0x07, 0x1e, 0x81, 0x31, 0xf7, 0xa7,
0xa7, 0x29, 0xbc, 0x2e, 0xd1, 0x9a, 0xe0, 0xc1,
0x29, 0xb8, 0xab, 0x03, 0x77, 0xfb, 0x55, 0xe0,
0x19, 0xe6, 0x49, 0x32, 0xeb, 0x28, 0xf8, 0x0b,
0x93, 0x34
};
static readonly byte[] g39_y = {
0x80, 0x3c, 0x8e, 0xbe, 0xaa, 0x49, 0x3d, 0x6f,
0xcb, 0x35, 0x6f, 0xb2, 0x56, 0x2b, 0x79, 0xd3,
0xa3, 0x76, 0xb5, 0xbc, 0xbf, 0x9b, 0x22, 0xbb,
0xee, 0x75, 0x61, 0x10, 0x75, 0xf7, 0xb7, 0xab,
0x16, 0x66, 0x6a, 0xbd, 0x97, 0xbe, 0x5b, 0x7e,
0xcb, 0x33, 0x88, 0x41, 0xfe, 0x77, 0xec, 0xe1,
0x35, 0xf4, 0xe2, 0xe4, 0xb8, 0x3a, 0xca, 0xfd,
0x71, 0xe7, 0x04, 0x52, 0x7a, 0x84, 0x78, 0x94,
0x68
};

// g40 (counter = 0)
static readonly byte[] g40_x = {
0xa0, 0x92, 0xee, 0xde, 0xd1, 0x85, 0xfa, 0x76,
0x82, 0xef, 0x7e, 0x4b, 0x2a, 0x19, 0x8a, 0x4e,
0xa0, 0x43, 0xa5, 0xeb, 0xa6, 0x2b, 0x44, 0x79,
0xa8, 0x65, 0x7d, 0x66, 0xd9, 0xda, 0xb7, 0xc5,
0xde, 0x65, 0xe4, 0x76, 0x42, 0xdc, 0xc6, 0x2e,
0x2a, 0x56, 0x1e, 0xb4, 0x94, 0xf1, 0xe2, 0x74,
0x8c, 0xdc, 0x79, 0x2f, 0x01, 0x14, 0xc7, 0x7f,
0x52, 0xb7, 0xd7, 0xc5, 0xf5, 0xc4, 0xb0, 0x58,
0x83
};
static readonly byte[] g40_y = {
0xd7, 0x4a, 0x59, 0x4a, 0x66, 0x1d, 0x40, 0xea,
0x62, 0x97, 0x02, 0xdd, 0x12, 0x47, 0x80, 0x8f,
0xb2, 0x88, 0xdd, 0x50, 0x40, 0xe5, 0x41, 0x0d,
0x7d, 0x6b, 0x04, 0x59, 0x1e, 0x39, 0x69, 0x6b,
0xf2, 0x99, 0x95, 0xa3, 0x52, 0x39, 0x28, 0x41,
0x3d, 0x74, 0x6a, 0xcf, 0x2c, 0xdb, 0x55, 0xc4,
0x0a, 0xa0, 0x84, 0xcb, 0x65, 0xa2, 0x89, 0xbd,
0xfc, 0x85, 0x01, 0xfd, 0xcc, 0xb4, 0x02, 0x09,
0xf9
};

// g41 (counter = 0)
static readonly byte[] g41_x = {
0xef, 0x45, 0x1e, 0x11, 0x0d, 0xaf, 0x9d, 0x5e,
0x39, 0x1a, 0x53, 0x59, 0x93, 0x4a, 0x9f, 0x69,
0x69, 0x51, 0x7f, 0x43, 0x8c, 0x7b, 0xea, 0x18,
0x3d, 0x5e, 0x96, 0x29, 0x0d, 0xdc, 0xe1, 0x20,
0x87, 0x3e, 0x45, 0x48, 0x6d, 0x0c, 0x41, 0xc8,
0x87, 0xc4, 0x83, 0xb4, 0x66, 0x1d, 0xfb, 0x48,
0xd8, 0xd7, 0x2c, 0x0f, 0xb8, 0x9e, 0x87, 0xe3,
0x9c, 0x8a, 0x37, 0x16, 0xe9, 0xcd, 0xc7, 0x6d,
0xdd
};
static readonly byte[] g41_y = {
0xef, 0xa5, 0xc6, 0x8d, 0x14, 0xde, 0x0e, 0xff,
0x1d, 0x72, 0x90, 0x05, 0xe2, 0x6e, 0x74, 0xf4,
0x9f, 0x8b, 0xc0, 0x83, 0x12, 0x4a, 0x43, 0x2a,
0xe1, 0x8d, 0xd5, 0x13, 0x1c, 0x04, 0x77, 0x6f,
0x59, 0xa3, 0xf9, 0x32, 0x74, 0xc5, 0x17, 0x7f,
0x3c, 0x91, 0xa2, 0x01, 0x60, 0x65, 0x7a, 0x11,
0x4f, 0x5b, 0xc0, 0xdb, 0xcc, 0x5c, 0x80, 0x69,
0x98, 0x4e, 0x34, 0x13, 0x20, 0x41, 0x2a, 0xdf,
0x1a
};

// g42 (counter = 0)
static readonly byte[] g42_x = {
0x01, 0x18, 0x71, 0x3f, 0xa1, 0x1a, 0x53, 0xa2,
0x4b, 0x33, 0xf9, 0xe1, 0x56, 0xce, 0xcf, 0x8b,
0x41, 0x58, 0x74, 0x79, 0x45, 0x08, 0xfe, 0x54,
0x38, 0xf8, 0xc4, 0xda, 0xe1, 0x38, 0x66, 0x6a,
0x5b, 0x88, 0x8e, 0x8e, 0x99, 0x43, 0x8d, 0xab,
0xc1, 0x55, 0x56, 0x46, 0x57, 0x47, 0x11, 0x6a,
0xc4, 0x72, 0xad, 0x95, 0xd2, 0x92, 0x0f, 0xe9,
0xb3, 0x6f, 0x30, 0xbe, 0x09, 0x63, 0xd7, 0xe0,
0xdb, 0x12
};
static readonly byte[] g42_y = {
0x95, 0x4b, 0xea, 0x5f, 0x61, 0xe4, 0xa9, 0xd8,
0xd0, 0x93, 0xd1, 0xb9, 0xd6, 0xab, 0x52, 0xb2,
0xfa, 0xc8, 0xb7, 0x86, 0x55, 0x80, 0x20, 0x10,
0xcf, 0x5e, 0xd9, 0x74, 0x59, 0x86, 0x42, 0x71,
0xef, 0x4c, 0x34, 0xd8, 0x04, 0x4c, 0x26, 0xc7,
0x76, 0x3d, 0xe3, 0x28, 0xe5, 0x09, 0xef, 0x0a,
0x68, 0xc1, 0x17, 0x31, 0x30, 0xad, 0x63, 0x01,
0xb2, 0x0c, 0x3b, 0xc8, 0xf8, 0xf2, 0x6a, 0xd2,
0x96
};

// g43 (counter = 0)
static readonly byte[] g43_x = {
0x01, 0x21, 0x0b, 0x8e, 0xfd, 0x90, 0xf3, 0xb9,
0xec, 0x0d, 0x8f, 0x3f, 0x82, 0xf6, 0x87, 0x8a,
0xbf, 0x54, 0xa8, 0x8f, 0xf0, 0x68, 0x72, 0x06,
0x4e, 0xb7, 0xf4, 0xe9, 0x03, 0xb5, 0x30, 0x94,
0x5c, 0x5a, 0x7f, 0x0b, 0xe3, 0xb3, 0x0d, 0x00,
0x48, 0x01, 0x70, 0x0a, 0xe3, 0xdd, 0x25, 0x32,
0x7f, 0x4b, 0xc2, 0x86, 0x08, 0xbe, 0x51, 0x2f,
0x80, 0x0a, 0x84, 0x54, 0xdf, 0x8d, 0x15, 0x06,
0x9a, 0x18
};
static readonly byte[] g43_y = {
0x8e, 0x6e, 0xfc, 0x12, 0xaa, 0xd5, 0x47, 0xb3,
0xe0, 0x8d, 0x39, 0xf7, 0xa7, 0x9e, 0x9b, 0x03,
0x46, 0x16, 0xf0, 0x98, 0x17, 0x5f, 0x71, 0x65,
0x61, 0xb7, 0x92, 0x7a, 0xb1, 0x15, 0xa5, 0x4b,
0xe4, 0xfd, 0xec, 0xd3, 0x84, 0x5f, 0x7f, 0x53,
0x58, 0x82, 0x8f, 0x17, 0x5b, 0x83, 0xed, 0x25,
0xef, 0x24, 0x92, 0xc8, 0x72, 0x96, 0xd6, 0xd8,
0x9e, 0xcb, 0x98, 0x81, 0x10, 0x4b, 0xe9, 0xbe,
0x7d
};

// g44 (counter = 0)
static readonly byte[] g44_x = {
0x01, 0xfb, 0xd7, 0x8c, 0x94, 0xec, 0x52, 0xe5,
0xd7, 0x69, 0x39, 0x2c, 0xf0, 0x84, 0xc5, 0xe6,
0x29, 0xb2, 0x9c, 0xf6, 0xe3, 0xbb, 0xa7, 0xf3,
0xcd, 0x0c, 0x61, 0x2c, 0x2f, 0x61, 0x0c, 0xe3,
0x88, 0xbb, 0x72, 0xa9, 0xc7, 0xd4, 0xdd, 0xc3,
0xf5, 0x61, 0x20, 0x86, 0xcb, 0x42, 0xbf, 0x34,
0x0f, 0x1e, 0x4a, 0xc7, 0x32, 0x4c, 0x73, 0x59,
0xef, 0x33, 0x51, 0x09, 0x5c, 0xb0, 0x0a, 0x7a,
0x30, 0x4e
};
static readonly byte[] g44_y = {
0x99, 0xdd, 0x83, 0xd1, 0xb4, 0xc0, 0xb2, 0x71,
0xf5, 0x1c, 0x48, 0x51, 0xb2, 0x36, 0xff, 0x7b,
0x7e, 0x86, 0x62, 0x66, 0x80, 0x68, 0x47, 0x96,
0x19, 0x37, 0xef, 0x20, 0xae, 0xcc, 0x98, 0x81,
0x22, 0xc0, 0x76, 0x03, 0x26, 0x74, 0xbe, 0x51,
0x3a, 0xa0, 0x39, 0x46, 0x16, 0x34, 0x77, 0xdd,
0xd2, 0x05, 0x94, 0x5b, 0x39, 0x17, 0xd1, 0xa3,
0xfc, 0xcc, 0xa2, 0xba, 0x9d, 0x42, 0x05, 0xaf,
0xf7
};

// g45 (counter = 0)
static readonly byte[] g45_x = {
0xec, 0xea, 0xda, 0xe3, 0x60, 0x7e, 0xa6, 0xaa,
0x2b, 0x65, 0x4f, 0xf6, 0x7f, 0x0c, 0x15, 0x69,
0xac, 0xc3, 0x23, 0xba, 0x3e, 0x6d, 0xcc, 0x5a,
0xf8, 0xec, 0xea, 0x30, 0xbb, 0xa4, 0x35, 0xbe,
0x1a, 0x62, 0x9a, 0x74, 0xd9, 0xe2, 0x37, 0x36,
0xfb, 0x93, 0xf3, 0xbe, 0xaf, 0xcf, 0xed, 0x2d,
0x36, 0xd5, 0x9e, 0x23, 0x31, 0x7b, 0xd0, 0xfd,
0x1f, 0x95, 0x9b, 0xd4, 0x58, 0x6e, 0x79, 0xf8,
0x9e
};
static readonly byte[] g45_y = {
0xf9, 0x73, 0x56, 0x28, 0xa3, 0x82, 0xe2, 0xad,
0x84, 0xfd, 0x1b, 0xad, 0xeb, 0xcf, 0x6c, 0x9e,
0xe5, 0xe3, 0x97, 0x89, 0x39, 0x1d, 0xe6, 0xbf,
0xb4, 0x09, 0xb0, 0xad, 0xac, 0xc1, 0x73, 0x10,
0x50, 0x54, 0x11, 0x8b, 0x44, 0x3c, 0xb4, 0xed,
0xa2, 0x32, 0x42, 0x8f, 0x44, 0x13, 0x18, 0x3a,
0x56, 0x48, 0x6d, 0x07, 0x3b, 0x84, 0xa5, 0x6d,
0x89, 0xcb, 0x72, 0x1e, 0xc9, 0x85, 0xeb, 0xc7,
0x51
};

// g46 (counter = 0)
static readonly byte[] g46_x = {
0xd2, 0x25, 0xf3, 0x9d, 0x10, 0x60, 0x0c, 0xc7,
0x61, 0xc1, 0x7f, 0xe1, 0x5a, 0xce, 0x8b, 0xba,
0xdc, 0x77, 0x6a, 0xba, 0x28, 0xbc, 0xba, 0xe4,
0x82, 0xd1, 0x5f, 0x79, 0xfe, 0x38, 0xdd, 0x0b,
0xc2, 0xc9, 0xbd, 0xe1, 0x2d, 0x6e, 0xe2, 0x50,
0x48, 0x9d, 0x0e, 0x7a, 0x23, 0x48, 0x87, 0x11,
0x85, 0x7f, 0xb9, 0x13, 0xde, 0x74, 0x44, 0x16,
0x76, 0xfd, 0x3c, 0x98, 0xda, 0x4e, 0x8f, 0xd8,
0xf6
};
static readonly byte[] g46_y = {
0x97, 0xb3, 0xf4, 0x35, 0xc0, 0x04, 0x16, 0xa5,
0x95, 0x86, 0xe7, 0x59, 0x31, 0x28, 0xc1, 0x2b,
0x73, 0x47, 0x23, 0xaa, 0x88, 0x7f, 0x98, 0xa0,
0xb4, 0x82, 0xd1, 0x8d, 0x38, 0x70, 0x7b, 0xb9,
0x33, 0xe4, 0xb4, 0xbc, 0xcc, 0xdc, 0x71, 0x07,
0x5f, 0x51, 0x74, 0xe8, 0xf2, 0x13, 0x3b, 0x74,
0xb5, 0x44, 0xb2, 0x9a, 0x79, 0x6b, 0x4f, 0xd8,
0xb7, 0x83, 0x66, 0x4a, 0xf1, 0xd3, 0x65, 0x9e,
0x99
};

// g47 (counter = 0)
static readonly byte[] g47_x = {
0xdb, 0xcc, 0x8b, 0xb8, 0x8c, 0x00, 0xc5, 0x31,
0x89, 0x0b, 0xcc, 0x22, 0xa2, 0xf2, 0x21, 0xb8,
0xca, 0x91, 0x6b, 0x9c, 0x32, 0x4c, 0x08, 0x1c,
0x71, 0x23, 0xe5, 0xca, 0x66, 0x06, 0x9e, 0x0b,
0xca, 0xc9, 0x1e, 0x09, 0xbf, 0x95, 0xfa, 0xe7,
0xd1, 0x5f, 0x8a, 0xac, 0xaa, 0x17, 0x26, 0xd7,
0x0a, 0xc0, 0x30, 0x57, 0xf7, 0xd8, 0x6f, 0x0e,
0x27, 0x37, 0x6c, 0xcc, 0x1f, 0xcd, 0x7b, 0x8c,
0x7e
};
static readonly byte[] g47_y = {
0x2d, 0x40, 0x01, 0x61, 0x5d, 0x62, 0x5d, 0x51,
0x26, 0x3a, 0x6f, 0x7b, 0xf2, 0x63, 0xc0, 0xe2,
0x1c, 0x65, 0x68, 0x64, 0xbd, 0x9b, 0x39, 0x5c,
0x17, 0x34, 0x36, 0x6c, 0xe3, 0x75, 0x51, 0x8c,
0x05, 0x8a, 0x34, 0xe7, 0x48, 0x11, 0xc1, 0x4c,
0xa0, 0x72, 0x27, 0xba, 0xa0, 0xbd, 0x0b, 0x59,
0xc9, 0x53, 0x56, 0xf4, 0x7f, 0x44, 0xf3, 0x90,
0xc6, 0xa4, 0xd1, 0x6c, 0x4a, 0x6e, 0x0c, 0x6f,
0xec
};

// g48 (counter = 0)
static readonly byte[] g48_x = {
0x01, 0x04, 0x5c, 0xad, 0x53, 0x5d, 0x92, 0x36,
0xd5, 0x46, 0xd0, 0xa1, 0x38, 0x5a, 0x3e, 0xa4,
0x1e, 0x65, 0x86, 0xee, 0x0a, 0x76, 0xc6, 0xff,
0x5b, 0x81, 0xcc, 0x88, 0x7c, 0xb5, 0x8e, 0xa2,
0x59, 0xe5, 0x52, 0x14, 0xc8, 0x28, 0x91, 0x60,
0xdc, 0x19, 0x21, 0xc4, 0x12, 0xe4, 0x15, 0x8b,
0xe0, 0xcb, 0x54, 0xb8, 0x06, 0x41, 0xcb, 0x65,
0x7c, 0x87, 0x3a, 0x5c, 0xa8, 0x4e, 0xf0, 0xe5,
0xbc, 0x5a
};
static readonly byte[] g48_y = {
0x56, 0xec, 0x9f, 0x00, 0xa5, 0xbd, 0x39, 0xa7,
0x85, 0x78, 0x36, 0x23, 0x5a, 0x83, 0xf2, 0x4b,
0x21, 0xff, 0x89, 0xb8, 0x54, 0x84, 0xc8, 0xa7,
0x95, 0xfe, 0x3e, 0x3b, 0xae, 0xed, 0x85, 0x32,
0xee, 0x2b, 0x94, 0x19, 0x8c, 0x98, 0x82, 0x03,
0x5c, 0x77, 0xb1, 0xe7, 0x50, 0x85, 0xc5, 0xcb,
0x93, 0x6d, 0xf3, 0x06, 0x3e, 0x71, 0x73, 0x59,
0x2e, 0x27, 0x49, 0x09, 0x9a, 0x62, 0x24, 0x7d,
0x6d
};

// g49 (counter = 4)
static readonly byte[] g49_x = {
0x8b, 0x7a, 0xf7, 0x37, 0x8c, 0xea, 0xe8, 0x5d,
0x55, 0x0b, 0xda, 0x52, 0xec, 0x67, 0x44, 0x13,
0x3b, 0xc0, 0xed, 0x59, 0x3d, 0x61, 0x2a, 0x47,
0x18, 0xa4, 0x3b, 0x0f, 0x85, 0xb0, 0x57, 0xf7,
0x9f, 0x43, 0x46, 0x29, 0xb1, 0x70, 0xd2, 0x03,
0xa5, 0x7a, 0xc4, 0xc0, 0x06, 0xdc, 0x4c, 0x87,
0xb5, 0xc9, 0x2f, 0xa7, 0xd4, 0xaf, 0x37, 0xd5,
0xe3, 0x65, 0x1a, 0x14, 0x15, 0x31, 0xfc, 0x15,
0x1a
};
static readonly byte[] g49_y = {
0x09, 0x44, 0x0c, 0x62, 0x16, 0xb4, 0x61, 0x07,
0xb2, 0x53, 0xc4, 0x0b, 0xd7, 0x0f, 0xc5, 0xfc,
0x96, 0xfe, 0x9b, 0x88, 0x6c, 0xf3, 0x29, 0x40,
0x95, 0x8a, 0x80, 0x24, 0xc0, 0x85, 0x95, 0xf9,
0x0a, 0x6d, 0x78, 0x3f, 0x40, 0xa4, 0x8b, 0x1e,
0xa9, 0xbb, 0xc8, 0xe0, 0xad, 0xbb, 0x8e, 0x4d,
0x87, 0x30, 0x0a, 0xca, 0x7f, 0x71, 0x2a, 0x80,
0x20, 0xb0, 0xc4, 0x3d, 0x3b, 0x1f, 0xd7, 0x94,
0xb1
};

// g50 (counter = 2)
static readonly byte[] g50_x = {
0x63, 0x10, 0x5a, 0x1f, 0xf0, 0x44, 0x9d, 0x4d,
0x5d, 0x78, 0x83, 0x95, 0xee, 0x71, 0x41, 0xfd,
0x5c, 0x44, 0xfd, 0x02, 0x27, 0xb3, 0xbd, 0xc3,
0x2b, 0x9e, 0x9d, 0x26, 0xbe, 0x5c, 0x7b, 0x79,
0x45, 0xfa, 0xf8, 0xab, 0x24, 0x7e, 0xed, 0x6e,
0x50, 0xcb, 0x82, 0x4a, 0xbe, 0x7f, 0x5c, 0x4b,
0x7e, 0xd9, 0xed, 0x72, 0x58, 0x92, 0xd2, 0x77,
0xad, 0xa4, 0x62, 0xbb, 0x4e, 0x72, 0xdd, 0x5e,
0x6d
};
static readonly byte[] g50_y = {
0xcb, 0x02, 0x80, 0xe5, 0xe9, 0x29, 0x3d, 0x2d,
0xb0, 0x8d, 0x25, 0xca, 0xad, 0x1e, 0x7b, 0x49,
0xaa, 0x52, 0xfc, 0x85, 0x4f, 0xa4, 0x94, 0x02,
0xe0, 0x5e, 0x4b, 0x6f, 0xbc, 0xf0, 0xef, 0xc1,
0x76, 0x48, 0x34, 0xc1, 0xc4, 0xca, 0x13, 0xa3,
0x6c, 0xa2, 0x0e, 0x8b, 0x8f, 0x57, 0xbd, 0x77,
0xac, 0xa8, 0xf5, 0x2b, 0xe0, 0x6b, 0xd9, 0xa6,
0x92, 0x9f, 0x93, 0xca, 0x17, 0x2d, 0x3d, 0x86,
0xd0
};

// gt (counter = 1)
static readonly byte[] gt_x = {
0xd0, 0xbf, 0xc6, 0x9d, 0x95, 0x7f, 0x2f, 0xc3,
0x8e, 0x51, 0x70, 0xac, 0x3a, 0xae, 0x81, 0x11,
0x0d, 0xcc, 0x7a, 0x07, 0x7c, 0x00, 0x94, 0xdd,
0xd2, 0x9f, 0xf1, 0x20, 0x57, 0xfc, 0xaf, 0x56,
0xe8, 0xd0, 0x14, 0xd0, 0x16, 0x99, 0x8e, 0x44,
0x71, 0x0d, 0xb3, 0xfd, 0xf7, 0x2d, 0xa6, 0x5e,
0x31, 0xcd, 0x66, 0x5a, 0xbc, 0xb3, 0x53, 0x08,
0xa6, 0xb0, 0xac, 0x5f, 0x18, 0xb3, 0xff, 0xb6,
0xf7
};
static readonly byte[] gt_y = {
0xbf, 0xab, 0xa1, 0xb7, 0xea, 0x54, 0x55, 0x2b,
0x93, 0x8c, 0xe8, 0x9d, 0x09, 0x07, 0x79, 0x7f,
0x5f, 0x55, 0xdd, 0x08, 0x1c, 0xa3, 0xfb, 0x5c,
0xf0, 0x1f, 0x26, 0x06, 0xd4, 0x64, 0xe3, 0x6e,
0x3a, 0x37, 0xe0, 0x50, 0xcf, 0xa0, 0xfb, 0x9c,
0xee, 0xe0, 0x35, 0x36, 0x70, 0x7c, 0x6d, 0x11,
0x76, 0x65, 0xb3, 0xb1, 0xe8, 0x34, 0x4d, 0x66,
0x93, 0x9b, 0x29, 0x79, 0x20, 0x04, 0x47, 0x50,
0x53
};

// gd (counter = 2)
static readonly byte[] gd_x = {
0x01, 0x5f, 0x3a, 0xe2, 0xee, 0x57, 0x67, 0x1e,
0x1b, 0xdc, 0x87, 0x04, 0x7f, 0x34, 0xcb, 0xe0,
0xaa, 0xbc, 0x69, 0x3f, 0x46, 0x3b, 0x13, 0xd4,
0x2d, 0xf6, 0xfd, 0x10, 0xf8, 0x6c, 0xc7, 0xb4,
0xb4, 0xb9, 0x99, 0xc9, 0xc3, 0x8b, 0xde, 0xfa,
0x8e, 0x13, 0xc7, 0x01, 0x74, 0xf1, 0xbe, 0x79,
0xf8, 0xe9, 0x34, 0xa3, 0xa0, 0x18, 0x03, 0x07,
0x3b, 0x34, 0x04, 0x6b, 0x8d, 0x8f, 0xfc, 0x79,
0x93, 0x42
};
static readonly byte[] gd_y = {
0xf0, 0x98, 0x4a, 0xc4, 0x3d, 0x1a, 0x34, 0xd5,
0x21, 0x43, 0xf0, 0x9a, 0xbd, 0x17, 0x46, 0xbc,
0xd1, 0x1e, 0x54, 0x95, 0xa0, 0x96, 0xe3, 0x07,
0xd1, 0xab, 0x21, 0x64, 0x8b, 0x11, 0x18, 0xc2,
0x9c, 0x20, 0x1f, 0x29, 0xd0, 0x85, 0xd6, 0xa8,
0xa4, 0xda, 0x6c, 0xf7, 0x60, 0xc8, 0x18, 0xdc,
0xdc, 0x4e, 0xde, 0x01, 0xb4, 0x11, 0x06, 0x5c,
0xac, 0x92, 0xee, 0xa1, 0xcc, 0x51, 0xbd, 0xaf,
0xb2
};

            private static readonly ECGroup ecGroup = ECGroup.CreateECGroup(p, a, b, g_x, g_y, n, ParamSet_EC_P521_V1Name, "P-521");

            internal static readonly Group Gq = ecGroup;

            internal static readonly GroupElement Gd = ecGroup.CreateGroupElement(gd_x, gd_y);

            internal static readonly GroupElement[] IssuerGenerators = {
                ecGroup.CreateGroupElement(g1_x, g1_y),
                ecGroup.CreateGroupElement(g2_x, g2_y),
                ecGroup.CreateGroupElement(g3_x, g3_y),
                ecGroup.CreateGroupElement(g4_x, g4_y),
                ecGroup.CreateGroupElement(g5_x, g5_y),
                ecGroup.CreateGroupElement(g6_x, g6_y),
                ecGroup.CreateGroupElement(g7_x, g7_y),
                ecGroup.CreateGroupElement(g8_x, g8_y),
                ecGroup.CreateGroupElement(g9_x, g9_y),
                ecGroup.CreateGroupElement(g10_x, g10_y),
                ecGroup.CreateGroupElement(g11_x, g11_y),
                ecGroup.CreateGroupElement(g12_x, g12_y),
                ecGroup.CreateGroupElement(g13_x, g13_y),
                ecGroup.CreateGroupElement(g14_x, g14_y),
                ecGroup.CreateGroupElement(g15_x, g15_y),
                ecGroup.CreateGroupElement(g16_x, g16_y),
                ecGroup.CreateGroupElement(g17_x, g17_y),
                ecGroup.CreateGroupElement(g18_x, g18_y),
                ecGroup.CreateGroupElement(g19_x, g19_y),
                ecGroup.CreateGroupElement(g20_x, g20_y),
                ecGroup.CreateGroupElement(g21_x, g21_y),
                ecGroup.CreateGroupElement(g22_x, g22_y),
                ecGroup.CreateGroupElement(g23_x, g23_y),
                ecGroup.CreateGroupElement(g24_x, g24_y),
                ecGroup.CreateGroupElement(g25_x, g25_y),
                ecGroup.CreateGroupElement(g26_x, g26_y),
                ecGroup.CreateGroupElement(g27_x, g27_y),
                ecGroup.CreateGroupElement(g28_x, g28_y),
                ecGroup.CreateGroupElement(g29_x, g29_y),
                ecGroup.CreateGroupElement(g30_x, g30_y),
                ecGroup.CreateGroupElement(g31_x, g31_y),
                ecGroup.CreateGroupElement(g32_x, g32_y),
                ecGroup.CreateGroupElement(g33_x, g33_y),
                ecGroup.CreateGroupElement(g34_x, g34_y),
                ecGroup.CreateGroupElement(g35_x, g35_y),
                ecGroup.CreateGroupElement(g36_x, g36_y),
                ecGroup.CreateGroupElement(g37_x, g37_y),
                ecGroup.CreateGroupElement(g38_x, g38_y),
                ecGroup.CreateGroupElement(g39_x, g39_y),
                ecGroup.CreateGroupElement(g40_x, g40_y),
                ecGroup.CreateGroupElement(g41_x, g41_y),
                ecGroup.CreateGroupElement(g42_x, g42_y),
                ecGroup.CreateGroupElement(g43_x, g43_y),
                ecGroup.CreateGroupElement(g44_x, g44_y),
                ecGroup.CreateGroupElement(g45_x, g45_y),
                ecGroup.CreateGroupElement(g46_x, g46_y),
                ecGroup.CreateGroupElement(g47_x, g47_y),
                ecGroup.CreateGroupElement(g48_x, g48_y),
                ecGroup.CreateGroupElement(g49_x, g49_y),
                ecGroup.CreateGroupElement(g50_x, g50_y),
                ecGroup.CreateGroupElement(gt_x, gt_y)
                };

        }
    }
}
